/* Process model C form file: QospfRouterProcesses.pr.c */
/* Portions of this file copyright 1992, 1996 by MIL 3, Inc. */



/* OPNET system definitions */
#include <opnet.h>
#include "/disk2/gduan/qosip/gwy_qospf/QospfRouterProcesses.pr.h"
FSM_EXT_DECS



/* Header block */
#define TRACE 		if(0)

#include <string.h>
#include <sys/time.h>

/**
*** Interrupts and conditions that make transitions from one state to another
**/
#define ARRIVAL        (op_intrpt_type() == OPC_INTRPT_STRM)
#define END_SIM		   (op_intrpt_type () ==  OPC_INTRPT_ENDSIM)
#define send_ospf_lsa  (op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == 0)
/* Add code 5 by DG, 11/24/97, for invork LSA when receive RSVP msg. */
#define send_qospf_lsa (op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == 1) || \
			(op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == 5) \
			||send_reslsa||send_rralsa
#define send_mospf_lsa (op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == 3) || \
		       (op_intrpt_type() == OPC_INTRPT_REMOTE && op_intrpt_code() == 6)
#define do_mospf_calc  (op_intrpt_type() == OPC_INTRPT_REMOTE && op_intrpt_code() == 7)
#define do_qospf_calc  (op_intrpt_type() == OPC_INTRPT_REMOTE && op_intrpt_code() == 8) || \
		       (op_intrpt_type() == OPC_INTRPT_REMOTE && op_intrpt_code() == 9)
#define hellopk_send   (op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == 2) || \
					   send_hellopk
#define send_lsreq     (op_intrpt_type() == OPC_INTRPT_SELF && op_intrpt_code() == 4) || \
					   send_lsreq_pk

/**
*** Probes that are needed to see the behaviour of the system
**/
#define CUMM_OSPF_LSA_RECV_STAT			0
#define CUMM_MOSPF_LSA_RECV_STAT		1
#define CUMM_QOSPF_LSA_RECV_STAT		2
#define CUMM_TOTAL_LSA_RECV_STAT		3
#define OSPF_DIJ_CALC_STAT				4
#define MOSPF_DIJ_CALC_STAT				5
#define QOSPF_DIJ_CALC_STAT				6
#define TOTAL_DIJ_CALC_STAT				7
#define CUMM_TOTAL_LSA_TRAFFIC_STAT		8
#define CUMM_OSPF_LSA_SENT_STAT			9
#define CUMM_MOSPF_LSA_SENT_STAT		10
#define CUMM_QOSPF_LSA_SENT_STAT		11
#define MAX_MOSPF_DIJ_CALC_STAT			12
#define MIN_MOSPF_DIJ_CALC_STAT			13
#define MAX_QOSPF_DIJ_CALC_STAT			14
#define MIN_QOSPF_DIJ_CALC_STAT			15
#define MAX_OSPF_DIJ_CALC_STAT			16
#define MIN_OSPF_DIJ_CALC_STAT			17

#define INST_OSPF_LSA_RECV_STAT			20
#define INST_MOSPF_LSA_RECV_STAT		21
#define INST_QOSPF_LSA_RECV_STAT		22
#define INST_OSPF_LSA_SENT_STAT			23
#define INST_MOSPF_LSA_SENT_STAT		24
#define INST_QOSPF_LSA_SENT_STAT		25



/** 
*** OSPF Architectural Constants 
**/
int MinLSInterval          = 5;
int MaxAge                 = 60;
int MaxAgeDiff             = 15;
double OspfLsaRefreshTime  = 20; 
double MospfLsaRefreshTime = 20; 
double HelloPkIntvl        = 20;
double QospfLsaRefreshTime = 20;
double LsaReqTime          = 20;
double LSInfinity 		   = 100000;

/** 
*** Link Attributes 
**/
#define LINKBANDWIDTH   		1500000
#define LINKDELAY       		(1/1500000)
#define LINKBUFFERSIZE  		100

#define ER_IS_USED
#define ROUTE_PINNING_IS_USED   1

/**
*** Protocol ID's
**/
#define RSVP_PROTOCOL_ID		46
#define QOSPF_PROTOCOL_ID		89
#define UDP_PROTOCOL_ID			20
#define TCP_PROTOCOL_ID         12
#define IGMP_PROTOCOL_ID        9
#define VERSION       			1

/** 
*** Packet types defintions
**/
#define HELLOPK       			1
#define DDPK          			2
#define LSREQPK       			3
#define OSPFLSUPDATEPK    		4
#define LSACKPK       			5
#define QOSPFLSUPDATEPK   		7
#define MOSPFLSUPDATEPK   		8

/**
*** LSA Types
**/
#define ROUTERLSA     			1
#define NETWORKLSA    			2
#define SUMMARYLSA    			3
#define RRALSA        			15
#define RESLSA        			16
#define GRPMBRSHIPLSA 			6
#define ILNone        			21

/**
*** Maximum Number Limitations on the data structures numbers
**/
#define MAX_NO_NEIGHBORS		10
#define MAX_NO_SPROUTERS		100
#define MAX_NO_LSAS	  		200
#define MAX_NO_RRA_LSAS	  		200
#define MAX_NO_AREAS			5
#define MAX_NO_GRPS             30
#define MAX_NO_NETWORKS         110
#define MaxNoNetsInArea         2
#define MaxNoRoutersInArea      1
#define MAX_INIT_TIME			10
#define MAX_REFRESH_TIME		5
#define MAX_NO_LINKS    		10
#define Router_Start_Index      60

#define OFF           			0
#define ON            			1
#define NONE          			0
#define TRUE          			1
#define FALSE         			0
#define UNDEFINED				-1

/**
*** Debugging Definitions 
**/
#define PROBE_ON				if(ON)
#define PROBE_OFF				if(OFF)
#define VL 					if(OFF)
#define DL 					if(OFF)
#define MDL					if(OFF)
#define ILDL					if(OFF)
#define OPDL 					if(OFF)
#define QPDL 					if(OFF)
#define MPDL 					if(OFF)
#define HPDL 					if(OFF)
#define SPFDL 					if(OFF)
#define MCSPFDL 				if(OFF)
#define QSPFDL 					if(OFF)

#define Tracking_Interval		0.1

/**
*** Streams that are coming to QOSPF Module
**/
#define BROADCASTLOCAL 			1
#define ABRONLY        			2
#define RSVPSTREAM     			2
#define IPSTREAM       			0
#define IGMPSTREAM				1
#define DEF_STREAM				110

#define AllSPFRoutersNetNum 	400
#define AllSPFRoutersNodeNum 	5
#define AllDRoutersNetNum   	400
#define AllDRoutersNodeNum   	6

#define ROUTER					1
#define TRANSITNET				2
#define STUBNET					3

enum Flodding_Scope {AreaLocal = 0, ABR_Only = 1};

/**
*** IP Address Data Structure
**/ 
struct ipAddress{
	int NetNum;
	int NodeNum;
};
typedef struct ipAddress IPAddress;

typedef struct{
    IPAddress SenderAddr;
    IPAddress DestAddr;
    int SenderPort;
    double Bandwidth;
    double Delay;
}sd;
 

/**
*** Structures for the list manipulation 
**/
#ifndef  _LIST_H_
#define  _LIST_H_

#define PrintErr(x)   printf("%s : Filename %s : Line NO %d:\n", \
								x, __FILE__, __LINE__);

struct List {
	struct List * next;
	struct List * curr;
 	char *ele;
};

typedef struct List ListL;

#endif

typedef struct {
	int Version;
	int Type;
	int PacketLength;
	IPAddress ThisRouterId;
	int ThisAreaId;
	int CheckSum;
	int AuType;
	double Authentication;
}ospfpackethdr;

typedef struct {
	int LsAge;
	int Options;
	int LSA_type;
	IPAddress LinkStateId;
	IPAddress AdvRouter;
	int ThisAreaId;
	double LSSeqNo;
	double LSChecksum;
	double Length;
	int rtype;
	int No_of_Links;
}reslsaheader;

typedef struct {
	int LsAge;
	int Options;
	int LSA_type;
	int OpaqueType;
	int OpaqueId;
	IPAddress AdvRouter;
	int ThisAreaId;
	double LSSeqNo;
	double LSChecksum;
	double Length;
	int FloddingScope;
}opaquelsaheader ;

typedef struct {
	int LsAge;
	int LsType;
	IPAddress LinkStateId;
	IPAddress AdvRouter;
	int ThisAreaId;
	double LSSeqNo;
	double LSChecksum;
	double Length;
}linkstateheader;

typedef struct {
	IPAddress destination_Addr;
	IPAddress Source_Addr;
	int src_prefix_length;
	int adjust_offset;
}EraHeader ;

/**
***  QOSPF Info. Structures
**/
typedef struct{
	int Types;
	double Delay;
	double Bandwidth;
	double BufferSize;
	double ReservedBandwidth;
	double ReservedBufferSize;
	double ReservedDelay;
	int IsNghborRou;
	/*int NghborNetNum;      commented by Jiemei.   The NeighborRtrId.NetNum can be used instead of this*/
	IPAddress IntfIPAddr;
	IPAddress NeighborRtrId;
	int CostMetric;
}Linkinfo;

typedef struct {
	int TOS;
	double metric;
}tosinfo;

typedef struct {
	IPAddress LinkId;
	IPAddress LinkData;
	int Type;
	int NoOfTOS;
	int TOSMetric;
	tosinfo Tos_Info[5];
}linkinfo;

typedef struct {
	IPAddress LinkId;
	IPAddress LinkData;
	int Type;
	int TOSMetric;
	double LinkDelay;
	double Bandwidth;
	double BufferSize;
}reslinkinfo ;

typedef struct {
	IPAddress LinkId;
	IPAddress LinkData;
	int Type;
	int PinFlag;
	double ReservedBandwidth;
	double ReservedBufferSize;
}linkresvinfo ;

typedef struct {
	IPAddress router_id;
	double delay;
}AbrRouterInfo ;

typedef struct {
	int incoming_intf_type;
	int MOSPF_IL_type;
	int MOSPF_Init_case;
	int No_Of_outgoing_Intf;
	IPAddress incoming_intf_addr;
	int outgoing_intf_type;
	int child_offset;
	IPAddress outgoing_intf_addr;
}EraBody ;

/* commented by 02/05/98   Never being used
typedef struct {
	struct Opaque_LSA_Header;
	struct ERA_Header;
	EraBody ERA_Body[100];
}EraSubTree ;

*/

typedef struct{
     int VertexType;
     IPAddress VertexId;
}vertexinfo;

typedef struct{
	int interface[MAX_NO_LINKS];
}mctable;
 
/**
*** QOSPF Link State Advertisements
**/
typedef struct {
	linkstateheader LinkStateHeader;
	int bit_b;
	int bit_e;
	int No_of_Links;
	linkinfo Link_Info[MAX_NO_LINKS];
}RouterLinkAdv;

typedef struct {
	linkstateheader LinkStateHeader;
	int ThisNetworkMask;
	IPAddress AttachedRouter[MAX_NO_LINKS];
}NetworkLinkAdv;

typedef struct {
	linkstateheader LinkStateHeader;
	int ThisNetworkMask;
	tosinfo Tos_Info[5];
}SummaryLinkAdv;

typedef struct {
     linkstateheader LinkStateHeader;
     vertexinfo VertexInfo[MAX_NO_LINKS];
}grpmbrlsa;

typedef struct {
	reslsaheader ResLsaHeader;
	reslinkinfo  ResLinkInfo[MAX_NO_LINKS];
}reslsa;

typedef struct {
	opaquelsaheader OpaqueLsaHeader;
	IPAddress DestAddr;
	IPAddress SourceAddr;
	int src_prefix_len;
	int dst_prefix_len;
	int No_of_Links;
	linkresvinfo LinkResvInfo[MAX_NO_LINKS];
}rralsa;

typedef struct {
	long LsType;
	long LinkStateId;
	IPAddress AdvRouter;
	long LsSeqNo;
	short LsCheckSum;
	short LsAge;
}ddlsadv;

typedef struct {
	long LsType;
	long LinkStateId;
	IPAddress AdvRouter;
}lsreqpklsadv;

/** 
*** QOSPF Packet Structures
**/

typedef struct {
	ospfpackethdr OspfPkHeader;
	int ThisNetworkMask;
	int Deadint;
	int Helloint;
	int ThisRtrPri;
	int Indep_Router;
	IPAddress DesgRouter;
	IPAddress BackupDesgRouter;
	IPAddress RtrsFrmWhichHelloRec[MAX_NO_NETWORKS];
}hellopacket;

typedef struct {
	ospfpackethdr OspfPkHeader;
	lsreqpklsadv LSReqPkLsAdv[MAX_NO_LINKS];
}LSReqPk;

typedef struct {
	ospfpackethdr OspfPkHeader;
	int IBit;
	int Mbit;
	int MSbit;
	int DDSeqNo;
	ddlsadv DDLsAdv[MAX_NO_LINKS];
}ddpk;

typedef struct {
	ospfpackethdr OspfPkHeader;
	int NoOfResAdvs;
	int NoOfRraAdvs;
	reslsa          ResLinkStateAdv[MAX_NO_LSAS];
	rralsa          RraLinkStateAdv[MAX_NO_LSAS];
}QospfLSUpdatePk;

typedef struct {
	ospfpackethdr OspfPkHeader;
	int NoOfRouterAdvs;
	int NoOfNetworkAdvs;
	int NoOfSummaryAdvs;
  	RouterLinkAdv   RouterLinkStateAdv[MAX_NO_LSAS];
  	NetworkLinkAdv  NetworkLinkStateAdv[MAX_NO_LSAS];
  	SummaryLinkAdv  SummaryLinkStateAdv[MAX_NO_LSAS];
}OspfLSUpdatePk;

typedef struct {
	ospfpackethdr OspfPkHeader;
	int NoOfGrpMbrAdvs;
  	grpmbrlsa       GrpmbrLinkStateAdv[MAX_NO_LSAS]; 
  	/*  grpmbrlsa       GrpmbrLinkStateAdv[MAX_NO_GRPS][MAX_NO_LSAS];   */
}MospfLSUpdatePk;

/**
*** STRUCTURES THAT ARE NEEDED TO BE MAINTAINED ON EACH ROUTER       
**/
/** 			        
*** Routing Table structure                                
**/
typedef struct {
	int DestinationType;
	IPAddress DestinationId;
	int AddressMask;
	int TypeOfService;
	int AreaId;
	int PathType;
	double cost;
	int ExternalMetricType;
	int Type2Cost;
	IPAddress NextHop;
	IPAddress AdvRouter;
}RoutingTable_Struct;

/**
*** Vertex Structure
**/
typedef struct{
    IPAddress VertexId;
    int DistanceToRoot;
    IPAddress NextHopsList[MAX_NO_NEIGHBORS];
}Vertex_Struct;

/**               
*** Shortest Path tree stucture                               
**/
typedef struct {
	Vertex_Struct SPVertex[MAX_NO_SPROUTERS];
}ShortestPathTree_Struct;

/**              
*** Address Ranges structure                                   
**/
typedef struct {
	int address;
	int mask;
}Address_Range_Struct;

/** 					  
*** Area Data Structure                                  
**/
typedef struct {
	int AreaId;
	Address_Range_Struct AddressRanges[MAX_NO_LSAS];
	int RouterInterfaces;
	rralsa          RRALsaList[MAX_NO_LSAS];
	grpmbrlsa       GrpMbrLsaList[MAX_NO_GRPS][MAX_NO_LSAS];
	RouterLinkAdv   RouterLsaList[MAX_NO_LSAS];
	NetworkLinkAdv  NetworkLsaList[MAX_NO_LSAS];
	SummaryLinkAdv  SummaryLsaList[MAX_NO_LSAS];
	reslsa          ResourceLsaList[MAX_NO_LSAS];
	ShortestPathTree_Struct ShortestPathTree;
	ShortestPathTree_Struct MospfShortestPathTree[MAX_NO_GRPS];
    ShortestPathTree_Struct QospfShortestPathTree[MAX_NO_GRPS];
}Area_Struct;

/**
*** Interface Data Structure
**/
typedef struct {
        int type;
        int state;
        IPAddress IPIntfAddr;
        int IPIntfMask;
        int AreaId;
        int HelloInterval;
        int RouterDeadInterval;
        int InfTransDelay;
        int RouterPriority;
        IPAddress NeighboringRouters[MAX_NO_NEIGHBORS];
        IPAddress DesignatedRouter;
        IPAddress BackupDesgRouter;
        int IntfOutputCosts[MAX_NO_LINKS];
        int RxmtInterval;
} Interface_Data_Str;
 
/**
*** Neighbor Data Structure
**/
typedef struct{
        int state;
        int InactivityTimer;
        int Master;
        int Slave;
        int SequenceNo;
        IPAddress NeighborRouterId;
        int NeighborPriority;
		int IsIndepRou;
        IPAddress NeighborIPAddr;
        IPAddress NghbrDesgRouter;
        IPAddress NghbrBackupDesgRouter;
        ListL LinkStateRetransmissionList;
        ListL DatabaseSummaryList;
        ListL LinkStateRequestList;
} Neighbor_Data_Str;


typedef struct{
	IPAddress VertexIPAddr;
	int	  	DistanceToVertex;
	IPAddress NextHops[MAX_NO_NEIGHBORS];
} CandidateStruct;

typedef struct {  
	int interface[MAX_NO_LINKS];
	int mtu[MAX_NO_LINKS];
	int infValid;		/* Added by DG, 11/8/97, keep consistance with the MC in IP */
}mc_id;

typedef struct{
    int RegNetNum[MAX_NO_NETWORKS];
}regmcnet;

/*  #ifdef IS_AREA_BORDER_ROUTER  */
/*	int NoOfVirtualLinks; */	/* The virtual links configured having this*/ 
							/* Router as an endpoint                   */
/*	int VirtualLinks[MAX_NO_LINKS]; */   /* This is Router Id of the other endpoint */
/*  #endif  */

/**
*** Structures that are needed to process the RSVP protocol messages 
**/

typedef struct commonheader {
	  int Version;
	  int Msg_Type;
	  int SendTTL;
	  double RSVPChecksum;
	  double RSVPLength;
}Common_Header;

typedef struct sessionobject {
	  int Class_Num;
      	  double Length;
	  IPAddress DestAddr;
	  int ProtocolId;
      int Flags;
      int DstPort;
}Session_Object;
 
typedef struct rsvphopobject {
	  int Class_Num;
	  double Length;
	  IPAddress NextHopAddr;
	  double LogicalIntfHandle;
}RsvpHop_Object;

typedef struct sendertmplobject {
	  int Class_Num;
	  double Length;
	  IPAddress SrcAddr;
	  double SrcPort;
}SenderTemplate_Object;

typedef struct sendertspecobject {
	  int Class_Num;
	  double Length;
	  double Bandwidth;
	  double LinkDelay;
}SenderTspec_Object;

typedef struct resvconfirm {
	 int Class_Num;
	 IPAddress RecieverAddr;
	 double Length;
}ResvConfirm_Object;

typedef struct flowspecobject {
	 int Class_Num;
	 double Length;
	 double ReqBandwidth;
	 double ReqLinkDelay;
}FlowSpec_Object;


/* Added by DG, 11/18 */


typedef struct {
	SenderTemplate_Object SenderTempObj;
	SenderTspec_Object SenderTspecObj;
}Sender_Descriptor;


typedef struct {
        int Class_Num;
        double Length;   
        double Refresh_Period;
} TimeVal_Object;


typedef struct {
        int Class_Num;
        double Length;
        IPAddress SrcAddress;
} Scope_Object;

typedef struct {
        int Class_Num;
        double Length;
        int Resv_Style;
} Style_Object;

typedef struct {
	  int Class_Num;
	  double Length;
	  IPAddress SrcAddr;
	  double SrcPort;
}FilterSpec_Object;

typedef struct {
	FlowSpec_Object FlowSpecObj;
	FilterSpec_Object FilterSpecObj;
}Flow_Descriptor_List;


typedef struct {
	Common_Header ComHdr;
	Session_Object SessObj;
	RsvpHop_Object RsvpHopObj;
	TimeVal_Object TimeValObj;
	Sender_Descriptor SenderDes; 
}PathMessage;

#define InLIST_SIZE     10

typedef struct {
	Common_Header ComHdr;
	Session_Object SessObj;
	RsvpHop_Object RsvpHopObj;
	TimeVal_Object TimeValObj;
	ResvConfirm_Object ResvConfObj;
	Scope_Object ScopeObj;
	Style_Object StyleObj;
	Flow_Descriptor_List FlowDesList[InLIST_SIZE]; 
}ResvMessage;


/* end */



typedef struct qospfpathstateblock {
	  Session_Object SessionObject;
	  SenderTemplate_Object SenderTemplateObject;
	  SenderTspec_Object SenderTspecObject;
	  RsvpHop_Object RsvpHopObject;
}Qospf_Path_State_Block;

typedef struct resvconfinfo {
	 Session_Object SessionObject;
	 ResvConfirm_Object ResvConfirmObject;
	 FlowSpec_Object FlowSpecObject;
}Resv_Conf_Info;

/** 
*** Functions prototypes that are being used 	                    
**/
void RsvpPkPro( Packet *x);
void QospfPkPro( Packet *x);

/* Added by DG 11/18, called by RsvpPkPro */
void Path_Msg_Pro( Packet *x );
void Resv_Msg_Pro( Packet *x );

void HelloPk_pro(Packet *x);
void DataDescPk_pro(Packet *x);
void LsReqPk_pro(Packet *x);
void QospfLsUpdatePk_pro(Packet *x);
void MospfLsUpdatePk_pro(Packet *x);
void OspfLsUpdatePk_pro(Packet *x);
void LsUpdatePk_pro(Packet *x);
void LsAckPk_pro(Packet *x);

void Router_Lsa_pro(Packet* x);
void Resource_Lsa_pro(Packet* x);
void RRA_Lsa_pro(Packet* x);
void Network_Lsa_pro(Packet* x);
void Summary_Lsa_pro(Packet* x);
void GrpMbrship_Lsa_pro(Packet* x);
void IgmpJoin_pro(Packet* x);

void trigger_the_route_calc();


void * AddListEle( ListL *lst, char *Ele, int size );
void  DeleteListEle( ListL *lst, char *Ele, int size );
char * GetNextListEle( ListL *lst, int flag );
void ListInit( ListL *lst ); 

/*changed by Jiemei: the struct parameters should be pointers */
void CalculateNextHops( Vertex_Struct*, Vertex_Struct*, RouterLinkAdv* );

int IsNetwork( IPAddress );

/** Print Lsa Functions **/	
void PrintLinkStateHdr( linkstateheader );
void PrintLinkInfo( linkinfo*, int );
void PrintAttRtrs( IPAddress* );
void PrintTosInfo( tosinfo* );
void PrintResLsaHdr( reslsaheader );
void PrintResLinkInfo( reslinkinfo* );
void PrintOpaqueLsaHdr( opaquelsaheader );
void PrintLinkResvInfo( linkresvinfo*, int );
void PrintVertexInfo( vertexinfo* );
void PrintLsa( char*, int );


/*changed by Jiemei:Parameter Vertex_Struct should be pointer*/
void OspfCandidateAddPro( Vertex_Struct*, ListL* );

void OspfSPTreeBuild( ListL* );
void QospfRouterInitialize( int );
void RoutingTableCalc( Area_Struct*, int* );
void CandListInt(int, ListL* );   
void MospfCandAddPro(int, Vertex_Struct*, ListL* );
void MospfCandidateAddPro(int, Vertex_Struct*, ListL* );  
void MMCRoutetableCalc( int, int ); /* last parameter added DG, 11/8, transfer source netnum to the function*/
void QMCRoutetableCalc( int, int ); /* added by DG 11/22 */

void MospfSPFTreeCalc( int, ListL* );
void QospfCandListInt(int, ListL* );
void QospfCandAddPro(int, Vertex_Struct*, ListL* );
void QospfCandidateAddPro(int, Vertex_Struct*, ListL* ); /* Added by DG 12/3 */ 
void QospfSPFTreeCalc( int, ListL* );

/*added by Jiemei */
void MospfSPFTreeCalculation(int, int, int, Objid, int*, mc_id*);

/**
*** Global variables for all the routers
**/
IPAddress AreaBorderRouters[MAX_NO_AREAS];
#if !defined (VOSD_NO_FIN)
#undef	BIN
#undef	BOUT
#define	BIN		_fstack_local_info.last_line_passed = __LINE__ - _block_origin;
#define	BOUT	BIN
#define	BINIT	_fstack_local_info.last_line_passed = 0; _block_origin = __LINE__;
#else
#define	BINIT
#endif /* #if !defined (VOSD_NO_FIN) */



/* State variable definitions */
typedef struct
	{
	FSM_SYS_STATE
	Objid                   		sv_my_id;
	Objid                   		sv_ip_objid;
	Objid                   		sv_rsvp_objid;
	Objid                   		sv_igmp_objid;
	Objid                   		sv_sub_netid;
	Objid                   		sv_ip_id;
	Objid                   		sv_q_id;
	Objid                   		sv_rx_id;
	Objid                   		sv_tx_id;
	Objid                   		sv_dup_lk_id;
	Objid                   		sv_qospf_id;
	int                     		sv_IS_ROUTER;
	int                     		sv_IS_DESIGNATED_ROUTER;
	int                     		sv_IS_AREA_BORDER_ROUTER;
	int                     		sv_net_num0;
	int                     		sv_node_num0;
	int                     		sv_net_num1;
	int                     		sv_node_num1;
	int                     		sv_net_num2;
	int                     		sv_node_num2;
	int                     		sv_net_num3;
	int                     		sv_node_num3;
	int                     		sv_net_num4;
	int                     		sv_node_num4;
	int                     		sv_net_num5;
	int                     		sv_node_num5;
	int                     		sv_net_num6;
	int                     		sv_node_num6;
	int                     		sv_net_num7;
	int                     		sv_node_num7;
	int                     		sv_net_num8;
	int                     		sv_node_num8;
	int                     		sv_net_num9;
	int                     		sv_node_num9;
	int                     		sv_res_ls_seq_no;
	int                     		sv_rra_ls_seq_no;
	int                     		sv_router_ls_seq_no;
	int                     		sv_summary_ls_seq_no;
	int                     		sv_network_ls_seq_no;
	int                     		sv_grpmbr_ls_seq_no;
	double                  		sv_NoOfOspfLsaRecv;
	double                  		sv_NoOfMospfLsaRecv;
	double                  		sv_NoOfQospfLsaRecv;
	double                  		sv_NoOfTotalLsaRecv;
	long                    		sv_OspfDijCalcTime;
	long                    		sv_MospfDijCalcTime;
	long                    		sv_QospfDijCalcTime;
	long                    		sv_TotalDijCalcTime;
	double                  		sv_TotalLsaTraffic;
	double                  		sv_NoOfOspfLsaSent;
	double                  		sv_NoOfMospfLsaSent;
	double                  		sv_NoOfQospfLsaSent;
	long                    		sv_MaxOspfDijCalcTime;
	long                    		sv_MinOspfDijCalcTime;
	long                    		sv_MaxMospfDijCalcTime;
	long                    		sv_MinMospfDijCalcTime;
	long                    		sv_MaxQospfDijCalcTime;
	long                    		sv_MinQospfDijCalcTime;
	int                     		sv_NoOfQospfCalcDone;
	int                     		sv_NoOfMospfCalcDone;
	int                     		sv_NoOfOspfCalcDone;
	int                     		sv_TotalDijCalcDone;
	int                     		sv_send_grpmbrlsa;
	int                     		sv_send_grp_mbr_lsa;
	int                     		sv_send_reslsa;
	int                     		sv_send_rralsa;
	int                     		sv_send_ospflsa;
	int                     		sv_send_lsreq_pk;
	int                     		sv_spfcalc;
	int                     		sv_detupstrnode;
	int                     		sv_send_dabra;
	int                     		sv_send_datadesc;
	int                     		sv_qospfcalc;
	int                     		sv_ospfspfcalc;
	int                     		sv_mospfspfcalc;
	int                     		sv_send_ls_update_pk;
	int                     		sv_send_hellopk;
	int                     		sv_Ip_Network_No;
	int                     		sv_Network_Mask;
	int                     		sv_NoOfLinks;
	int                     		sv_Indep_Rou;
	int                     		sv_AreaId;
	int                     		sv_DeadInt;
	int                     		sv_HelloInt;
	int                     		sv_RtrPri;
	int                     		sv_NetworkMask;
	int                     		sv_No_of_Networks;
	int                     		sv_NoOfVertices;
	int                     		sv_No_of_TOS;
	IPAddress               		sv_Network_Routers[MAX_NO_LINKS];
	double                  		sv_TOS_Metric[5];
	Linkinfo                		sv_LinksInfo[MAX_NO_LINKS];
	IPAddress               		sv_DsgRouterIntfAddr;
	IPAddress               		sv_BackupDsgRouterIntfAddr;
	IPAddress               		sv_DsgRtr_IpAddr;
	IPAddress               		sv_RouterId;
	IPAddress               		sv_DefaultDestination;
	Neighbor_Data_Str       		sv_NeighborsInfo[MAX_NO_NEIGHBORS];
	ListL                   		sv_QospfPSBList;
	ListL                   		sv_ResvConfInfoList;
	ListL                   		sv_NeighborList;
	int                     		sv_NoOfAreasRouterIsConnected;
	Area_Struct             		sv_AreaStructure;
	Area_Struct             		sv_BackboneStructure;
	RoutingTable_Struct     		sv_RoutingTable;
	regmcnet                		sv_RegMcNets[MAX_NO_GRPS];
	mctable                 		sv_mc_rte_table[MAX_NO_GRPS];
	int                     		sv_obt_pk_srcnet;
	int                     		sv_obt_pk_srcnode;
	int                     		sv_NoOfNeighborElements;
	int                     		sv_InstNoLsa;
	double                  		sv_print_time;
	double                  		sv_BCOspfLsa_sim_time;
	double                  		sv_BCQospfLsa_sim_time;
	double                  		sv_BCMospfLsa_sim_time;
	sd                      		sv_SessionDirectory[MAX_NO_GRPS];
	int                     		sv_CurrSess;
	double                  		sv_NoOfInstOspfLsaRecv;
	double                  		sv_NoOfInstMospfLsaRecv;
	double                  		sv_NoOfInstQospfLsaRecv;
	double                  		sv_NoOfInstOspfLsaSent;
	double                  		sv_NoOfInstMospfLsaSent;
	double                  		sv_NoOfInstQospfLsaSent;
	int                     		sv_MRecvLsaCycleFirst;
	int                     		sv_URecvLsaCycleFirst;
	int                     		sv_QRecvLsaCycleFirst;
	int                     		sv_MSentLsaCycleFirst;
	int                     		sv_USentLsaCycleFirst;
	int                     		sv_QSentLsaCycleFirst;
	double                  		sv_OspfLsa_Recv_sim_time;
	double                  		sv_MospfLsa_Recv_sim_time;
	double                  		sv_QospfLsa_Recv_sim_time;
	} QospfRouterProcesses_state;

#define pr_state_ptr            		((QospfRouterProcesses_state*) SimI_Mod_State_Ptr)
#define my_id                   		pr_state_ptr->sv_my_id
#define ip_objid                		pr_state_ptr->sv_ip_objid
#define rsvp_objid              		pr_state_ptr->sv_rsvp_objid
#define igmp_objid              		pr_state_ptr->sv_igmp_objid
#define sub_netid               		pr_state_ptr->sv_sub_netid
#define ip_id                   		pr_state_ptr->sv_ip_id
#define q_id                    		pr_state_ptr->sv_q_id
#define rx_id                   		pr_state_ptr->sv_rx_id
#define tx_id                   		pr_state_ptr->sv_tx_id
#define dup_lk_id               		pr_state_ptr->sv_dup_lk_id
#define qospf_id                		pr_state_ptr->sv_qospf_id
#define IS_ROUTER               		pr_state_ptr->sv_IS_ROUTER
#define IS_DESIGNATED_ROUTER    		pr_state_ptr->sv_IS_DESIGNATED_ROUTER
#define IS_AREA_BORDER_ROUTER   		pr_state_ptr->sv_IS_AREA_BORDER_ROUTER
#define net_num0                		pr_state_ptr->sv_net_num0
#define node_num0               		pr_state_ptr->sv_node_num0
#define net_num1                		pr_state_ptr->sv_net_num1
#define node_num1               		pr_state_ptr->sv_node_num1
#define net_num2                		pr_state_ptr->sv_net_num2
#define node_num2               		pr_state_ptr->sv_node_num2
#define net_num3                		pr_state_ptr->sv_net_num3
#define node_num3               		pr_state_ptr->sv_node_num3
#define net_num4                		pr_state_ptr->sv_net_num4
#define node_num4               		pr_state_ptr->sv_node_num4
#define net_num5                		pr_state_ptr->sv_net_num5
#define node_num5               		pr_state_ptr->sv_node_num5
#define net_num6                		pr_state_ptr->sv_net_num6
#define node_num6               		pr_state_ptr->sv_node_num6
#define net_num7                		pr_state_ptr->sv_net_num7
#define node_num7               		pr_state_ptr->sv_node_num7
#define net_num8                		pr_state_ptr->sv_net_num8
#define node_num8               		pr_state_ptr->sv_node_num8
#define net_num9                		pr_state_ptr->sv_net_num9
#define node_num9               		pr_state_ptr->sv_node_num9
#define res_ls_seq_no           		pr_state_ptr->sv_res_ls_seq_no
#define rra_ls_seq_no           		pr_state_ptr->sv_rra_ls_seq_no
#define router_ls_seq_no        		pr_state_ptr->sv_router_ls_seq_no
#define summary_ls_seq_no       		pr_state_ptr->sv_summary_ls_seq_no
#define network_ls_seq_no       		pr_state_ptr->sv_network_ls_seq_no
#define grpmbr_ls_seq_no        		pr_state_ptr->sv_grpmbr_ls_seq_no
#define NoOfOspfLsaRecv         		pr_state_ptr->sv_NoOfOspfLsaRecv
#define NoOfMospfLsaRecv        		pr_state_ptr->sv_NoOfMospfLsaRecv
#define NoOfQospfLsaRecv        		pr_state_ptr->sv_NoOfQospfLsaRecv
#define NoOfTotalLsaRecv        		pr_state_ptr->sv_NoOfTotalLsaRecv
#define OspfDijCalcTime         		pr_state_ptr->sv_OspfDijCalcTime
#define MospfDijCalcTime        		pr_state_ptr->sv_MospfDijCalcTime
#define QospfDijCalcTime        		pr_state_ptr->sv_QospfDijCalcTime
#define TotalDijCalcTime        		pr_state_ptr->sv_TotalDijCalcTime
#define TotalLsaTraffic         		pr_state_ptr->sv_TotalLsaTraffic
#define NoOfOspfLsaSent         		pr_state_ptr->sv_NoOfOspfLsaSent
#define NoOfMospfLsaSent        		pr_state_ptr->sv_NoOfMospfLsaSent
#define NoOfQospfLsaSent        		pr_state_ptr->sv_NoOfQospfLsaSent
#define MaxOspfDijCalcTime      		pr_state_ptr->sv_MaxOspfDijCalcTime
#define MinOspfDijCalcTime      		pr_state_ptr->sv_MinOspfDijCalcTime
#define MaxMospfDijCalcTime     		pr_state_ptr->sv_MaxMospfDijCalcTime
#define MinMospfDijCalcTime     		pr_state_ptr->sv_MinMospfDijCalcTime
#define MaxQospfDijCalcTime     		pr_state_ptr->sv_MaxQospfDijCalcTime
#define MinQospfDijCalcTime     		pr_state_ptr->sv_MinQospfDijCalcTime
#define NoOfQospfCalcDone       		pr_state_ptr->sv_NoOfQospfCalcDone
#define NoOfMospfCalcDone       		pr_state_ptr->sv_NoOfMospfCalcDone
#define NoOfOspfCalcDone        		pr_state_ptr->sv_NoOfOspfCalcDone
#define TotalDijCalcDone        		pr_state_ptr->sv_TotalDijCalcDone
#define send_grpmbrlsa          		pr_state_ptr->sv_send_grpmbrlsa
#define send_grp_mbr_lsa        		pr_state_ptr->sv_send_grp_mbr_lsa
#define send_reslsa             		pr_state_ptr->sv_send_reslsa
#define send_rralsa             		pr_state_ptr->sv_send_rralsa
#define send_ospflsa            		pr_state_ptr->sv_send_ospflsa
#define send_lsreq_pk           		pr_state_ptr->sv_send_lsreq_pk
#define spfcalc                 		pr_state_ptr->sv_spfcalc
#define detupstrnode            		pr_state_ptr->sv_detupstrnode
#define send_dabra              		pr_state_ptr->sv_send_dabra
#define send_datadesc           		pr_state_ptr->sv_send_datadesc
#define qospfcalc               		pr_state_ptr->sv_qospfcalc
#define ospfspfcalc             		pr_state_ptr->sv_ospfspfcalc
#define mospfspfcalc            		pr_state_ptr->sv_mospfspfcalc
#define send_ls_update_pk       		pr_state_ptr->sv_send_ls_update_pk
#define send_hellopk            		pr_state_ptr->sv_send_hellopk
#define Ip_Network_No           		pr_state_ptr->sv_Ip_Network_No
#define Network_Mask            		pr_state_ptr->sv_Network_Mask
#define NoOfLinks               		pr_state_ptr->sv_NoOfLinks
#define Indep_Rou               		pr_state_ptr->sv_Indep_Rou
#define AreaId                  		pr_state_ptr->sv_AreaId
#define DeadInt                 		pr_state_ptr->sv_DeadInt
#define HelloInt                		pr_state_ptr->sv_HelloInt
#define RtrPri                  		pr_state_ptr->sv_RtrPri
#define NetworkMask             		pr_state_ptr->sv_NetworkMask
#define No_of_Networks          		pr_state_ptr->sv_No_of_Networks
#define NoOfVertices            		pr_state_ptr->sv_NoOfVertices
#define No_of_TOS               		pr_state_ptr->sv_No_of_TOS
#define Network_Routers         		pr_state_ptr->sv_Network_Routers
#define TOS_Metric              		pr_state_ptr->sv_TOS_Metric
#define LinksInfo               		pr_state_ptr->sv_LinksInfo
#define DsgRouterIntfAddr       		pr_state_ptr->sv_DsgRouterIntfAddr
#define BackupDsgRouterIntfAddr 		pr_state_ptr->sv_BackupDsgRouterIntfAddr
#define DsgRtr_IpAddr           		pr_state_ptr->sv_DsgRtr_IpAddr
#define RouterId                		pr_state_ptr->sv_RouterId
#define DefaultDestination      		pr_state_ptr->sv_DefaultDestination
#define NeighborsInfo           		pr_state_ptr->sv_NeighborsInfo
#define QospfPSBList            		pr_state_ptr->sv_QospfPSBList
#define ResvConfInfoList        		pr_state_ptr->sv_ResvConfInfoList
#define NeighborList            		pr_state_ptr->sv_NeighborList
#define NoOfAreasRouterIsConnected		pr_state_ptr->sv_NoOfAreasRouterIsConnected
#define AreaStructure           		pr_state_ptr->sv_AreaStructure
#define BackboneStructure       		pr_state_ptr->sv_BackboneStructure
#define RoutingTable            		pr_state_ptr->sv_RoutingTable
#define RegMcNets               		pr_state_ptr->sv_RegMcNets
#define mc_rte_table            		pr_state_ptr->sv_mc_rte_table
#define obt_pk_srcnet           		pr_state_ptr->sv_obt_pk_srcnet
#define obt_pk_srcnode          		pr_state_ptr->sv_obt_pk_srcnode
#define NoOfNeighborElements    		pr_state_ptr->sv_NoOfNeighborElements
#define InstNoLsa               		pr_state_ptr->sv_InstNoLsa
#define print_time              		pr_state_ptr->sv_print_time
#define BCOspfLsa_sim_time      		pr_state_ptr->sv_BCOspfLsa_sim_time
#define BCQospfLsa_sim_time     		pr_state_ptr->sv_BCQospfLsa_sim_time
#define BCMospfLsa_sim_time     		pr_state_ptr->sv_BCMospfLsa_sim_time
#define SessionDirectory        		pr_state_ptr->sv_SessionDirectory
#define CurrSess                		pr_state_ptr->sv_CurrSess
#define NoOfInstOspfLsaRecv     		pr_state_ptr->sv_NoOfInstOspfLsaRecv
#define NoOfInstMospfLsaRecv    		pr_state_ptr->sv_NoOfInstMospfLsaRecv
#define NoOfInstQospfLsaRecv    		pr_state_ptr->sv_NoOfInstQospfLsaRecv
#define NoOfInstOspfLsaSent     		pr_state_ptr->sv_NoOfInstOspfLsaSent
#define NoOfInstMospfLsaSent    		pr_state_ptr->sv_NoOfInstMospfLsaSent
#define NoOfInstQospfLsaSent    		pr_state_ptr->sv_NoOfInstQospfLsaSent
#define MRecvLsaCycleFirst      		pr_state_ptr->sv_MRecvLsaCycleFirst
#define URecvLsaCycleFirst      		pr_state_ptr->sv_URecvLsaCycleFirst
#define QRecvLsaCycleFirst      		pr_state_ptr->sv_QRecvLsaCycleFirst
#define MSentLsaCycleFirst      		pr_state_ptr->sv_MSentLsaCycleFirst
#define USentLsaCycleFirst      		pr_state_ptr->sv_USentLsaCycleFirst
#define QSentLsaCycleFirst      		pr_state_ptr->sv_QSentLsaCycleFirst
#define OspfLsa_Recv_sim_time   		pr_state_ptr->sv_OspfLsa_Recv_sim_time
#define MospfLsa_Recv_sim_time  		pr_state_ptr->sv_MospfLsa_Recv_sim_time
#define QospfLsa_Recv_sim_time  		pr_state_ptr->sv_QospfLsa_Recv_sim_time

/* This macro definition will define a local variable called	*/
/* "op_sv_ptr" in each function containing a FIN statement.	*/
/* This variable points to the state variable data structure,	*/
/* and can be used from a C debugger to display their values.	*/
#undef FIN_PREAMBLE
#define FIN_PREAMBLE	QospfRouterProcesses_state *op_sv_ptr = pr_state_ptr;




/* Process model interrupt handling procedure */



void
QospfRouterProcesses ()
	{
	
		int in_strm;
		int ip_strm;
		int data_len;
	
		int tmp_msg_type;
		int protocolid;
	
		int NeighborFoundFlag;
	
		Packet *ReceivedPkptr;
		Packet *ReceivedRsvpPkptr;
		Packet *ReceivedQospfPkptr;
	
		mc_id           *mc_table;
	
		double tmp_sim_time;
	int _block_origin = 0;


	FSM_ENTER (QospfRouterProcesses)

	FSM_BLOCK_SWITCH
		{
		/*---------------------------------------------------------*/
		/** state (init) enter executives **/
		FSM_STATE_ENTER_FORCED (0, state0_enter_exec, "init", "QospfRouterProcesses () [init enter execs]")
			{
			int i;
			my_id = op_id_self();
			 
			IS_ROUTER = 0;
			IS_DESIGNATED_ROUTER = 0;
			IS_AREA_BORDER_ROUTER = 0;
			
			op_ima_obj_attr_get( my_id, "AreaId", &AreaId );
			op_ima_obj_attr_get( my_id, "IsRouter", &IS_ROUTER );
			op_ima_obj_attr_get( my_id, "AreaBorderRouter", &IS_AREA_BORDER_ROUTER );
			op_ima_obj_attr_get( my_id, "DesignatedRouter", &IS_DESIGNATED_ROUTER );
			
			/*ip_objid   = op_topo_assoc( my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, 0 );
			rsvp_objid = op_topo_assoc( my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, 1 );
			igmp_objid = op_topo_assoc( my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, 2 );
			commented by Jiemei*/
			 
			/*Changed by Jiemei*/
			ip_objid = op_topo_assoc(my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, IPSTREAM);
			rsvp_objid = op_topo_assoc(my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, RSVPSTREAM);
			igmp_objid = op_topo_assoc(my_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, IGMPSTREAM);
			
			/**
			***  Using this IP layer Object Id get all the required attributes from the IP layer
			**/
			op_ima_obj_attr_get( ip_objid, "net_num0", &net_num0 ); 
			op_ima_obj_attr_get( ip_objid, "node_num0", &node_num0 );
			op_ima_obj_attr_get( ip_objid, "net_num1", &net_num1 );
			op_ima_obj_attr_get( ip_objid, "node_num1", &node_num1 );
			op_ima_obj_attr_get( ip_objid, "net_num2", &net_num2 );
			op_ima_obj_attr_get( ip_objid, "node_num2", &node_num2 );
			op_ima_obj_attr_get( ip_objid, "net_num3", &net_num3 );
			op_ima_obj_attr_get( ip_objid, "node_num3", &node_num3 );
			op_ima_obj_attr_get( ip_objid, "net_num4", &net_num4 );
			op_ima_obj_attr_get( ip_objid, "node_num4", &node_num4 );
			op_ima_obj_attr_get( ip_objid, "net_num5", &net_num5 );
			op_ima_obj_attr_get( ip_objid, "node_num5", &node_num5 );
			op_ima_obj_attr_get( ip_objid, "net_num6", &net_num6 );
			op_ima_obj_attr_get( ip_objid, "node_num6", &node_num6 );
			op_ima_obj_attr_get( ip_objid, "net_num7", &net_num7 );
			op_ima_obj_attr_get( ip_objid, "node_num7", &node_num7 );
			op_ima_obj_attr_get( ip_objid, "net_num8", &net_num8 );
			op_ima_obj_attr_get( ip_objid, "node_num8", &node_num8 );
			op_ima_obj_attr_get( ip_objid, "net_num9", &net_num9 );
			op_ima_obj_attr_get( ip_objid, "node_num9", &node_num9 );
			op_ima_obj_attr_get( ip_objid, "Indep_Rou", &Indep_Rou );
			
			
			if( !Indep_Rou )
				sub_netid=op25_topo_parent( op25_topo_parent(my_id) );
			else  
			{
				sub_netid=op25_topo_parent( my_id );
			}
			
			/*** Initialize all the variables that are used for LSA            added by Jiemei***/
			NoOfLinks = op_topo_assoc_count( sub_netid,OPC_TOPO_ASSOC_IN,OPC_OBJTYPE_LKSIMP );
			No_of_Networks=NoOfLinks;
			DeadInt = 0;
			HelloInt = 0;
			RtrPri = 0;
			NetworkMask = 0;
			NoOfVertices = 0;   /* never used in this module*/
			No_of_TOS = 0;
			
			/*
			if( !Indep_Rou )
				NoOfLinks++;
			*/
			
			memset( LinksInfo, 0, sizeof(Linkinfo)*MAX_NO_LINKS );
			memset ( Network_Routers, 0, sizeof(IPAddress)*MAX_NO_LINKS );   /* added by Jiemei*/
			memset ( TOS_Metric, 0, sizeof(double)*5);  /*added by Jiemei*/
			
			RouterId.NetNum = net_num0;
			RouterId.NodeNum = node_num0;
			
			if(IS_AREA_BORDER_ROUTER)
			{
				AreaBorderRouters[AreaId] = RouterId;
			}
			
			QospfRouterInitialize(0);
			
			op_intrpt_schedule_self( op_sim_time()+2, 0 );  /* ospf  */
			/*op_intrpt_schedule_self( op_sim_time()+2, 1 );*/  /* qospf */
			op_intrpt_schedule_self( op_sim_time()+1, 2 );  /* hello */
			op_intrpt_schedule_self( op_sim_time()+10, 3 ); /* mospf */ 
			/*op_intrpt_schedule_self( op_sim_time()+1, 4 ); */	/* lsreq */
			
			BCMospfLsa_sim_time = MospfLsaRefreshTime;
			
			NoOfNeighborElements = 0;
			
			ListInit( &NeighborList );
			
			router_ls_seq_no = 0;
			grpmbr_ls_seq_no =0;
			InstNoLsa = 0;
			send_ls_update_pk = 0;
			send_grp_mbr_lsa = 0;
			print_time = 0;
			/* Added by DG, 11/25/97, init RSVP lsa seq no. */
			res_ls_seq_no = 0;
			rra_ls_seq_no = 0;
			
			/**
			*** Initialize all the variables that are used for probes
			**/
				NoOfOspfLsaRecv = 0;
				NoOfMospfLsaRecv = 0;
				NoOfQospfLsaRecv = 0;
				NoOfTotalLsaRecv = 0;
				OspfDijCalcTime = 0;
				MospfDijCalcTime = 0;
				QospfDijCalcTime = 0;
				TotalDijCalcTime = 0;
				TotalLsaTraffic = 0;
				NoOfOspfLsaSent = 0;
				NoOfMospfLsaSent = 0;
				NoOfQospfLsaSent = 0;
				 
				MaxOspfDijCalcTime = 0;
				MinOspfDijCalcTime = 0;
				MaxMospfDijCalcTime = 0;
				MinMospfDijCalcTime = 0;
				MaxQospfDijCalcTime = 0;
				MinQospfDijCalcTime = 0;
				NoOfQospfCalcDone = 0;
				NoOfMospfCalcDone = 0;
				NoOfOspfCalcDone = 0;
				TotalDijCalcDone = 0;
			
			/* Added by DG and Jiemei, 11/27/97 */
				NoOfInstOspfLsaRecv = 0;
				NoOfInstMospfLsaRecv = 0;
				NoOfInstQospfLsaRecv = 0;
				NoOfInstOspfLsaSent = 0;
				NoOfInstMospfLsaSent = 0;
				NoOfInstQospfLsaSent = 0;
			
			
				MRecvLsaCycleFirst = 0;
				URecvLsaCycleFirst = 0;
				QRecvLsaCycleFirst = 0;
				MSentLsaCycleFirst = 0;
				USentLsaCycleFirst = 0;
				QSentLsaCycleFirst = 0;
			
				OspfLsa_Recv_sim_time = MAX_INIT_TIME;
				MospfLsa_Recv_sim_time = 0;
				QospfLsa_Recv_sim_time = MAX_INIT_TIME;
			/* Add end */
			
			VL	ILDL	
			{
				printf("--------------  INIT BLOCK PROCESSING  ----------------\n");
				printf("Router(%d.%d) Link Information \n", RouterId.NetNum, RouterId.NodeNum );
				printf("Number of Links is %d\n ", NoOfLinks);
				for( i = 0; i < NoOfLinks; i++ )
				{
					printf("Network_Router[%d]=%d.%d  ",i,Network_Routers[i].NetNum,Network_Routers[i].NodeNum);
					printf("LinksInfo[%d].NeighborRtrId=%d.%d  ",i, 
							LinksInfo[i].NeighborRtrId.NetNum, LinksInfo[i].NeighborRtrId.NodeNum );
					printf("LinksInfo[%d].IntfIPAddr=%d.%d  ",i, 
							 LinksInfo[i].IntfIPAddr.NetNum, LinksInfo[i].IntfIPAddr.NodeNum );
					printf("Link Cost=%d\n", LinksInfo[i].CostMetric );
				}
				printf("------------  END INIT BLOCK PROCESSING  --------------\n");
			}
			/* added by Jiemei*/
			memset( &(AreaStructure), 0, sizeof(Area_Struct) );
			
			}


		/** state (init) exit executives **/
		FSM_STATE_EXIT_FORCED (0, state0_exit_exec, "init", "QospfRouterProcesses () [init exit execs]")
			{
			}


		/** state (init) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (idle) enter executives **/
		FSM_STATE_ENTER_UNFORCED (1, state1_enter_exec, "idle", "QospfRouterProcesses () [idle enter execs]")
			{
			/*added by Jiemei */
			
			TRACE if (op_sim_time()>9) printf("Qospf: idle state\n");
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (3,QospfRouterProcesses)


		/** state (idle) exit executives **/
		FSM_STATE_EXIT_UNFORCED (1, state1_exit_exec, "idle", "QospfRouterProcesses () [idle exit execs]")
			{
			}


		/** state (idle) transition processing **/
		FSM_INIT_COND (ARRIVAL)
		FSM_TEST_COND (send_ospf_lsa)
		FSM_TEST_COND (send_qospf_lsa)
		FSM_TEST_COND (hellopk_send)
		FSM_TEST_COND (send_lsreq)
		FSM_TEST_COND (send_datadesc)
		FSM_TEST_COND (send_mospf_lsa)
		FSM_TEST_COND (do_mospf_calc)
		FSM_TEST_COND (do_qospf_calc)
		FSM_TEST_LOGIC ("idle")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 4, state4_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 9, state9_enter_exec, ;)
			FSM_CASE_TRANSIT (2, 8, state8_enter_exec, ;)
			FSM_CASE_TRANSIT (3, 11, state11_enter_exec, ;)
			FSM_CASE_TRANSIT (4, 12, state12_enter_exec, ;)
			FSM_CASE_TRANSIT (5, 14, state14_enter_exec, ;)
			FSM_CASE_TRANSIT (6, 10, state10_enter_exec, ;)
			FSM_CASE_TRANSIT (7, 5, state5_enter_exec, ;)
			FSM_CASE_TRANSIT (8, 2, state2_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (qospfspfcalc) enter executives **/
		FSM_STATE_ENTER_FORCED (2, state2_enter_exec, "qospfspfcalc", "QospfRouterProcesses () [qospfspfcalc enter execs]")
			{
			int s_net;
			int GrpId;
			Vertex_Struct* VertexV;
			ListL QospfCandidatesList;  /*changed by Jiemei from a pointer to struct*/
			double qospfcalc_time = 0;
			long diff;
			struct timeval enter, leave;
			FILE* fp;
			 
			gettimeofday(&enter, (struct timezone*)0);
			
			
			if(do_qospf_calc)
			{
				s_net = *((int*) op_ima_obj_svar_get(ip_objid, "SourceNet"));
				GrpId = *((int*) op_ima_obj_svar_get(ip_objid, "GrpId"));
			}
			else
			{
				s_net = 0;
				GrpId = 0;
			}
			
			VL
			{
				printf("----Router(%d.%d) doing the Qospf Dijkstra Calculation T=%fs\n", 
					RouterId.NetNum,RouterId.NodeNum, op_sim_time() );
			}
			
			
			ListInit( &QospfCandidatesList );
			
			
			CandListInt( s_net, &QospfCandidatesList ); 
			
			if( !IsListEmpty( &QospfCandidatesList ) )
			{
				VertexV = (Vertex_Struct*) GetNextListEle(&QospfCandidatesList, 1 );
			}
			
			
			
			QospfCandidateAddPro( GrpId, VertexV, &QospfCandidatesList );
			QospfSPFTreeCalc( GrpId, &QospfCandidatesList );
			QMCRoutetableCalc( GrpId, s_net ); 
			
			
			VL
			{
				printf("----Router(%d.%d) finish the Qospf Dijkstra Calculation\n", RouterId.NetNum,RouterId.NodeNum );
			}
			
			
			
			/*qospfcalc = 0;
			detupstrnode = 1;     commented by Jiemei*/
			
			NoOfQospfCalcDone++;
			TotalDijCalcDone++;
			
			gettimeofday(&leave, (struct timezone*)0);
			 
			        
			if( enter.tv_usec > leave.tv_usec)
				diff = (((leave.tv_sec - 1) - enter.tv_sec)*1000000) + (leave.tv_usec + 1000000) - enter.tv_usec;
			else
				diff = ((leave.tv_sec - enter.tv_sec)*1000000) + leave.tv_usec - enter.tv_usec;
			
			        PROBE_ON
			        {   
			                if( diff > MaxQospfDijCalcTime )
			                {
			                        MaxQospfDijCalcTime = diff;
			                        op_stat_local_write( MAX_QOSPF_DIJ_CALC_STAT, MinQospfDijCalcTime );
			                }
			                else if( diff < MinQospfDijCalcTime )
			                {
			                        MinQospfDijCalcTime = diff;
			                        op_stat_local_write( MIN_QOSPF_DIJ_CALC_STAT, MinQospfDijCalcTime );
			                }
			 
			                QospfDijCalcTime = QospfDijCalcTime + diff;
			                TotalDijCalcTime = TotalDijCalcTime + diff;
			 
			                op_stat_local_write( QOSPF_DIJ_CALC_STAT, QospfDijCalcTime/NoOfQospfCalcDone );
			                op_stat_local_write( TOTAL_DIJ_CALC_STAT, TotalDijCalcTime/TotalDijCalcDone );
			
					if(op_sim_time() > print_time + 50*Tracking_Interval)
					{
			/*   		fp = fopen("../LogFiles/QospfDijCalcTime.log", "rw" );
			    		if(!fp)
			        		op_sim_end("Unable to open the file QospfDijCalcTime.log....  ","","","");
			    		fprintf(fp, "Current Simulation time is %f. Average time taken for doing QOSPF "
			                	"Dijikstra Algorithm is %f \n",
			                	op_sim_time(), QospfDijCalcTime/NoOfQospfCalcDone );
			    		fclose(fp); */
			 		
			    		print_time = op_sim_time();
					}
				}
			/*added by Jiemei*/
			qospfcalc=0;
			detupstrnode=1; 
			}


		/** state (qospfspfcalc) exit executives **/
		FSM_STATE_EXIT_FORCED (2, state2_exit_exec, "qospfspfcalc", "QospfRouterProcesses () [qospfspfcalc exit execs]")
			{
			}


		/** state (qospfspfcalc) transition processing **/
		FSM_TRANSIT_ONLY ((detupstrnode), 13, state13_enter_exec, ;, "qospfspfcalc")
		/*---------------------------------------------------------*/



		/** state (dabra) enter executives **/
		FSM_STATE_ENTER_FORCED (3, state3_enter_exec, "dabra", "QospfRouterProcesses () [dabra enter execs]")
			{
			
			send_dabra = 0;
			}


		/** state (dabra) exit executives **/
		FSM_STATE_EXIT_FORCED (3, state3_exit_exec, "dabra", "QospfRouterProcesses () [dabra exit execs]")
			{
			}


		/** state (dabra) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (arr) enter executives **/
		FSM_STATE_ENTER_FORCED (4, state4_enter_exec, "arr", "QospfRouterProcesses () [arr enter execs]")
			{
			
			/**  
			***  The Arrival state will check for the packet protocol Id field that is in the incoming IP A
			***  datagram. Depending on the protocol ID the packet is known as either RSVP or QOSPF packet 
			***  and the appropriate functions are called
			**/
			
			in_strm = op_intrpt_strm();
			
			ReceivedPkptr = op_pk_get( in_strm );
			/*if(in_strm == IPSTREAM )
			{*/
				op_pk_nfd_get( ReceivedPkptr, "strm", &ip_strm );
				op_pk_nfd_get( ReceivedPkptr, "prot_id", &protocolid );
				op_pk_nfd_get( ReceivedPkptr, "src_net", &obt_pk_srcnet );
				op_pk_nfd_get( ReceivedPkptr, "src_node", &obt_pk_srcnode );
			/*} */
			
			/*
			VL	DL	printf( "Router(%d.%d). A Packet is obtained from the router %d.%d through the stream %d "
							"with the protocol Id of %d \n", RouterId.NetNum, RouterId.NodeNum, 
							 obt_pk_srcnet, obt_pk_srcnode, ip_strm, protocolid );
			*/
			switch (in_strm)
			{
				case IGMPSTREAM : 
			VL DL		printf("--------- (%d.%d) Igmp Packet Arrived -------- \n", 
							RouterId.NetNum, RouterId.NodeNum );
					send_grp_mbr_lsa = TRUE;
				break;
				case RSVPSTREAM :
			VL DL		printf("--------- (%d.%d) RSVP Packet Arrived, Protocol %d -------- \n", 
							RouterId.NetNum, RouterId.NodeNum, protocolid );
				break;
				default :
				break;
			}
			
			
			switch(protocolid) 
			{
				case QOSPF_PROTOCOL_ID  :   /* changed to symbol by Jiemei */  
					op_pk_nfd_get(ReceivedPkptr, "data", &ReceivedQospfPkptr);
					QospfPkPro( ReceivedQospfPkptr );
				break;
				case RSVP_PROTOCOL_ID  :    /* changed to symbol by Jiemei */
			VL DL		printf(" ^^^ (%d.%d) Recived an RSVP packet \n",
			                	RouterId.NetNum,RouterId.NodeNum );
					op_pk_nfd_get(ReceivedPkptr, "data", &ReceivedRsvpPkptr);
					RsvpPkPro( ReceivedRsvpPkptr );
				break;
				default  :  
				break;
			}
			
			op_pk_destroy( ReceivedPkptr );
			}


		/** state (arr) exit executives **/
		FSM_STATE_EXIT_FORCED (4, state4_exit_exec, "arr", "QospfRouterProcesses () [arr exit execs]")
			{
			}


		/** state (arr) transition processing **/
		FSM_INIT_COND (ospfspfcalc)
		FSM_DFLT_COND
		FSM_TEST_LOGIC ("arr")

		FSM_TRANSIT_SWITCH
			{
			FSM_CASE_TRANSIT (0, 7, state7_enter_exec, ;)
			FSM_CASE_TRANSIT (1, 1, state1_enter_exec, ;)
			}
		/*---------------------------------------------------------*/



		/** state (mospfspfcalc) enter executives **/
		FSM_STATE_ENTER_FORCED (5, state5_enter_exec, "mospfspfcalc", "QospfRouterProcesses () [mospfspfcalc enter execs]")
			{
			{
				int s_net;
				int GrpId;
				ListL MospfCandidatesList;    /*changed by Jiemei from pointer to struct*/
				/*Vertex_Struct* VertexV;       commented by Jiemei */
				double mospfcalc_time = 0;
				long diff;
				struct timeval enter, leave;
				FILE* fp;
			
				gettimeofday(&enter, (struct timezone*)0);
			
				if(do_mospf_calc)
				{
					s_net = *((int*) op_ima_obj_svar_get(ip_objid, "SourceNet"));
					GrpId = *((int*) op_ima_obj_svar_get(ip_objid, "GrpId"));
				}
				else 
				{
					s_net = 0;
					GrpId = 0;
				}
			
				VL	MCSPFDL
				{
					printf("Doing Mospf Dijikstra's Algorithm \n");
				}
			/* line added by Jiemei */
				memset( &(AreaStructure.MospfShortestPathTree), 0, sizeof( ShortestPathTree_Struct ) );
			
				ListInit( &MospfCandidatesList );
			
				CandListInt(s_net, &MospfCandidatesList );   
			
				/*if( !IsListEmpty( &MospfCandidatesList ) )
				{
					VertexV = (Vertex_Struct*) GetNextListEle( &MospfCandidatesList, 1 );
				}
			
				MospfCandAddPro( GrpId, VertexV, &MospfCandidatesList ); commented by Jiemei*/
			
				MospfSPFTreeCalc( GrpId, &MospfCandidatesList );
			
			/* Add s_net parameter by DG, 11/8, let MCRoutetableCalc set infValid of the group to s_net */
				MMCRoutetableCalc( GrpId, s_net );
			
				detupstrnode = 1;
				mospfspfcalc = 0;
			
				NoOfMospfCalcDone++;
				TotalDijCalcDone++;
			
				gettimeofday(&leave, (struct timezone*)0);
			
				if( enter.tv_usec > leave.tv_usec)
					diff = (((leave.tv_sec - 1) - enter.tv_sec)*1000000) + (leave.tv_usec + 1000000) - enter.tv_usec;
				else
					diff = ((leave.tv_sec - enter.tv_sec)*1000000) + leave.tv_usec - enter.tv_usec;
			
				PROBE_ON
				{
					if( diff > MaxMospfDijCalcTime )
					{
						MaxMospfDijCalcTime = diff;
						op_stat_local_write( MAX_MOSPF_DIJ_CALC_STAT, MinMospfDijCalcTime );
					}
					else if( diff < MinMospfDijCalcTime )
					{
						MinMospfDijCalcTime = diff;
						op_stat_local_write( MIN_MOSPF_DIJ_CALC_STAT, MinMospfDijCalcTime );
					}
			
					MospfDijCalcTime = MospfDijCalcTime + diff;
					TotalDijCalcTime = TotalDijCalcTime + diff;
			
					op_stat_local_write( MOSPF_DIJ_CALC_STAT, MospfDijCalcTime/NoOfMospfCalcDone );
					op_stat_local_write( TOTAL_DIJ_CALC_STAT, TotalDijCalcTime/TotalDijCalcDone );
			
			
					if(op_sim_time() > print_time + 50*Tracking_Interval)
					{
			/*			fp = fopen("../LogFiles/MospfDijCalcTime.log", "a" );
						if(!fp)
							op_sim_end("Unable to open the file MospfDijCalcTime.log....  ","","","");
						fprintf(fp, "Average time spent by Router %d.%d for MOSPF Dijkstra's Calculation is %ld. "
									"Current Simulation time is %f \n", RouterId.NetNum, RouterId.NodeNum, 
									MospfDijCalcTime/NoOfMospfCalcDone, op_sim_time() );
						fclose(fp); */
						print_time = op_sim_time();
					}
			
					if( op_sim_time() > 60 )
					{
			/*			fp = fopen("../LogFiles/MospfDijCalcTime.log", "a" );
						if(!fp)
							op_sim_end("Unable to open the file MospfDijCalcTime.log....  ","","","");
						fprintf(fp, "Total Number of MOSPF Dijkstra's done by Router %d.%d is %d. Total time"
									"spent for these calculation is %ld\n", RouterId.NetNum, RouterId.NodeNum, 
									TotalDijCalcDone, MospfDijCalcTime );
						fclose(fp); */		/* Added by DG 11/12 */
					}
				}
			}
			}


		/** state (mospfspfcalc) exit executives **/
		FSM_STATE_EXIT_FORCED (5, state5_exit_exec, "mospfspfcalc", "QospfRouterProcesses () [mospfspfcalc exit execs]")
			{
			}


		/** state (mospfspfcalc) transition processing **/
		FSM_TRANSIT_ONLY ((detupstrnode), 13, state13_enter_exec, ;, "mospfspfcalc")
		/*---------------------------------------------------------*/



		/** state (st_12) enter executives **/
		FSM_STATE_ENTER_UNFORCED (6, state6_enter_exec, "st_12", "QospfRouterProcesses () [st_12 enter execs]")
			{
			}


		/** blocking after enter executives of unforced state. **/
		FSM_EXIT (13,QospfRouterProcesses)


		/** state (st_12) exit executives **/
		FSM_STATE_EXIT_UNFORCED (6, state6_exit_exec, "st_12", "QospfRouterProcesses () [st_12 exit execs]")
			{
			}


		/** state (st_12) transition processing **/
		FSM_TRANSIT_MISSING ("st_12")
		/*---------------------------------------------------------*/



		/** state (ospfspfcalc) enter executives **/
		FSM_STATE_ENTER_FORCED (7, state7_enter_exec, "ospfspfcalc", "QospfRouterProcesses () [ospfspfcalc enter execs]")
			{
			/*int backup_rte_table[501];   commented by Jiemei*/
			ListL CandidateList;
			Vertex_Struct* CandidateEle;
			Vertex_Struct VertexV; 
			Vertex_Struct VertexW;  
			int i;
			int CandidateFoundFlag = OFF;
			int Inx;
			int *rte_table;  
			double ospfcalc_time = 0;
			struct timeval enter, leave;
			long diff;
			FILE* fp;
			
			VL	SPFDL
				printf("-------  Unicast Dijikstra's algorithm (%d.%d)   --------- \n", 
								RouterId.NetNum, RouterId.NodeNum);
			
			
			gettimeofday(&enter, (struct timezone*)0);
			
			rte_table = (int *)op_ima_obj_svar_get(ip_objid, "rte_tab" );
			/*op_prg_mem_copy( rte_table, backup_rte_table, sizeof(int)*501 ); commented by Jiemei, never used*/
			
			memset( &(AreaStructure.ShortestPathTree), 0, sizeof( ShortestPathTree_Struct ) );
			ListInit( &CandidateList );  
			
			AreaStructure.ShortestPathTree.SPVertex[0].VertexId = RouterId;
			AreaStructure.ShortestPathTree.SPVertex[0].DistanceToRoot = 0;
			
			VertexV.VertexId = RouterId;
			VertexV.DistanceToRoot = 0;
			memset( VertexV.NextHopsList, 0, sizeof(IPAddress)*MAX_NO_NEIGHBORS );
			memset( &VertexW, 0, sizeof(Vertex_Struct) );
			
			OspfCandidateAddPro( &VertexV, &CandidateList );
			OspfSPTreeBuild( &CandidateList );
			RoutingTableCalc( &AreaStructure, rte_table ); 
			
			VL SPFDL
			{
				printf("~~~~~~~~~~~~~~~  Shortest Path Information ~~~~~~~~~~~~~~~~~~~~\n");
				i = 0;
				while( !(AreaStructure.ShortestPathTree.SPVertex[i].VertexId.NetNum == 0 || 
					 AreaStructure.ShortestPathTree.SPVertex[i].VertexId.NodeNum == 0) )
				{
					printf("SP Router No %d is %d.%d and its distance to root is %d \n",
							i, AreaStructure.ShortestPathTree.SPVertex[i].VertexId.NetNum,
							AreaStructure.ShortestPathTree.SPVertex[i].VertexId.NodeNum,
							AreaStructure.ShortestPathTree.SPVertex[i].DistanceToRoot );
					printf("The Next Hops for this Router is %d.%d \n", 
							AreaStructure.ShortestPathTree.SPVertex[i].NextHopsList[0].NetNum,
							AreaStructure.ShortestPathTree.SPVertex[i].NextHopsList[0].NodeNum );
					i++;
				}
				printf("~~~~~~~~~~~~~  End Shortest Path Information ~~~~~~~~~~~~~~~~~~\n");
				printf("-------  End Unicast Dijikstra's algorithm (%d.%d)   --------- \n", 
								RouterId.NetNum, RouterId.NodeNum);
			}
			
			PROBE_ON
			{
				NoOfOspfCalcDone++;
				TotalDijCalcDone++;
				gettimeofday(&leave, (struct timezone*)0);
				if( enter.tv_usec > leave.tv_usec)
					diff = (((leave.tv_sec - 1) - enter.tv_sec)*1000000) + (leave.tv_usec + 1000000) - enter.tv_usec;
				else
					diff = ((leave.tv_sec - enter.tv_sec)*1000000) + leave.tv_usec - enter.tv_usec;
				OspfDijCalcTime = OspfDijCalcTime + diff;
				TotalDijCalcTime = TotalDijCalcTime + diff;
			
				op_stat_local_write( OSPF_DIJ_CALC_STAT, OspfDijCalcTime/NoOfOspfCalcDone );
				op_stat_local_write( TOTAL_DIJ_CALC_STAT, TotalDijCalcTime/TotalDijCalcDone );
				if(op_sim_time() > print_time + 50*Tracking_Interval)
				{
			/*		fp = fopen("../LogFiles/OspfDijCalcTime.log", "a" );
					if(!fp)
						op_sim_end("Unable to open the file OspfDijCalcTime.log....  ","","","");
					fprintf(fp, "Average time spent by Router %d.%d for OSPF Dijkstra's Calculation is %ld. "
								"Current Simulation time is %f \n", RouterId.NetNum, RouterId.NodeNum, 
								OspfDijCalcTime/NoOfOspfCalcDone, op_sim_time() );
					fclose(fp); */
					print_time = op_sim_time();
				}
			}
			ospfspfcalc = 0;
			detupstrnode = 1;
			
			}


		/** state (ospfspfcalc) exit executives **/
		FSM_STATE_EXIT_FORCED (7, state7_exit_exec, "ospfspfcalc", "QospfRouterProcesses () [ospfspfcalc exit execs]")
			{
			}


		/** state (ospfspfcalc) transition processing **/
		FSM_TRANSIT_ONLY ((detupstrnode), 13, state13_enter_exec, ;, "ospfspfcalc")
		/*---------------------------------------------------------*/



		/** state (BCQospfLsa) enter executives **/
		FSM_STATE_ENTER_FORCED (8, state8_enter_exec, "BCQospfLsa", "QospfRouterProcesses () [BCQospfLsa enter execs]")
			{
			int send_qospf_ls_update_pk = 0;
			
			/**
			*** Creating the RESOURCE LSA
			**/
				int i;
			/*	IPAddress Neighbors_Router_Id; */
				reslsa* temp_res_lsa;
			
				res_ls_seq_no++;
			
				temp_res_lsa = (reslsa *) op_prg_mem_alloc(sizeof (reslsa));
				memset(temp_res_lsa, 0, sizeof(reslsa));
			
				temp_res_lsa->ResLsaHeader.LsAge = 0;
				temp_res_lsa->ResLsaHeader.Options = 0;
				temp_res_lsa->ResLsaHeader.LSA_type = RESLSA;
				temp_res_lsa->ResLsaHeader.LinkStateId.NetNum = RouterId.NetNum;
				temp_res_lsa->ResLsaHeader.LinkStateId.NodeNum = RouterId.NodeNum;
				temp_res_lsa->ResLsaHeader.AdvRouter.NetNum = RouterId.NetNum;
				temp_res_lsa->ResLsaHeader.AdvRouter.NodeNum = RouterId.NodeNum;
				temp_res_lsa->ResLsaHeader.ThisAreaId = AreaId;
				temp_res_lsa->ResLsaHeader.LSSeqNo = res_ls_seq_no;
				temp_res_lsa->ResLsaHeader.LSChecksum = 0;
				temp_res_lsa->ResLsaHeader.Length = sizeof(reslsa);
				temp_res_lsa->ResLsaHeader.rtype = ROUTE_PINNING_IS_USED; 
				temp_res_lsa->ResLsaHeader.No_of_Links = NoOfLinks; 
				for(i = 0; i < temp_res_lsa->ResLsaHeader.No_of_Links; i++)
				{
					temp_res_lsa->ResLinkInfo[i].Type = LinksInfo[i].Types; 
			
			
			/* added by Jiemei, commented by DG 11/18
			Neighbors_Router_Id.NetNum=0;
			Neighbors_Router_Id.NodeNum=0;
			*/
			
			/*
			printf("i=%d Type=%d\n",i,temp_res_lsa->ResLinkInfo[i].Type);
			printf("Neighbors_Router_Id=%d.%d\n",Neighbors_Router_Id.NetNum,Neighbors_Router_Id.NodeNum);
			*/
					switch (temp_res_lsa->ResLinkInfo[i].Type) 
					{
						case 1  :	/* Router */
							  /* commented by Jiemei  
							  op_prg_mem_copy( &(temp_res_lsa->ResLinkInfo[i].LinkId), &Neighbors_Router_Id, 
				   					     sizeof(IPAddress) );
							  op_prg_mem_copy( &(temp_res_lsa->ResLinkInfo[i].LinkData), &(LinksInfo[i].IntfIPAddr), 
					 				     sizeof(IPAddress) );*/
			
							  /* added by Jiemei. */
							  /* Changed by DG 11/18, Follow the same info in OSPF Router LSA */
							  op_prg_mem_copy( &(LinksInfo[i].NeighborRtrId),
									   &(temp_res_lsa->ResLinkInfo[i].LinkId), sizeof(IPAddress) );
							  op_prg_mem_copy( &(LinksInfo[i].IntfIPAddr), 
									   &(temp_res_lsa->ResLinkInfo[i].LinkData), sizeof(IPAddress) );
							  break;
						case 2  : 	/* TransitNet */
						       	  /* commented by Jiemei 
							  op_prg_mem_copy( &(temp_res_lsa->ResLinkInfo[i].LinkData), &(LinksInfo[i].IntfIPAddr), 
									     sizeof(IPAddress) );*/
			
							  /* added by Jiemei */
							  op_prg_mem_copy( &(LinksInfo[i].IntfIPAddr), &(temp_res_lsa->ResLinkInfo[i].LinkData), 
									     sizeof(IPAddress) );
							  op_prg_mem_copy( &RouterId, &(temp_res_lsa->ResLinkInfo[i].LinkData), 
									     sizeof(IPAddress) );
							  break;
						case 3  :	/* SubNet */
			/*				  temp_res_lsa->ResLinkInfo[i].LinkId.NetNum = LinksInfo[i].IntfIPAddr.NetNum; */
			/* Changed by DG, 11/8, follow the info on OSPF Router LSA  */
							  temp_res_lsa->ResLinkInfo[i].LinkId.NetNum = LinksInfo[i].NeighborRtrId.NetNum; 
							  temp_res_lsa->ResLinkInfo[i].LinkData.NetNum = NetworkMask; 
							  break;
						default : break;
					}
					temp_res_lsa->ResLinkInfo[i].TOSMetric = LinksInfo[i].CostMetric;
					temp_res_lsa->ResLinkInfo[i].LinkDelay = LinksInfo[i].Delay;
					temp_res_lsa->ResLinkInfo[i].Bandwidth = LinksInfo[i].Bandwidth;
					temp_res_lsa->ResLinkInfo[i].BufferSize = LinksInfo[i].BufferSize;
				}
			
				op_prg_mem_copy( temp_res_lsa, &(AreaStructure.ResourceLsaList[0]), sizeof(reslsa) );
				op_prg_mem_free( temp_res_lsa );
			
				send_qospf_ls_update_pk = TRUE;
			/**
			*** Creating the RRA LSA
			**/
			{
				int i;
			/*	IPAddress Neighbors_Router_Id; */
				rralsa* temp_rra_lsa;
				rra_ls_seq_no++;
			
				temp_rra_lsa = (rralsa *) op_prg_mem_alloc (sizeof (rralsa));
				memset(temp_rra_lsa, 0, sizeof(rralsa));
			
				temp_rra_lsa->OpaqueLsaHeader.LsAge = 0;
				temp_rra_lsa->OpaqueLsaHeader.Options = 0;
				temp_rra_lsa->OpaqueLsaHeader.LSA_type = RRALSA;
				temp_rra_lsa->OpaqueLsaHeader.OpaqueType = 11;
				temp_rra_lsa->OpaqueLsaHeader.OpaqueId = 5;     
				temp_rra_lsa->OpaqueLsaHeader.AdvRouter.NetNum = RouterId.NetNum;
				temp_rra_lsa->OpaqueLsaHeader.AdvRouter.NodeNum = RouterId.NodeNum;
				temp_rra_lsa->OpaqueLsaHeader.ThisAreaId = AreaId;
				temp_rra_lsa->OpaqueLsaHeader.LSSeqNo = rra_ls_seq_no;
				temp_rra_lsa->OpaqueLsaHeader.LSChecksum = 0;
				temp_rra_lsa->OpaqueLsaHeader.Length = sizeof(rralsa);
				#ifdef ER_IS_USED
					temp_rra_lsa->OpaqueLsaHeader.FloddingScope = 1;
				#endif;
				#ifndef ER_IS_USED
					temp_rra_lsa->OpaqueLsaHeader.FloddingScope = 0;
				#endif;
			/*
				temp_rra_lsa->DestAddr = QospfPathStateBlock->SessionObject.DestAddr; 
				temp_rra_lsa->SourceAddr = QospfPathStateBlock->SenderTemplateObject.SrcAddr;    
			*/
			/* Added by DG, 11/18 */
				temp_rra_lsa->DestAddr = SessionDirectory[CurrSess].DestAddr; 
				temp_rra_lsa->SourceAddr = SessionDirectory[CurrSess].SenderAddr;    
			
				temp_rra_lsa->src_prefix_len = 0;
				temp_rra_lsa->dst_prefix_len = 0;
				temp_rra_lsa->No_of_Links = NoOfLinks;
				for(i = 0; i < NoOfLinks; i++)
				{
					temp_rra_lsa->LinkResvInfo[i].Type = LinksInfo[i].Types;
					switch(temp_rra_lsa->LinkResvInfo[i].Type) 
					{
						case 1  :
							/* commented by Jiemei  
							op_prg_mem_copy( &(temp_rra_lsa->LinkResvInfo[i].LinkId), &Neighbors_Router_Id, 
									sizeof(IPAddress) );
							op_prg_mem_copy( &(temp_rra_lsa->LinkResvInfo[i].LinkData), 
									&(LinksInfo[i].IntfIPAddr), sizeof(IPAddress) );*/
			
							/* added by Jiemei */
							/* Changed by DG 11/18, Follow the same info in OSPF Router LSA */
							op_prg_mem_copy( &(LinksInfo[i].NeighborRtrId),
									&(temp_rra_lsa->LinkResvInfo[i].LinkId), sizeof(IPAddress) );
							op_prg_mem_copy( &(LinksInfo[i].IntfIPAddr), 
									&(temp_rra_lsa->LinkResvInfo[i].LinkData), sizeof(IPAddress) );
							break;
						case 2  : 
							/* commented by Jiemei 
							op_prg_mem_copy( &(temp_rra_lsa->LinkResvInfo[i].LinkData), 
									&(LinksInfo[i].IntfIPAddr), sizeof(IPAddress) );*/
			
							/* added by Jiemei */
							op_prg_mem_copy( &(LinksInfo[i].IntfIPAddr),
									&(temp_rra_lsa->LinkResvInfo[i].LinkData), sizeof(IPAddress) );
							op_prg_mem_copy( &RouterId, &(temp_rra_lsa->LinkResvInfo[i].LinkData),  
									sizeof(IPAddress) );
							break;
						case 3  : 
			/*				temp_rra_lsa->LinkResvInfo[i].LinkId.NetNum = LinksInfo[i].IntfIPAddr.NetNum; */
			/* Changed by DG, 11/8, follow the info on OSPF Router LSA  */
							temp_rra_lsa->LinkResvInfo[i].LinkId.NetNum = LinksInfo[i].NeighborRtrId.NetNum; 
							temp_rra_lsa->LinkResvInfo[i].LinkData.NetNum = NetworkMask;
							break;
						default : break;
					}
					temp_rra_lsa->LinkResvInfo[i].PinFlag = 1;
					temp_rra_lsa->LinkResvInfo[i].ReservedBandwidth = LinksInfo[i].ReservedBandwidth;
					temp_rra_lsa->LinkResvInfo[i].ReservedBufferSize = LinksInfo[i].ReservedBufferSize;
				}
			
				op_prg_mem_copy( temp_rra_lsa, &(AreaStructure.RRALsaList[0]), sizeof(rralsa) );
				op_prg_mem_free( temp_rra_lsa );  
				send_qospf_ls_update_pk = TRUE;
			}
			
			
			
			/**
			*** Sending the Link state Update packets
			**/
			
			
			if(send_qospf_ls_update_pk == TRUE)
			{
				QospfLSUpdatePk* temp_ls_update_pk;
				Packet* QospfLsUpdatePkptr;
				Packet* ipdgramptr;
				int Prev_sended_ResLsaSeqNo, Prev_sended_RRALsaSeqNo;
				int LSUpdatePkCheckSum = 0;
				int i;
			
			TRACE VL QPDL printf("----------------(%d.%d) BCQospfLsa Time=%fs------------------\n",
			        	RouterId.NetNum,RouterId.NodeNum,op_sim_time());
			
				temp_ls_update_pk = ( QospfLSUpdatePk *) op_prg_mem_alloc( sizeof(QospfLSUpdatePk) );
				memset( temp_ls_update_pk, 0, sizeof(QospfLSUpdatePk) );
			
				temp_ls_update_pk->OspfPkHeader.Version = VERSION;
				temp_ls_update_pk->OspfPkHeader.Type = QOSPFLSUPDATEPK;
				temp_ls_update_pk->OspfPkHeader.PacketLength = sizeof(QospfLSUpdatePk);
				op_prg_mem_copy( &(RouterId), &(temp_ls_update_pk->OspfPkHeader.ThisRouterId), 
						sizeof(IPAddress) );
				temp_ls_update_pk->OspfPkHeader.ThisAreaId = AreaId;
				temp_ls_update_pk->OspfPkHeader.CheckSum = LSUpdatePkCheckSum;
				temp_ls_update_pk->OspfPkHeader.AuType = NONE;
				temp_ls_update_pk->OspfPkHeader.Authentication = 0;
			
			TRACE VL QPDL
			{
				printf( "Resource LSA:\n" );
				printf( "    0--AdvRouter=%d.%d \n",
			                AreaStructure.ResourceLsaList[0].ResLsaHeader.AdvRouter.NetNum,
			                AreaStructure.ResourceLsaList[0].ResLsaHeader.AdvRouter.NodeNum );
			}
				op_prg_mem_copy( &(AreaStructure.ResourceLsaList[0]), &(temp_ls_update_pk->ResLinkStateAdv[0]), 
						   sizeof(reslsa) );
				temp_ls_update_pk->NoOfResAdvs++;
				NoOfQospfLsaSent++;
			
			/* Added by DG and Jiemei, 11/27 */
				NoOfInstQospfLsaSent++;
			
				for(i = 1; i <MAX_NO_LSAS; i++ ) /* found by Sitaram */
				{
					if( temp_ls_update_pk->NoOfResAdvs >= (MAX_NO_LSAS - 1) ||
						AreaStructure.ResourceLsaList[i].ResLsaHeader.LSA_type != RESLSA ||
						( AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NetNum == 0 &&
						  AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NodeNum == 0 ) )
					{
						continue;
					}
					else
					{
			TRACE VL QPDL
				printf( "    %d--AdvRouter=%d.%d \n", i,
			                AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NetNum,
			                AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NodeNum );
						op_prg_mem_copy( &(AreaStructure.ResourceLsaList[i]), 
								 &(temp_ls_update_pk->ResLinkStateAdv[i]), 
								 sizeof(reslsa) );
						temp_ls_update_pk->NoOfResAdvs++;
						NoOfQospfLsaSent++;
			
			/* Added by DG and Jiemei, 11/27 */
						NoOfInstQospfLsaSent++;
			
					}
			
				}
			
			TRACE VL QPDL
			{
				printf( "Resv LSA:\n" );
				printf( "    0--AdvRouter=%d.%d Dest(%d.%d) Source(%d.%d)\n",
			                AreaStructure.RRALsaList[0].OpaqueLsaHeader.AdvRouter.NetNum,
			                AreaStructure.RRALsaList[0].OpaqueLsaHeader.AdvRouter.NodeNum,
					AreaStructure.RRALsaList[0].DestAddr.NetNum,
					AreaStructure.RRALsaList[0].DestAddr.NodeNum,
					AreaStructure.RRALsaList[0].SourceAddr.NetNum,
					AreaStructure.RRALsaList[0].SourceAddr.NodeNum );
			}
				op_prg_mem_copy( &(AreaStructure.RRALsaList[0]), &(temp_ls_update_pk->RraLinkStateAdv[0]),
								   sizeof(rralsa) );
				temp_ls_update_pk->NoOfRraAdvs++;
				NoOfQospfLsaSent++;
			
			/* Added by DG and Jiemei, 11/27 */
				NoOfInstQospfLsaSent++;
			
				for(i = 1; i <MAX_NO_LSAS; i++ ) /* found by Sitaram */
				{
					if( temp_ls_update_pk->NoOfRraAdvs >= (MAX_NO_LSAS - 1) ||
						AreaStructure.RRALsaList[i].OpaqueLsaHeader.LSA_type != RRALSA ||
						( AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NetNum == 0 &&
						  AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NodeNum == 0 ) )
					{
						continue;
					}
					else
					{
			TRACE VL QPDL
				printf( "    %d--AdvRouter=%d.%d Dest(%d.%d) Source(%d.%d)\n",i,
			                AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NetNum,
			                AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NodeNum,
					AreaStructure.RRALsaList[i].DestAddr.NetNum,
					AreaStructure.RRALsaList[i].DestAddr.NodeNum,
					AreaStructure.RRALsaList[i].SourceAddr.NetNum,
					AreaStructure.RRALsaList[i].SourceAddr.NodeNum );
						op_prg_mem_copy( &(AreaStructure.RRALsaList[i]), 
								 &(temp_ls_update_pk->RraLinkStateAdv[i]), 
								 sizeof(rralsa) );
						temp_ls_update_pk->NoOfRraAdvs++;
						NoOfQospfLsaSent++;
			/* Added by DG and Jiemei, 11/27 */
						NoOfInstQospfLsaSent++;
					}
			
				}
			TRACE VL QPDL printf("----------------(%d.%d) BCQospfLsa Information End--------------------\n",
			        RouterId.NetNum,RouterId.NodeNum);
			
				QospfLsUpdatePkptr = op_pk_create_fmt("qospflsupdate_pkfmt");
				op25_pk_nfd_set( QospfLsUpdatePkptr, "Lsupdate", temp_ls_update_pk, op_prg_mem_alloc,
							   op_prg_mem_free, sizeof(QospfLSUpdatePk) );
				op25_pk_nfd_set( QospfLsUpdatePkptr, "Msg_type", QOSPFLSUPDATEPK );
			
				data_len = op_pk_total_size_get( QospfLsUpdatePkptr );
				ipdgramptr = op_pk_create_fmt("ip_dgram");
				op25_pk_nfd_set( ipdgramptr, "data", QospfLsUpdatePkptr );
				op25_pk_nfd_set( ipdgramptr, "prot_id", QOSPF_PROTOCOL_ID );
				op_pk_bulk_size_set( ipdgramptr, data_len*8 );
			
				op25_pk_nfd_set( ipdgramptr, "dest_net", AllSPFRoutersNetNum );
				op25_pk_nfd_set( ipdgramptr, "dest_node", AllSPFRoutersNodeNum );
				op25_pk_nfd_set( ipdgramptr, "ttl", 2 );
			
				op25_pk_nfd_set( ipdgramptr, "src_net", RouterId.NetNum );
				op25_pk_nfd_set( ipdgramptr, "src_node", RouterId.NodeNum );
			
				op25_pk_nfd_set( ipdgramptr, "orig_len", data_len );
				op25_pk_nfd_set( ipdgramptr, "frag_len", data_len );
			
				op25_pk_nfd_set( ipdgramptr, "frag", 0 );
			
				op_pk_send( ipdgramptr, IPSTREAM );
			
				PROBE_ON
				{
					op_stat_local_write(CUMM_QOSPF_LSA_SENT_STAT, NoOfQospfLsaSent);
			/* Added by DG and Jiemei, 11/27/97 */
					op_stat_local_write(INST_QOSPF_LSA_SENT_STAT, NoOfInstQospfLsaSent);
				}
			}
			
			/* Added by DG, 11/18, Don't reschedule self if code == 1. In that case, the process itself was invoked 
			 by receiving RSVP MSG. */
			
			if ( op_intrpt_code() == 1 )
			{
				if(op_sim_time() < MAX_INIT_TIME)
				{
					op_intrpt_schedule_self(op_sim_time()+1, 1);
					BCQospfLsa_sim_time = op_sim_time();
				}
				else
				{
					if(op_sim_time() < BCQospfLsa_sim_time + MAX_REFRESH_TIME)
					{
				                if ( QSentLsaCycleFirst == 0 )
			        	        {
			                	        QSentLsaCycleFirst = 1;
			                        	NoOfInstQospfLsaSent = 0;
				                }
			
						op_intrpt_schedule_self(op_sim_time()+1, 1);
					}
					else
					{
						op_intrpt_schedule_self(op_sim_time()+QospfLsaRefreshTime - MAX_REFRESH_TIME, 1);
						BCQospfLsa_sim_time = op_sim_time()+QospfLsaRefreshTime - MAX_REFRESH_TIME;
			                        QSentLsaCycleFirst = 0;
					}
				}
			}
			}


		/** state (BCQospfLsa) exit executives **/
		FSM_STATE_EXIT_FORCED (8, state8_exit_exec, "BCQospfLsa", "QospfRouterProcesses () [BCQospfLsa exit execs]")
			{
			}


		/** state (BCQospfLsa) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (BCOspfLsa) enter executives **/
		FSM_STATE_ENTER_FORCED (9, state9_enter_exec, "BCOspfLsa", "QospfRouterProcesses () [BCOspfLsa enter execs]")
			{
			
			/**
			*** Creating the OSPF LSA's
			**/
			int i,j;
			int Prev_sended_RouterLsaSeqNo;
			int Prev_sended_NetworkLsaSeqNo;
			int Prev_sended_SummaryLsaSeqNo;
			
			/** 
			*** Create the Router Link State Advertisement for the Aera 
			**/
			{
				RouterLinkAdv* tmp_router_lsa;
			
				router_ls_seq_no++;
				tmp_router_lsa = (RouterLinkAdv *) op_prg_mem_alloc (sizeof (RouterLinkAdv));
				memset(tmp_router_lsa, 0, sizeof(RouterLinkAdv));
			
				tmp_router_lsa->LinkStateHeader.LsAge = 0;
				tmp_router_lsa->LinkStateHeader.LsType = ROUTERLSA;
				op_prg_mem_copy( &(RouterId), &(tmp_router_lsa->LinkStateHeader.LinkStateId), 
								   sizeof(IPAddress) );
				op_prg_mem_copy( &(RouterId), &(tmp_router_lsa->LinkStateHeader.AdvRouter), 
								   sizeof(IPAddress) );
				tmp_router_lsa->LinkStateHeader.ThisAreaId = AreaId;
				tmp_router_lsa->LinkStateHeader.LSSeqNo = router_ls_seq_no;
				tmp_router_lsa->LinkStateHeader.LSChecksum = 0;
				tmp_router_lsa->LinkStateHeader.Length = sizeof(RouterLinkAdv);
				tmp_router_lsa->bit_b = 1;
				tmp_router_lsa->bit_e = 0;
				tmp_router_lsa->No_of_Links = NoOfLinks;
				for(i = 0; i < tmp_router_lsa->No_of_Links; i++)
				{
					tmp_router_lsa->Link_Info[i].Type = LinksInfo[i].Types; 
					switch(tmp_router_lsa->Link_Info[i].Type)
					{
			/* ROUTER, by DG */
						case  1 : op_prg_mem_copy( &(LinksInfo[i].NeighborRtrId), 
									   &(tmp_router_lsa->Link_Info[i].LinkId), sizeof(IPAddress) );
							  op_prg_mem_copy( &(LinksInfo[i].IntfIPAddr),
									   &(tmp_router_lsa->Link_Info[i].LinkData), sizeof(IPAddress) );
							  break;
			/*TRANSITNET, by DG */
						case  2 : op_prg_mem_copy( &(LinksInfo[i].IntfIPAddr),
									   &(tmp_router_lsa->Link_Info[i].LinkId), sizeof(IPAddress) );
							  op_prg_mem_copy( &(RouterId), &(tmp_router_lsa->Link_Info[i].LinkData), 
									     sizeof(IPAddress) );
							  break;
			/* SUBNET, by DG */
						case  3 : tmp_router_lsa->Link_Info[i].LinkId.NetNum = LinksInfo[i].NeighborRtrId.NetNum;
							  tmp_router_lsa->Link_Info[i].LinkData.NetNum = NetworkMask;
							  break;
						default : break;
					}
					tmp_router_lsa->Link_Info[i].NoOfTOS = 0;
					tmp_router_lsa->Link_Info[i].TOSMetric = LinksInfo[i].CostMetric;
				}
				VL	OPDL
				{
					printf("-------------------  BC OSPF ROUTER LSA STATE  ------------------\n");
					printf("Advertising Router is %d.%d and its Link information is : \n", 
							RouterId.NetNum, RouterId.NodeNum );
					for( i = 0; i < NoOfLinks; i++ )
					{
						printf("Link Type is %d and its cost is %d \n", 
								tmp_router_lsa->Link_Info[i].Type, 
								tmp_router_lsa->Link_Info[i].TOSMetric );
						printf("Link_Info[%d].LinkId = %d.%d, Link_Info[%d].LinkData = %d.%d\n",
								i,tmp_router_lsa->Link_Info[i].LinkId.NetNum, 
								tmp_router_lsa->Link_Info[i].LinkId.NodeNum,
								i,tmp_router_lsa->Link_Info[i].LinkData.NetNum,
								tmp_router_lsa->Link_Info[i].LinkData.NodeNum );
						/*  PrintLsa( (char*)tmp_router_lsa, ROUTERLSA );  */
					}
					printf("-----------------  END BC OSPF ROUTER LSA STATE  ----------------\n");
				}
				op_prg_mem_copy( tmp_router_lsa, &AreaStructure.RouterLsaList[0], sizeof(RouterLinkAdv) );
				op_prg_mem_free( tmp_router_lsa );
			}
			/**
			*** Create the Network Link State Advertisement 
			**/
			if( IS_DESIGNATED_ROUTER == 1)
			{
				NetworkLinkAdv* tmp_network_lsa;
			
				tmp_network_lsa = (NetworkLinkAdv*) op_prg_mem_alloc(sizeof(NetworkLinkAdv));
				memset(tmp_network_lsa, 0, sizeof(NetworkLinkAdv));
			
				network_ls_seq_no++;
				tmp_network_lsa->LinkStateHeader.LsAge = 0;
				tmp_network_lsa->LinkStateHeader.LsType = NETWORKLSA; 
				op_prg_mem_copy( &(RouterId), &(tmp_network_lsa->LinkStateHeader.LinkStateId), sizeof(IPAddress) ); 
				op_prg_mem_copy( &(RouterId), &(tmp_network_lsa->LinkStateHeader.AdvRouter), sizeof(IPAddress) ); 
				tmp_network_lsa->LinkStateHeader.ThisAreaId = AreaId;
				tmp_network_lsa->LinkStateHeader.LSSeqNo = network_ls_seq_no; 
				tmp_network_lsa->LinkStateHeader.LSChecksum = 0; 
				tmp_network_lsa->LinkStateHeader.Length = sizeof(NetworkLinkAdv);; 
				tmp_network_lsa->ThisNetworkMask = NetworkMask;
				op_prg_mem_copy( &(RouterId), &(tmp_network_lsa->AttachedRouter[0]), sizeof(IPAddress) );
				j=1;
				for(i=0; i<No_of_Networks; i++)
				{
					op_prg_mem_copy( &(Network_Routers[i]),
							&(tmp_network_lsa->AttachedRouter[j]), sizeof(IPAddress) );
					j++;
				}
			
				VL	OPDL
				{
					printf("-------------------  BC OSPF NETWORK LSA STATE  ------------------\n");
					printf("Advertising Router is %d.%d and its Link information is : \n", 
							RouterId.NetNum, RouterId.NodeNum );
					for( i = 0; i <=No_of_Networks; i++ )    /* 0- self, 1~No_of_networks - neighbor.  Jiemei*/
					{
						printf("AttachedRouter[%d] = %d.%d\n",i,
							tmp_network_lsa->AttachedRouter[i].NetNum, 
							tmp_network_lsa->AttachedRouter[i].NodeNum);
					}
					printf("-----------------  END BC OSPF NETWORK LSA STATE  ----------------\n");
				}
				op_prg_mem_copy( tmp_network_lsa, &AreaStructure.NetworkLsaList[0], sizeof(NetworkLinkAdv) );
			
				op_prg_mem_free(tmp_network_lsa);
			}
			
			/**
			*** IF the router is an Area Border Router Create the Summary Link State Adv
			**/
			if( IS_AREA_BORDER_ROUTER == 1)
			{ 
				SummaryLinkAdv* tmp_summary_lsa;
			
				tmp_summary_lsa = (SummaryLinkAdv*) op_prg_mem_alloc( sizeof(SummaryLinkAdv) );
				memset(tmp_summary_lsa, 0, sizeof(SummaryLinkAdv));
			
				summary_ls_seq_no++;
				tmp_summary_lsa->LinkStateHeader.LsAge = 0;
				tmp_summary_lsa->LinkStateHeader.LsType = SUMMARYLSA; 
				tmp_summary_lsa->LinkStateHeader.LinkStateId.NetNum = Ip_Network_No;
				op_prg_mem_copy( &(RouterId), &(tmp_summary_lsa->LinkStateHeader.AdvRouter), sizeof(IPAddress) );
				tmp_summary_lsa->LinkStateHeader.ThisAreaId = AreaId;
				tmp_summary_lsa->LinkStateHeader.LSSeqNo = summary_ls_seq_no;
				tmp_summary_lsa->LinkStateHeader.LSChecksum = 0;
				tmp_summary_lsa->LinkStateHeader.Length = sizeof(SummaryLinkAdv);
				tmp_summary_lsa->ThisNetworkMask = NetworkMask;
				for (i=0; i<No_of_TOS; i++)
				{
					tmp_summary_lsa->Tos_Info[i].TOS = 0;
					tmp_summary_lsa->Tos_Info[i].metric = TOS_Metric[tmp_summary_lsa->Tos_Info[i].TOS];
				}
			
				VL	OPDL
				{
					printf("-------------------  BC OSPF SUMMARY LSA STATE  ------------------\n");
					printf("Advertising Router is %d.%d and its Link information is : \n", 
							RouterId.NetNum, RouterId.NodeNum );
					for( i = 0; i < No_of_TOS; i++ )
					{
						printf("Tos_Info[%d].TOS = %d  Tos_Info[%d].metric=%d\n",
							i,tmp_summary_lsa->Tos_Info[i].TOS, 
							i,tmp_summary_lsa->Tos_Info[i].metric);
					}
					printf("-----------------  END BC OSPF SUMMARY LSA STATE  ----------------\n");
				}
				op_prg_mem_copy( tmp_summary_lsa, &AreaStructure.SummaryLsaList[0], sizeof(SummaryLinkAdv) );
				op_prg_mem_free( tmp_summary_lsa );
			}
			
				/**
				*** Send the Ospf Lsa Update packet 
				**/
			{
				Packet* OspfLsUpdatePkptr;
				Packet* ipdgramptr;
				OspfLSUpdatePk* temp_ls_update_pk;
				int LSUpdatePkCheckSum = 0;
			
				temp_ls_update_pk = ( OspfLSUpdatePk *) op_prg_mem_alloc( sizeof(OspfLSUpdatePk) );
				memset( temp_ls_update_pk, 0, sizeof(OspfLSUpdatePk) );
			
				temp_ls_update_pk->OspfPkHeader.Version = VERSION;
				temp_ls_update_pk->OspfPkHeader.Type = OSPFLSUPDATEPK;
				temp_ls_update_pk->OspfPkHeader.PacketLength = sizeof(OspfLSUpdatePk);
				op_prg_mem_copy( &(RouterId), &(temp_ls_update_pk->OspfPkHeader.ThisRouterId), 
								sizeof(IPAddress) );
				temp_ls_update_pk->OspfPkHeader.ThisAreaId = AreaId;
				temp_ls_update_pk->OspfPkHeader.CheckSum = LSUpdatePkCheckSum;
				temp_ls_update_pk->OspfPkHeader.AuType = NONE;
				temp_ls_update_pk->OspfPkHeader.Authentication = 0;
			
				op_prg_mem_copy( &AreaStructure.RouterLsaList[0], &(temp_ls_update_pk->RouterLinkStateAdv[0]),
								 sizeof(RouterLinkAdv) );
				temp_ls_update_pk->NoOfRouterAdvs++;
				NoOfOspfLsaSent++;
			
			/* Added by DG and Jiemei, 11/27/97 */
				NoOfInstOspfLsaSent++;
			
				for(i = 1; i <MAX_NO_LSAS; i++ ) /* found by Sitaram */
				{
					if( AreaStructure.RouterLsaList[i].LinkStateHeader.LsType != ROUTERLSA || 
					  ( AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 &&
					    AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NodeNum == 0 ) ) 
					{
					}
					else 
					{
						op_prg_mem_copy(&AreaStructure.RouterLsaList[i], 
								&(temp_ls_update_pk->RouterLinkStateAdv[i]),
								sizeof(RouterLinkAdv) );
						temp_ls_update_pk->NoOfRouterAdvs++;
						NoOfOspfLsaSent++;
			/* Added by DG and Jiemei, 11/27/97 */
						NoOfInstOspfLsaSent++;
					}
					
				}
			
			
				if(IS_DESIGNATED_ROUTER == 1)
				{
					op_prg_mem_copy( &AreaStructure.NetworkLsaList[0], &(temp_ls_update_pk->NetworkLinkStateAdv[0]), 
									 sizeof(NetworkLinkAdv) );
					temp_ls_update_pk->NoOfNetworkAdvs++;
					NoOfOspfLsaSent++;
			
			/* Added by DG and Jiemei, 11/27/97 */
					NoOfInstOspfLsaSent++;
			
					for(i = 1; i <MAX_NO_LSAS; i++ ) /* found by Sitaram */
					{
						if( AreaStructure.NetworkLsaList[i].LinkStateHeader.LsType != NETWORKLSA || 
						  ( AreaStructure.NetworkLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 &&
							AreaStructure.NetworkLsaList[i].LinkStateHeader.AdvRouter.NodeNum == 0 ) ) 
						{
						}
						else 
						{
							op_prg_mem_copy( &AreaStructure.NetworkLsaList[i],
									 &(temp_ls_update_pk->NetworkLinkStateAdv[i]), 
									 sizeof(NetworkLinkAdv) );
							temp_ls_update_pk->NoOfNetworkAdvs++;
							NoOfOspfLsaSent++;
			/* Added by DG and Jiemei, 11/27/97 */
							NoOfInstOspfLsaSent++;
						}
						
					}
				}
			
				if(IS_AREA_BORDER_ROUTER == 1)
				{
					op_prg_mem_copy( &AreaStructure.SummaryLsaList[0], &(temp_ls_update_pk->SummaryLinkStateAdv[0]), 
									   sizeof(SummaryLinkAdv) );
					temp_ls_update_pk->NoOfSummaryAdvs++;
					NoOfOspfLsaSent++;
			/*  Added by DG and Jiemei, 11/27/97 */
					NoOfInstOspfLsaSent++;
			
					for(i = 1; i <MAX_NO_LSAS; i++ ) /* found by Sitaram */
					{
						if( AreaStructure.SummaryLsaList[i].LinkStateHeader.LsType != SUMMARYLSA || 
						  ( AreaStructure.SummaryLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 &&
						    AreaStructure.SummaryLsaList[i].LinkStateHeader.AdvRouter.NodeNum == 0 ) ) 
						{
						}
						else 
						{
							op_prg_mem_copy( &AreaStructure.SummaryLsaList[i], 
									 &(temp_ls_update_pk->SummaryLinkStateAdv[i]), 
									 sizeof(SummaryLinkAdv) );
						   temp_ls_update_pk->NoOfSummaryAdvs++;
						   NoOfOspfLsaSent++;
			/* Added by DG and Jiemei, 11/27/97 */
						   NoOfInstOspfLsaSent++;
						}
					}
				}
			
				OspfLsUpdatePkptr = op_pk_create_fmt("ospflsupdate_pkfmt");
				op25_pk_nfd_set( OspfLsUpdatePkptr, "Lsupdate", temp_ls_update_pk, op_prg_mem_alloc,
							   op_prg_mem_free, sizeof(OspfLSUpdatePk) );
				op25_pk_nfd_set( OspfLsUpdatePkptr, "Msg_type", OSPFLSUPDATEPK );
			
				data_len = op_pk_total_size_get( OspfLsUpdatePkptr );
				ipdgramptr = op_pk_create_fmt("ip_dgram");
				op25_pk_nfd_set( ipdgramptr, "data", OspfLsUpdatePkptr );
				op25_pk_nfd_set( ipdgramptr, "prot_id", QOSPF_PROTOCOL_ID );
				op_pk_bulk_size_set( ipdgramptr, data_len*8 );
			
				op25_pk_nfd_set( ipdgramptr, "dest_net", AllSPFRoutersNetNum );
				op25_pk_nfd_set( ipdgramptr, "dest_node", AllSPFRoutersNodeNum );
				op25_pk_nfd_set( ipdgramptr, "ttl", 2 );
			
				op25_pk_nfd_set( ipdgramptr, "src_net", RouterId.NetNum );
				op25_pk_nfd_set( ipdgramptr, "src_node", RouterId.NodeNum );
			
				op25_pk_nfd_set( ipdgramptr, "orig_len", data_len );
				op25_pk_nfd_set( ipdgramptr, "frag_len", data_len );
			
				op25_pk_nfd_set( ipdgramptr, "frag", 0 );
			
				op_pk_send( ipdgramptr, IPSTREAM );
			
				if( op_sim_time() < MAX_INIT_TIME )
				{
					op_intrpt_schedule_self(op_sim_time()+1, 0); 
					BCOspfLsa_sim_time = op_sim_time();   /* changed by Jiemei: BCOspfLsa should be SV*/
				}
			/*	else
				{
					if( op_sim_time() < BCOspfLsa_sim_time + MAX_REFRESH_TIME )
					{
				                if ( USentLsaCycleFirst == 0 )
			        	        {
			                	        USentLsaCycleFirst = 1;
			                        	NoOfInstOspfLsaSent = 0;
				                }
						op_intrpt_schedule_self(op_sim_time()+1, 0);
					}
					else
					{
						op_intrpt_schedule_self(op_sim_time()+OspfLsaRefreshTime-MAX_REFRESH_TIME, 0);
						BCOspfLsa_sim_time = op_sim_time()+QospfLsaRefreshTime - MAX_REFRESH_TIME;  
			                	USentLsaCycleFirst = 0;
					}
				}*/
			
				PROBE_ON
				{
					op_stat_local_write(CUMM_OSPF_LSA_SENT_STAT, NoOfOspfLsaSent);
					op_stat_local_write(INST_OSPF_LSA_SENT_STAT, NoOfInstOspfLsaSent);
				}
			}
			}


		/** state (BCOspfLsa) exit executives **/
		FSM_STATE_EXIT_FORCED (9, state9_exit_exec, "BCOspfLsa", "QospfRouterProcesses () [BCOspfLsa exit execs]")
			{
			}


		/** state (BCOspfLsa) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (BCMospfLsa) enter executives **/
		FSM_STATE_ENTER_FORCED (10, state10_enter_exec, "BCMospfLsa", "QospfRouterProcesses () [BCMospfLsa enter execs]")
			{
			/**
			*** Creating the GRPMBRSHIP LSA for a all the groups and send the LSA as
			*** a mospf ls update packet.
			**/
			int send_mospf_ls_update_pk = FALSE;
			{
				int i,j,k;
				int DestGrp;
				int VertexNo;
				int RegGrp_Flag;
				grpmbrlsa* tmp_grpmbrship_lsa;
				int No_of_Reg_Grps = 0;
				int RegGrps[MAX_NO_GRPS];
				int Reg_Grp_Flag;
			
			if ( !IS_ROUTER )  /* add 10/30 */
			{
				mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");
			
				VL MPDL  printf("Router %d.%d \n", RouterId.NetNum, RouterId.NodeNum );
				for( i = 0; i < MAX_NO_GRPS; i++ )
				{
					for( j = 0; j < NoOfLinks; j++ ) /* changed by Jiemei for j<= NoOfLinks */
					{
						Reg_Grp_Flag = OFF;
						if( mc_table[i].interface[j] == 1)
						{
							VL MPDL
							{
								printf(" GRP ID : %d, REGISTERED INTERFACE : %d \n", i, j ); 
							}
							RegGrps[No_of_Reg_Grps] = i;
							Reg_Grp_Flag = ON;
						}
						if (Reg_Grp_Flag == ON )
							No_of_Reg_Grps++;
					}
				}
			
				VL MPDL printf("This Router %d.%d is registered to %d No. of groups\n",
								RouterId.NetNum, RouterId.NodeNum, No_of_Reg_Grps );
			
			
				for( i = 0; i < No_of_Reg_Grps; i++ ) 
				{
					RegGrp_Flag = OFF;
					tmp_grpmbrship_lsa = (grpmbrlsa *) op_prg_mem_alloc(sizeof(grpmbrlsa));
					memset(tmp_grpmbrship_lsa, 0, sizeof(grpmbrlsa));
						 
					grpmbr_ls_seq_no++;
					tmp_grpmbrship_lsa->LinkStateHeader.LsAge = 0;
					tmp_grpmbrship_lsa->LinkStateHeader.LsType = GRPMBRSHIPLSA;
					tmp_grpmbrship_lsa->LinkStateHeader.AdvRouter.NetNum = RouterId.NetNum;
					tmp_grpmbrship_lsa->LinkStateHeader.AdvRouter.NodeNum = RouterId.NodeNum;
					tmp_grpmbrship_lsa->LinkStateHeader.ThisAreaId = AreaId;
					tmp_grpmbrship_lsa->LinkStateHeader.LSSeqNo = grpmbr_ls_seq_no;
					tmp_grpmbrship_lsa->LinkStateHeader.LSChecksum = 0;
					tmp_grpmbrship_lsa->LinkStateHeader.Length = sizeof(grpmbrlsa);
					VertexNo = 0;
					DestGrp = RegGrps[i];
					tmp_grpmbrship_lsa->LinkStateHeader.LinkStateId.NetNum = DestGrp;
			
					for(j=0; j < NoOfLinks; j++)
					{
						if(mc_table[RegGrps[i]].interface[j] == 1)
						{
						    if ( j == 0 ) {
							tmp_grpmbrship_lsa->VertexInfo[j].VertexType = ROUTER;
							tmp_grpmbrship_lsa->VertexInfo[j].VertexId.NetNum = RouterId.NetNum;
							tmp_grpmbrship_lsa->VertexInfo[j].VertexId.NodeNum = RouterId.NodeNum;
							
						    } else {
							tmp_grpmbrship_lsa->VertexInfo[j].VertexType = LinksInfo[j-1].Types;
							tmp_grpmbrship_lsa->VertexInfo[j].VertexId.NetNum =
			LinksInfo[j-1].NeighborRtrId.NetNum;
							tmp_grpmbrship_lsa->VertexInfo[j].VertexId.NodeNum =
			LinksInfo[j-1].NeighborRtrId.NodeNum;
						    }
			/*
							tmp_grpmbrship_lsa->VertexInfo[j].VertexType = LinksInfo[j].Types;
							tmp_grpmbrship_lsa->VertexInfo[j].VertexId.NetNum =LinksInfo[j].NeighborRtrId.NetNum;
							tmp_grpmbrship_lsa->VertexInfo[j].VertexId.NodeNum =LinksInfo[j].NeighborRtrId.NodeNum;
			*/
							RegGrp_Flag = ON;
						}
					}
					if( RegGrp_Flag == ON )
					{
						op_prg_mem_copy( tmp_grpmbrship_lsa, &AreaStructure.GrpMbrLsaList[DestGrp][0], 
								 sizeof(grpmbrlsa) );
						send_mospf_ls_update_pk = TRUE;
					}
					op_prg_mem_free(tmp_grpmbrship_lsa);
				}
			VL MPDL
			{
			printf("----------------(%d.%d) BCMospfLsa Information Begin------------------\n",
				RouterId.NetNum,RouterId.NodeNum);
			for (i=0; i< MAX_NO_GRPS; i++)
			{
				for (k=0; k< MAX_NO_LSAS; k++)
				{
					if(!(AreaStructure.GrpMbrLsaList[i][k].LinkStateHeader.AdvRouter.NetNum==0 ||
			 	     	     AreaStructure.GrpMbrLsaList[i][k].LinkStateHeader.AdvRouter.NodeNum==0))
					{
						printf("AreaStructure.GrpMbrLsaList[%d][%d].LinkStateHeader.AdvRouter=%d.%d\n",i,k,
							AreaStructure.GrpMbrLsaList[i][k].LinkStateHeader.AdvRouter.NetNum,
							AreaStructure.GrpMbrLsaList[i][k].LinkStateHeader.AdvRouter.NodeNum);
						for (j=0; j< MAX_NO_LINKS; j++)
						{
							if (!(AreaStructure.GrpMbrLsaList[i][k].VertexInfo[j].VertexId.NetNum == 0 ||
							      AreaStructure.GrpMbrLsaList[i][k].VertexInfo[j].VertexId.NodeNum == 0))
							{
								printf("VertexInfo[%d].VertexId=%d.%d\n",j,
									AreaStructure.GrpMbrLsaList[i][k].VertexInfo[j].VertexId.NetNum,
									AreaStructure.GrpMbrLsaList[i][k].VertexInfo[j].VertexId.NodeNum);
							}
						}
					}
				} 
			} 
			printf("----------------(%d.%d) BCMospfLsa Information End--------------------\n",
				RouterId.NetNum,RouterId.NodeNum);
			}
			
			}
			
			/* Block added 10/30 */
			if ( IS_ROUTER )
			{
				for( i=0; i < MAX_NO_GRPS; i++ )
				{
					if( !(AreaStructure.GrpMbrLsaList[i][0].LinkStateHeader.AdvRouter.NetNum == 0 || 
						AreaStructure.GrpMbrLsaList[i][0].LinkStateHeader.AdvRouter.NodeNum == 0 ) ) 
					{
						send_mospf_ls_update_pk = TRUE;
						break;
					}
			
				}
			}
			
			/*******	SEND THE MOSPF LSA UPDATE PACKET	**********/
				if(send_mospf_ls_update_pk == TRUE )
				{
					int j=0;
					int i=0;
					int LSUpdatePkCheckSum = 0;
			
					Packet* MospfLsUpdatePkptr;
					Packet* ipdgramptr;
			
					MospfLSUpdatePk* temp_ls_update_pk;
			
					temp_ls_update_pk = ( MospfLSUpdatePk *) op_prg_mem_alloc( sizeof(MospfLSUpdatePk) );
					memset( temp_ls_update_pk, 0, sizeof(MospfLSUpdatePk) );
			
					temp_ls_update_pk->OspfPkHeader.Version = VERSION;
					temp_ls_update_pk->OspfPkHeader.Type = MOSPFLSUPDATEPK;
					temp_ls_update_pk->OspfPkHeader.PacketLength = sizeof(MospfLSUpdatePk);
					op_prg_mem_copy( &(RouterId), &(temp_ls_update_pk->OspfPkHeader.ThisRouterId), 
									sizeof(IPAddress) );
					temp_ls_update_pk->OspfPkHeader.ThisAreaId = AreaId;
					temp_ls_update_pk->OspfPkHeader.CheckSum = LSUpdatePkCheckSum;
					temp_ls_update_pk->OspfPkHeader.AuType = NONE;
					temp_ls_update_pk->OspfPkHeader.Authentication = 0;
					temp_ls_update_pk->NoOfGrpMbrAdvs = 0;
			
					for( i=0; i < MAX_NO_GRPS; i++ )
					{
						j=0;
						while( temp_ls_update_pk->NoOfGrpMbrAdvs < (MAX_NO_LSAS - 1) &&
							!(AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.AdvRouter.NetNum == 0 || 
							 AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.AdvRouter.NodeNum == 0 ) ) 
						{	
							if( AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.LsType == GRPMBRSHIPLSA )
							{
								op_prg_mem_copy( &AreaStructure.GrpMbrLsaList[i][j],
								 &(temp_ls_update_pk->GrpmbrLinkStateAdv[temp_ls_update_pk->NoOfGrpMbrAdvs] ), 
								 sizeof(grpmbrlsa) );
								temp_ls_update_pk->NoOfGrpMbrAdvs++;
								NoOfMospfLsaSent++;
			/* Added by DG and Jiemei, 11/27/97 */
								NoOfInstMospfLsaSent++;
								
			
			VL MPDL 
			{	        
				printf( "TIME: %f\n", op_sim_time() );
				printf( "Sent an MospfLSUpdate Packet to IP Layer for the Group %d LSA %d\n", i ,j);
				printf("AreaStructure.GrpMbrLsaList[%d][%d].VertexInfo[0].VertexId=%d.%d\n",i,j,
					AreaStructure.GrpMbrLsaList[i][j].VertexInfo[0].VertexId.NetNum,
					AreaStructure.GrpMbrLsaList[i][j].VertexInfo[0].VertexId.NodeNum);
											
				printf("AreaStructure.GrpMbrLsaList[%d][%d].LinkStateHeader.AdvRouter=%d.%d\n",i,j,
					AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.AdvRouter.NetNum,
					AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.AdvRouter.NodeNum);
			}
							} 
							j++;
						}
					}
					MospfLsUpdatePkptr = op_pk_create_fmt("mospflsupdate_pkfmt");
					op25_pk_nfd_set( MospfLsUpdatePkptr, "Lsupdate", temp_ls_update_pk, op_prg_mem_alloc,
								   op_prg_mem_free, sizeof(MospfLSUpdatePk) );
					op25_pk_nfd_set( MospfLsUpdatePkptr, "Msg_type", MOSPFLSUPDATEPK );
			
					data_len = op_pk_total_size_get( MospfLsUpdatePkptr );
					ipdgramptr = op_pk_create_fmt("ip_dgram");
					op25_pk_nfd_set( ipdgramptr, "data", MospfLsUpdatePkptr );
					op25_pk_nfd_set( ipdgramptr, "prot_id", QOSPF_PROTOCOL_ID );
					op_pk_bulk_size_set( ipdgramptr, data_len*8 );
			
					op25_pk_nfd_set( ipdgramptr, "dest_net", AllSPFRoutersNetNum );
					op25_pk_nfd_set( ipdgramptr, "dest_node", AllSPFRoutersNodeNum );
					op25_pk_nfd_set( ipdgramptr, "ttl", 2 );
			
					op25_pk_nfd_set( ipdgramptr, "src_net", RouterId.NetNum );
					op25_pk_nfd_set( ipdgramptr, "src_node", RouterId.NodeNum );
			
					op25_pk_nfd_set( ipdgramptr, "orig_len", data_len );
					op25_pk_nfd_set( ipdgramptr, "frag_len", data_len );
			
					op25_pk_nfd_set( ipdgramptr, "frag", 0 );
			
			
					op_pk_send( ipdgramptr, IPSTREAM );
					PROBE_ON
					{
						op_stat_local_write(CUMM_MOSPF_LSA_SENT_STAT, NoOfMospfLsaSent);
			/* Added by DG and Jiemei, 11/27/97 */
						op_stat_local_write(INST_MOSPF_LSA_SENT_STAT, NoOfInstMospfLsaSent);
					}
			
				}
			}
			
			if ( op_intrpt_code() == 3 )
			{
			
			    if ( IS_ROUTER) 
			    {
			        if( op_sim_time() < BCMospfLsa_sim_time + MAX_REFRESH_TIME ) /*changed by Jiemei*/
				{
					if ( MSentLsaCycleFirst == 0 )
					{
						MSentLsaCycleFirst = 1;
						NoOfInstMospfLsaSent = 0;
					}
			                op_intrpt_schedule_self(op_sim_time()+1, 3);
				}
			        else     
			                {
			                        op_intrpt_schedule_self(op_sim_time()+MospfLsaRefreshTime-MAX_REFRESH_TIME, 3);
			                        BCMospfLsa_sim_time = op_sim_time()+MospfLsaRefreshTime-MAX_REFRESH_TIME;  /*changed by Jiemei*/
						MSentLsaCycleFirst = 0;
			                }
			
			
			    } 
			    else 
			    {
				NoOfInstMospfLsaSent = 0;
				op_intrpt_schedule_self(op_sim_time()+MospfLsaRefreshTime, 3);
			    }
			}
			
			send_grp_mbr_lsa = 0;
			}


		/** state (BCMospfLsa) exit executives **/
		FSM_STATE_EXIT_FORCED (10, state10_exit_exec, "BCMospfLsa", "QospfRouterProcesses () [BCMospfLsa exit execs]")
			{
			}


		/** state (BCMospfLsa) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (hello_pks) enter executives **/
		FSM_STATE_ENTER_FORCED (11, state11_enter_exec, "hello_pks", "QospfRouterProcesses () [hello_pks enter execs]")
			{
			/**
			*** This state creates the hello packet that needs to be sent to all the neighbors. The hello
			*** packet sets the destination as AllSPFRouters which indicates that this packet should be 
			*** sent to only the other end of the links and the packet shouldn't be forwarded from there.
			*** The source address of the packet is the Router Id of the router originating the packet
			**/
			
			int NoOfRtrsFrmWhichHelloRec;
			hellopacket* tmp_hello_pk;
			Packet* helloPkptr;
			Packet* ipdgramptr;
			
			tmp_hello_pk = (hellopacket*) op_prg_mem_alloc( sizeof(hellopacket) );
			memset( tmp_hello_pk, 0, sizeof(hellopacket) );
			
			
			tmp_hello_pk->OspfPkHeader.Version = VERSION;
			tmp_hello_pk->OspfPkHeader.Type = HELLOPK;
			tmp_hello_pk->OspfPkHeader.PacketLength = sizeof(hellopacket);
			tmp_hello_pk->OspfPkHeader.ThisRouterId.NetNum = RouterId.NetNum;
			tmp_hello_pk->OspfPkHeader.ThisRouterId.NodeNum = RouterId.NodeNum;
			tmp_hello_pk->OspfPkHeader.ThisAreaId = AreaId;
			tmp_hello_pk->OspfPkHeader.AuType = NONE;
			tmp_hello_pk->OspfPkHeader.Authentication = 0;
			tmp_hello_pk->ThisNetworkMask = NetworkMask;
			tmp_hello_pk->Deadint = DeadInt;
			tmp_hello_pk->Helloint = HelloInt;
			tmp_hello_pk->ThisRtrPri = RtrPri;
			tmp_hello_pk->Indep_Router = Indep_Rou;
			op_prg_mem_copy( &(tmp_hello_pk->DesgRouter), &(DsgRouterIntfAddr), sizeof(IPAddress) );
			op_prg_mem_copy( &(tmp_hello_pk->BackupDesgRouter), &(BackupDsgRouterIntfAddr), sizeof(IPAddress) );
			NoOfRtrsFrmWhichHelloRec = 0;
			if(!IsListEmpty(&NeighborList) )
			{
			
			
				Neighbor_Data_Str* neighborEle;
			
			
				neighborEle = (Neighbor_Data_Str*) op_prg_mem_alloc( sizeof(Neighbor_Data_Str) );
			
				neighborEle = (Neighbor_Data_Str*) GetNextListEle( &(NeighborList), 1 );
				while(neighborEle)
				{
					op_prg_mem_copy( &(tmp_hello_pk->RtrsFrmWhichHelloRec[NoOfRtrsFrmWhichHelloRec]), 
									 &(neighborEle->NeighborRouterId), sizeof(IPAddress) );
					NoOfRtrsFrmWhichHelloRec++;
					neighborEle = (Neighbor_Data_Str*) GetNextListEle( &(NeighborList), 0 );
				}
				op_prg_mem_free( neighborEle );
			}
			
			helloPkptr = op_pk_create_fmt("hello_pkfmt");
			op25_pk_nfd_set( helloPkptr, "hellopk", tmp_hello_pk, op_prg_mem_alloc, op_prg_mem_free, sizeof(hellopacket) );
			
			data_len = op_pk_total_size_get( helloPkptr )/8;
			ipdgramptr = op_pk_create_fmt("ip_dgram");
			op25_pk_nfd_set( ipdgramptr, "data", helloPkptr );
			op25_pk_nfd_set( ipdgramptr, "prot_id", QOSPF_PROTOCOL_ID );
			op_pk_bulk_size_set( ipdgramptr, data_len*8 );
			op25_pk_nfd_set( ipdgramptr, "dest_net", AllSPFRoutersNetNum );
			op25_pk_nfd_set( ipdgramptr, "dest_node", AllSPFRoutersNodeNum );
			op25_pk_nfd_set( ipdgramptr, "ttl", 2 );
			
			op25_pk_nfd_set( ipdgramptr, "src_net", net_num0);
			op25_pk_nfd_set( ipdgramptr, "src_node", node_num0);
			
			op25_pk_nfd_set( ipdgramptr, "orig_len", data_len);
			op25_pk_nfd_set( ipdgramptr, "frag_len", data_len);
			
			op25_pk_nfd_set( ipdgramptr, "frag", 0 );
			
			/*  printf("\nSending Hello Packet from the Router %d.%d \n", RouterId.NetNum, RouterId.NodeNum);  */
			op_pk_send( ipdgramptr, IPSTREAM );
			
			op_intrpt_schedule_self( op_sim_time()+HelloPkIntvl, 2 );
			}


		/** state (hello_pks) exit executives **/
		FSM_STATE_EXIT_FORCED (11, state11_exit_exec, "hello_pks", "QospfRouterProcesses () [hello_pks exit execs]")
			{
			}


		/** state (hello_pks) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (BCLsReqPk) enter executives **/
		FSM_STATE_ENTER_FORCED (12, state12_enter_exec, "BCLsReqPk", "QospfRouterProcesses () [BCLsReqPk enter execs]")
			{
			
			}


		/** state (BCLsReqPk) exit executives **/
		FSM_STATE_EXIT_FORCED (12, state12_exit_exec, "BCLsReqPk", "QospfRouterProcesses () [BCLsReqPk exit execs]")
			{
			}


		/** state (BCLsReqPk) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		/** state (upstr_node) enter executives **/
		FSM_STATE_ENTER_FORCED (13, state13_enter_exec, "upstr_node", "QospfRouterProcesses () [upstr_node enter execs]")
			{
			
			send_dabra = 1;
			detupstrnode = 0;
			
			}


		/** state (upstr_node) exit executives **/
		FSM_STATE_EXIT_FORCED (13, state13_exit_exec, "upstr_node", "QospfRouterProcesses () [upstr_node exit execs]")
			{
			}


		/** state (upstr_node) transition processing **/
		FSM_TRANSIT_ONLY ((send_dabra), 3, state3_enter_exec, ;, "upstr_node")
		/*---------------------------------------------------------*/



		/** state (BCdatadescpk) enter executives **/
		FSM_STATE_ENTER_FORCED (14, state14_enter_exec, "BCdatadescpk", "QospfRouterProcesses () [BCdatadescpk enter execs]")
			{
			
			}


		/** state (BCdatadescpk) exit executives **/
		FSM_STATE_EXIT_FORCED (14, state14_exit_exec, "BCdatadescpk", "QospfRouterProcesses () [BCdatadescpk exit execs]")
			{
			}


		/** state (BCdatadescpk) transition processing **/
		FSM_TRANSIT_FORCE (1, state1_enter_exec, ;)
		/*---------------------------------------------------------*/



		}


	FSM_EXIT (0,QospfRouterProcesses)
	}




void
QospfRouterProcesses_svar (prs_ptr,var_name,var_p_ptr)
	QospfRouterProcesses_state		*prs_ptr;
	char			*var_name, **var_p_ptr;
	{

	FIN (QospfRouterProcesses_svar (prs_ptr))

	*var_p_ptr = VOS_NIL;
	if (Vos_String_Equal ("my_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_my_id);
	if (Vos_String_Equal ("ip_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_objid);
	if (Vos_String_Equal ("rsvp_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_rsvp_objid);
	if (Vos_String_Equal ("igmp_objid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_igmp_objid);
	if (Vos_String_Equal ("sub_netid" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_sub_netid);
	if (Vos_String_Equal ("ip_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ip_id);
	if (Vos_String_Equal ("q_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_q_id);
	if (Vos_String_Equal ("rx_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_rx_id);
	if (Vos_String_Equal ("tx_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_tx_id);
	if (Vos_String_Equal ("dup_lk_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_dup_lk_id);
	if (Vos_String_Equal ("qospf_id" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_qospf_id);
	if (Vos_String_Equal ("IS_ROUTER" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_IS_ROUTER);
	if (Vos_String_Equal ("IS_DESIGNATED_ROUTER" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_IS_DESIGNATED_ROUTER);
	if (Vos_String_Equal ("IS_AREA_BORDER_ROUTER" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_IS_AREA_BORDER_ROUTER);
	if (Vos_String_Equal ("net_num0" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num0);
	if (Vos_String_Equal ("node_num0" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num0);
	if (Vos_String_Equal ("net_num1" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num1);
	if (Vos_String_Equal ("node_num1" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num1);
	if (Vos_String_Equal ("net_num2" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num2);
	if (Vos_String_Equal ("node_num2" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num2);
	if (Vos_String_Equal ("net_num3" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num3);
	if (Vos_String_Equal ("node_num3" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num3);
	if (Vos_String_Equal ("net_num4" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num4);
	if (Vos_String_Equal ("node_num4" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num4);
	if (Vos_String_Equal ("net_num5" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num5);
	if (Vos_String_Equal ("node_num5" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num5);
	if (Vos_String_Equal ("net_num6" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num6);
	if (Vos_String_Equal ("node_num6" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num6);
	if (Vos_String_Equal ("net_num7" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num7);
	if (Vos_String_Equal ("node_num7" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num7);
	if (Vos_String_Equal ("net_num8" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num8);
	if (Vos_String_Equal ("node_num8" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num8);
	if (Vos_String_Equal ("net_num9" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_net_num9);
	if (Vos_String_Equal ("node_num9" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_node_num9);
	if (Vos_String_Equal ("res_ls_seq_no" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_res_ls_seq_no);
	if (Vos_String_Equal ("rra_ls_seq_no" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_rra_ls_seq_no);
	if (Vos_String_Equal ("router_ls_seq_no" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_router_ls_seq_no);
	if (Vos_String_Equal ("summary_ls_seq_no" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_summary_ls_seq_no);
	if (Vos_String_Equal ("network_ls_seq_no" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_network_ls_seq_no);
	if (Vos_String_Equal ("grpmbr_ls_seq_no" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_grpmbr_ls_seq_no);
	if (Vos_String_Equal ("NoOfOspfLsaRecv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfOspfLsaRecv);
	if (Vos_String_Equal ("NoOfMospfLsaRecv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfMospfLsaRecv);
	if (Vos_String_Equal ("NoOfQospfLsaRecv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfQospfLsaRecv);
	if (Vos_String_Equal ("NoOfTotalLsaRecv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfTotalLsaRecv);
	if (Vos_String_Equal ("OspfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_OspfDijCalcTime);
	if (Vos_String_Equal ("MospfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MospfDijCalcTime);
	if (Vos_String_Equal ("QospfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_QospfDijCalcTime);
	if (Vos_String_Equal ("TotalDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_TotalDijCalcTime);
	if (Vos_String_Equal ("TotalLsaTraffic" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_TotalLsaTraffic);
	if (Vos_String_Equal ("NoOfOspfLsaSent" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfOspfLsaSent);
	if (Vos_String_Equal ("NoOfMospfLsaSent" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfMospfLsaSent);
	if (Vos_String_Equal ("NoOfQospfLsaSent" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfQospfLsaSent);
	if (Vos_String_Equal ("MaxOspfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MaxOspfDijCalcTime);
	if (Vos_String_Equal ("MinOspfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MinOspfDijCalcTime);
	if (Vos_String_Equal ("MaxMospfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MaxMospfDijCalcTime);
	if (Vos_String_Equal ("MinMospfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MinMospfDijCalcTime);
	if (Vos_String_Equal ("MaxQospfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MaxQospfDijCalcTime);
	if (Vos_String_Equal ("MinQospfDijCalcTime" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MinQospfDijCalcTime);
	if (Vos_String_Equal ("NoOfQospfCalcDone" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfQospfCalcDone);
	if (Vos_String_Equal ("NoOfMospfCalcDone" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfMospfCalcDone);
	if (Vos_String_Equal ("NoOfOspfCalcDone" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfOspfCalcDone);
	if (Vos_String_Equal ("TotalDijCalcDone" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_TotalDijCalcDone);
	if (Vos_String_Equal ("send_grpmbrlsa" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_grpmbrlsa);
	if (Vos_String_Equal ("send_grp_mbr_lsa" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_grp_mbr_lsa);
	if (Vos_String_Equal ("send_reslsa" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_reslsa);
	if (Vos_String_Equal ("send_rralsa" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_rralsa);
	if (Vos_String_Equal ("send_ospflsa" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_ospflsa);
	if (Vos_String_Equal ("send_lsreq_pk" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_lsreq_pk);
	if (Vos_String_Equal ("spfcalc" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_spfcalc);
	if (Vos_String_Equal ("detupstrnode" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_detupstrnode);
	if (Vos_String_Equal ("send_dabra" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_dabra);
	if (Vos_String_Equal ("send_datadesc" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_datadesc);
	if (Vos_String_Equal ("qospfcalc" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_qospfcalc);
	if (Vos_String_Equal ("ospfspfcalc" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ospfspfcalc);
	if (Vos_String_Equal ("mospfspfcalc" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_mospfspfcalc);
	if (Vos_String_Equal ("send_ls_update_pk" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_ls_update_pk);
	if (Vos_String_Equal ("send_hellopk" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_send_hellopk);
	if (Vos_String_Equal ("Ip_Network_No" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_Ip_Network_No);
	if (Vos_String_Equal ("Network_Mask" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_Network_Mask);
	if (Vos_String_Equal ("NoOfLinks" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfLinks);
	if (Vos_String_Equal ("Indep_Rou" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_Indep_Rou);
	if (Vos_String_Equal ("AreaId" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_AreaId);
	if (Vos_String_Equal ("DeadInt" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_DeadInt);
	if (Vos_String_Equal ("HelloInt" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_HelloInt);
	if (Vos_String_Equal ("RtrPri" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RtrPri);
	if (Vos_String_Equal ("NetworkMask" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NetworkMask);
	if (Vos_String_Equal ("No_of_Networks" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_No_of_Networks);
	if (Vos_String_Equal ("NoOfVertices" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfVertices);
	if (Vos_String_Equal ("No_of_TOS" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_No_of_TOS);
	if (Vos_String_Equal ("Network_Routers" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_Network_Routers);
	if (Vos_String_Equal ("TOS_Metric" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_TOS_Metric);
	if (Vos_String_Equal ("LinksInfo" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_LinksInfo);
	if (Vos_String_Equal ("DsgRouterIntfAddr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_DsgRouterIntfAddr);
	if (Vos_String_Equal ("BackupDsgRouterIntfAddr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_BackupDsgRouterIntfAddr);
	if (Vos_String_Equal ("DsgRtr_IpAddr" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_DsgRtr_IpAddr);
	if (Vos_String_Equal ("RouterId" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RouterId);
	if (Vos_String_Equal ("DefaultDestination" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_DefaultDestination);
	if (Vos_String_Equal ("NeighborsInfo" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_NeighborsInfo);
	if (Vos_String_Equal ("QospfPSBList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_QospfPSBList);
	if (Vos_String_Equal ("ResvConfInfoList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_ResvConfInfoList);
	if (Vos_String_Equal ("NeighborList" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NeighborList);
	if (Vos_String_Equal ("NoOfAreasRouterIsConnected" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfAreasRouterIsConnected);
	if (Vos_String_Equal ("AreaStructure" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_AreaStructure);
	if (Vos_String_Equal ("BackboneStructure" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_BackboneStructure);
	if (Vos_String_Equal ("RoutingTable" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_RoutingTable);
	if (Vos_String_Equal ("RegMcNets" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_RegMcNets);
	if (Vos_String_Equal ("mc_rte_table" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_mc_rte_table);
	if (Vos_String_Equal ("obt_pk_srcnet" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_obt_pk_srcnet);
	if (Vos_String_Equal ("obt_pk_srcnode" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_obt_pk_srcnode);
	if (Vos_String_Equal ("NoOfNeighborElements" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfNeighborElements);
	if (Vos_String_Equal ("InstNoLsa" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_InstNoLsa);
	if (Vos_String_Equal ("print_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_print_time);
	if (Vos_String_Equal ("BCOspfLsa_sim_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_BCOspfLsa_sim_time);
	if (Vos_String_Equal ("BCQospfLsa_sim_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_BCQospfLsa_sim_time);
	if (Vos_String_Equal ("BCMospfLsa_sim_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_BCMospfLsa_sim_time);
	if (Vos_String_Equal ("SessionDirectory" , var_name))
		*var_p_ptr = (char *) (prs_ptr->sv_SessionDirectory);
	if (Vos_String_Equal ("CurrSess" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_CurrSess);
	if (Vos_String_Equal ("NoOfInstOspfLsaRecv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfInstOspfLsaRecv);
	if (Vos_String_Equal ("NoOfInstMospfLsaRecv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfInstMospfLsaRecv);
	if (Vos_String_Equal ("NoOfInstQospfLsaRecv" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfInstQospfLsaRecv);
	if (Vos_String_Equal ("NoOfInstOspfLsaSent" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfInstOspfLsaSent);
	if (Vos_String_Equal ("NoOfInstMospfLsaSent" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfInstMospfLsaSent);
	if (Vos_String_Equal ("NoOfInstQospfLsaSent" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_NoOfInstQospfLsaSent);
	if (Vos_String_Equal ("MRecvLsaCycleFirst" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MRecvLsaCycleFirst);
	if (Vos_String_Equal ("URecvLsaCycleFirst" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_URecvLsaCycleFirst);
	if (Vos_String_Equal ("QRecvLsaCycleFirst" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_QRecvLsaCycleFirst);
	if (Vos_String_Equal ("MSentLsaCycleFirst" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MSentLsaCycleFirst);
	if (Vos_String_Equal ("USentLsaCycleFirst" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_USentLsaCycleFirst);
	if (Vos_String_Equal ("QSentLsaCycleFirst" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_QSentLsaCycleFirst);
	if (Vos_String_Equal ("OspfLsa_Recv_sim_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_OspfLsa_Recv_sim_time);
	if (Vos_String_Equal ("MospfLsa_Recv_sim_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_MospfLsa_Recv_sim_time);
	if (Vos_String_Equal ("QospfLsa_Recv_sim_time" , var_name))
		*var_p_ptr = (char *) (&prs_ptr->sv_QospfLsa_Recv_sim_time);

	FOUT;
	}




void
QospfRouterProcesses_diag ()
	{
	
		int in_strm;
		int ip_strm;
		int data_len;
	
		int tmp_msg_type;
		int protocolid;
	
		int NeighborFoundFlag;
	
		Packet *ReceivedPkptr;
		Packet *ReceivedRsvpPkptr;
		Packet *ReceivedQospfPkptr;
	
		mc_id           *mc_table;
	
		double tmp_sim_time;
	int _block_origin;

	FIN (QospfRouterProcesses_diag ())

	BINIT
	

	FOUT;
	}




void
QospfRouterProcesses_terminate ()
	{
	
		int in_strm;
		int ip_strm;
		int data_len;
	
		int tmp_msg_type;
		int protocolid;
	
		int NeighborFoundFlag;
	
		Packet *ReceivedPkptr;
		Packet *ReceivedRsvpPkptr;
		Packet *ReceivedQospfPkptr;
	
		mc_id           *mc_table;
	
		double tmp_sim_time;
	int _block_origin;

	FIN (QospfRouterProcesses_terminate ())

	BINIT
	

	FOUT;
	}



Compcode
QospfRouterProcesses_init (pr_state_pptr)
	QospfRouterProcesses_state			**pr_state_pptr;
	{
	int _block_origin = 0;
	static VosT_Cm_Obtype	obtype = OPC_NIL;

	extern int				Vos_Nop ();

	FIN (QospfRouterProcesses_init (pr_state_pptr))

	if (obtype == OPC_NIL)
		{
		if (Vos_Catmem_Register ("proc state vars (QospfRouterProcesses)",
			sizeof (QospfRouterProcesses_state), Vos_Nop, &obtype) == VOSC_FAILURE)
			FRET (OPC_COMPCODE_FAILURE)
		}

	if ((*pr_state_pptr = (QospfRouterProcesses_state*) Vos_Catmem_Alloc (obtype, 1)) == OPC_NIL)
		FRET (OPC_COMPCODE_FAILURE)
	else
		{
		(*pr_state_pptr)->current_block = 0;
		FRET (OPC_COMPCODE_SUCCESS)
		}
	}



enum { _block_origin = __LINE__ };
int NeighborFoundFlag;
Packet* ReceivedRsvpPkptr; 
Packet* ReceivedOspfPkptr; 

int 
get_transmiting_link_id(double destaddr)
{
	 int tmp_link_id; 
	 
	 return tmp_link_id;
}

void 
trigger_the_route_calc() 
{ 
	 spfcalc = 1;
	 return; 
} 

/**
*** Hello Packet Processing 
**/
void
HelloPk_pro(Packet * ReceivedOspfPkptr)
{
	int NeighborFoundFlag = 0;
	int i;
   	hellopacket* tmp_hellopk;

	/*tmp_hellopk = (hellopacket*) op_prg_mem_alloc( sizeof(hellopacket) ); commented by Jiemei*/
	op_pk_nfd_get(ReceivedOspfPkptr, "hellopk", &tmp_hellopk); /* changed by Jiemei from &tmp_hellopk */
	if( tmp_hellopk->Deadint != DeadInt ||
    		tmp_hellopk->Helloint != HelloInt ||
        	tmp_hellopk->ThisNetworkMask != Network_Mask)
	{
		op_prg_mem_free( tmp_hellopk );  /* added by Sitaram */
		return;
	} 
	for(i = 0; i< MAX_NO_NEIGHBORS; i++ )  /* found by Sitaram */
	{
		if( tmp_hellopk->OspfPkHeader.ThisRouterId.NetNum == 
			NeighborsInfo[i].NeighborRouterId.NetNum && 
			tmp_hellopk->OspfPkHeader.ThisRouterId.NodeNum == 
			NeighborsInfo[i].NeighborRouterId.NodeNum )
		{
				NeighborFoundFlag = 1;
				break;
		}
	}
   	if(NeighborFoundFlag == 0)
   	{		
		for(i = 0; i< MAX_NO_NEIGHBORS; i++ ) /* changed by Jiemei from i<=MAX_NO_NEIGHBORS*/
		{
			if( NeighborsInfo[i].NeighborRouterId.NodeNum == 0 &&
				NeighborsInfo[i].NeighborRouterId.NetNum == 0 )
			{
				NeighborsInfo[i].NeighborRouterId.NetNum =
					tmp_hellopk->OspfPkHeader.ThisRouterId.NetNum;
				NeighborsInfo[i].NeighborRouterId.NodeNum =
					tmp_hellopk->OspfPkHeader.ThisRouterId.NodeNum;
				NeighborsInfo[i].IsIndepRou = tmp_hellopk->Indep_Router;
				break;
			}
		}
	}
		
	VL	HPDL	
	{
		printf("\n Router(%d.%d), has the following Neighbors ", RouterId.NetNum, RouterId.NodeNum );
		for(i = 0; i< MAX_NO_NEIGHBORS; i++ ) /* changed by Jiemei from i<=MAX_NO_NEIGHBORS*/
		{
			if( !(NeighborsInfo[i].NeighborRouterId.NodeNum == 0 || 
				  NeighborsInfo[i].NeighborRouterId.NetNum == 0 ) )
				  printf( "%d.%d ,", NeighborsInfo[i].NeighborRouterId.NetNum, 
						  NeighborsInfo[i].NeighborRouterId.NodeNum );
		}
	}
	op_prg_mem_free( tmp_hellopk );
	return;
}
/**
*** Function : QospfLsUpdatePk_pro
*** Input    : A pointer to the recived Qospf packet pointer is passed.
*** Output   : This function will check for the Lsa's that are in the Qospf Ls update packet and update the 
***            local databases
**/
void
QospfLsUpdatePk_pro( Packet *ReceivedQospfPkptr )
{
	int i = 0;
	int j = 0;
	int FoundFlag = OFF;
	QospfLSUpdatePk* qupdate;
	
	/*qupdate = (QospfLSUpdatePk *) op_prg_mem_alloc( sizeof(QospfLSUpdatePk) );
	memset( qupdate, 0, sizeof(QospfLSUpdatePk) );     commented by Jiemei */

	op_pk_nfd_get( ReceivedQospfPkptr, "Lsupdate" , &qupdate );
	if( qupdate->OspfPkHeader.Type != QOSPFLSUPDATEPK )
	{
TRACE
		VL	QPDL	printf(" Not Valid OspfPacket Header Type %d \n", qupdate->OspfPkHeader.Type );
		op_prg_mem_free( qupdate );
		return;
	}
	else
	{
			for( i = 0; i < qupdate->NoOfResAdvs; i++ )
			{
				FoundFlag = OFF;
				for(j = 0; j < MAX_NO_LSAS; j++ )
				{
					if( (AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NetNum ==
						 qupdate->ResLinkStateAdv[i].ResLsaHeader.AdvRouter.NetNum &&
						 AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NodeNum ==
						 qupdate->ResLinkStateAdv[i].ResLsaHeader.AdvRouter.NodeNum) )

					{
						FoundFlag = ON;
						if( IS_AREA_BORDER_ROUTER == 1 ||
							qupdate->ResLinkStateAdv[i].ResLsaHeader.ThisAreaId == AreaId )
						{
							op_prg_mem_copy( &(qupdate->ResLinkStateAdv[i]), 
									 &AreaStructure.ResourceLsaList[j], sizeof(reslsa) );
						}
						break;
					}

				}
				if( FoundFlag == OFF )
				{
					for(j = 0; j < MAX_NO_LSAS; j++ )
					{
						if( (AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NetNum == 0 &&
							 AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NodeNum == 0 &&
							 AreaStructure.ResourceLsaList[j].ResLsaHeader.LSSeqNo == 0) )
						{
							if( IS_AREA_BORDER_ROUTER == 1 ||
								qupdate->ResLinkStateAdv[i].ResLsaHeader.ThisAreaId == AreaId )
								op_prg_mem_copy( &(qupdate->ResLinkStateAdv[i]), 
										 &AreaStructure.ResourceLsaList[j], sizeof(reslsa) );
							break;
						}
					}
				}
			}
TRACE VL QPDL	printf("---------------------- Qospf LsUpdate Packet ----------------------\n");

			for( i = 0; i < qupdate->NoOfRraAdvs; i++ )
			{
TRACE VL printf( "Receive RRA from (%d.%d) SeqNo=%f\n",
	qupdate->RraLinkStateAdv[i].OpaqueLsaHeader.AdvRouter.NetNum,
	qupdate->RraLinkStateAdv[i].OpaqueLsaHeader.AdvRouter.NodeNum,
	qupdate->RraLinkStateAdv[i].OpaqueLsaHeader.LSSeqNo );


				FoundFlag = OFF;
				for(j = 0; j < MAX_NO_LSAS; j++ )
				{
					if( (AreaStructure.RRALsaList[j].OpaqueLsaHeader.AdvRouter.NetNum ==
						 qupdate->RraLinkStateAdv[i].OpaqueLsaHeader.AdvRouter.NetNum) &&
						(AreaStructure.RRALsaList[j].OpaqueLsaHeader.AdvRouter.NodeNum ==
						 qupdate->RraLinkStateAdv[i].OpaqueLsaHeader.AdvRouter.NodeNum) )
					{
TRACE VL printf( "	Found Matched ADV, seqNo= %f\n", AreaStructure.RRALsaList[j].OpaqueLsaHeader.LSSeqNo);

						FoundFlag = ON;

						if (qupdate->RraLinkStateAdv[i].OpaqueLsaHeader.LSSeqNo >=
						 	AreaStructure.RRALsaList[j].OpaqueLsaHeader.LSSeqNo)
						{
						if( IS_AREA_BORDER_ROUTER == 1 ||
							qupdate->RraLinkStateAdv[i].OpaqueLsaHeader.ThisAreaId == AreaId )
						{
							op_prg_mem_copy( &(qupdate->RraLinkStateAdv[i]), 
									&AreaStructure.RRALsaList[j],sizeof(rralsa) );
						}
						}
						break;
					}

				}
				if( FoundFlag == OFF )
				{
					for(j = 0; j < MAX_NO_LSAS; j++ )
					{
						if( (AreaStructure.RRALsaList[j].OpaqueLsaHeader.AdvRouter.NetNum == 0 &&
							 AreaStructure.RRALsaList[j].OpaqueLsaHeader.AdvRouter.NodeNum == 0 &&
							 AreaStructure.RRALsaList[j].OpaqueLsaHeader.LSSeqNo == 0) )
						{
							if( IS_AREA_BORDER_ROUTER == 1 ||
								qupdate->RraLinkStateAdv[i].OpaqueLsaHeader.ThisAreaId == AreaId )
								op_prg_mem_copy( &(qupdate->RraLinkStateAdv[i]), 
										&AreaStructure.RRALsaList[j], sizeof(rralsa) );
							break;
						}
						   
					}
				}
			}
	}

	PROBE_ON
	{
		NoOfQospfLsaRecv = NoOfQospfLsaRecv + qupdate->NoOfRraAdvs + qupdate->NoOfResAdvs;
		op_stat_local_write( CUMM_QOSPF_LSA_RECV_STAT, NoOfQospfLsaRecv );
		NoOfTotalLsaRecv = NoOfTotalLsaRecv + qupdate->NoOfRraAdvs + qupdate->NoOfResAdvs;
		op_stat_local_write( CUMM_TOTAL_LSA_RECV_STAT, NoOfTotalLsaRecv );

/* Added by Jiemei and Dg 11/27 */
		NoOfInstQospfLsaRecv = NoOfInstQospfLsaRecv + qupdate->NoOfRraAdvs + qupdate->NoOfResAdvs;
		op_stat_local_write( INST_QOSPF_LSA_RECV_STAT, NoOfInstQospfLsaRecv );

		if( op_sim_time() > QospfLsa_Recv_sim_time + QospfLsaRefreshTime )
		{
        		NoOfInstQospfLsaRecv = 0;
        		QospfLsa_Recv_sim_time = op_sim_time();
 		}
	}

	op_prg_mem_free( qupdate );
	qospfcalc = TRUE;

	VL	QPDL
	{
		printf("---------------------- Qospf LsUpdate Packet ----------------------\n");
		printf( "(%d.%d)Reservation Lsa's AdvRouter: \n", RouterId.NetNum, RouterId.NodeNum );
		for( i = 0; i < MAX_NO_LSAS; i++ )
		{
			if( !(AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NetNum == 0 || 
				  AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NetNum == 0 ) )
				printf(" (%d.%d) DEST(%d.%d)\n ",
				AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NetNum,
				AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NodeNum,
				AreaStructure.RRALsaList[i].DestAddr.NetNum,
				AreaStructure.RRALsaList[i].DestAddr.NodeNum );
		}
		printf("\n");
		printf( "(%d.%d)Resource Lsa's AdvRouter: ", RouterId.NetNum, RouterId.NodeNum );
		for( i = 0; i < MAX_NO_LSAS; i++ )
		{
			if( !(AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NetNum == 0 || 
				  AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NetNum == 0 ) )
			printf(" %d.%d ",
			AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NetNum,
			AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NodeNum );
		}
		printf("\n");
		printf("-------------------- End Qospf LsUpdate Packet --------------------\n");
	}
	return;
}

void
MospfLsUpdatePk_pro( Packet *ReceivedQospfPkptr )
{
	int i = 0;
	int j = 0;
	int FoundFlag = OFF;
	int Dest_Grp;

	MospfLSUpdatePk *qupdate;

/* Add two lines by DG, 11/8/97, Get MC_table in IP, clear infValid when receiving a update pkt. */
	mc_id *mc_table;
	mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");



VL	MPDL	printf("---------- Mospf LS Update Packet Processing ---------- \n");
	/*qupdate = (MospfLSUpdatePk *) op_prg_mem_alloc( sizeof(MospfLSUpdatePk) );
	memset( qupdate, 0, sizeof(MospfLSUpdatePk) );   commented by Jiemei*/

	op_pk_nfd_get( ReceivedQospfPkptr, "Lsupdate" , &qupdate );
	if( qupdate->OspfPkHeader.Type != MOSPFLSUPDATEPK )
	{
		op_prg_mem_free( qupdate );
		return;
	}
	else
	{
/* DG,10/7/97	if(qupdate->OspfPkHeader.ThisAreaId = AreaId || IS_AREA_BORDER_ROUTER ) */
/* Comment next line, move into inner loop, 10/31/97 */
/*		if(qupdate->OspfPkHeader.ThisAreaId == AreaId || IS_AREA_BORDER_ROUTER ) */
		{
			PROBE_ON
			{
				NoOfMospfLsaRecv = NoOfMospfLsaRecv + qupdate->NoOfGrpMbrAdvs;
				op_stat_local_write( CUMM_MOSPF_LSA_RECV_STAT, NoOfMospfLsaRecv );
				NoOfTotalLsaRecv = NoOfTotalLsaRecv + qupdate->NoOfGrpMbrAdvs;
				op_stat_local_write( CUMM_TOTAL_LSA_RECV_STAT, NoOfTotalLsaRecv );
/* Added by Jiemei and Dg 11/27 */
				NoOfInstMospfLsaRecv = NoOfInstMospfLsaRecv + qupdate->NoOfGrpMbrAdvs;
				op_stat_local_write( INST_MOSPF_LSA_RECV_STAT, NoOfInstMospfLsaRecv );

				if( op_sim_time() > MospfLsa_Recv_sim_time + MospfLsaRefreshTime  )
				{
        				NoOfInstMospfLsaRecv = 0;
       		 			MospfLsa_Recv_sim_time = op_sim_time();
 				}
			}
VL    MPDL printf( " No of Grp Mbr Advs = %d\n", qupdate->NoOfGrpMbrAdvs );



			for( i = 0; i < qupdate->NoOfGrpMbrAdvs && i < MAX_NO_LSAS; i++ )    
			{

				FoundFlag = OFF;
				Dest_Grp = qupdate->GrpmbrLinkStateAdv[i].LinkStateHeader.LinkStateId.NetNum;

VL    MPDL printf("Recevied a MospfLsa for the group %d from %d.%d\n", 
			Dest_Grp, obt_pk_srcnet, obt_pk_srcnode );

				j=0;
				while( !(AreaStructure.GrpMbrLsaList[Dest_Grp][j].LinkStateHeader.AdvRouter.NetNum == 0 || 
					 AreaStructure.GrpMbrLsaList[Dest_Grp][j].LinkStateHeader.AdvRouter.NodeNum == 0) ) 
				{
VL    MPDL
{
	printf("AreaStructure.GrpMbrLsaList[%d][%d].VertexInfo[0].VertexId=%d.%d\n",Dest_Grp,j,
		AreaStructure.GrpMbrLsaList[Dest_Grp][j].VertexInfo[0].VertexId.NetNum,
		AreaStructure.GrpMbrLsaList[Dest_Grp][j].VertexInfo[0].VertexId.NodeNum);
										
	printf("AreaStructure.GrpMbrLsaList[%d][%d].LinkStateHeader.AdvRouter=%d.%d\n",Dest_Grp,j,
		AreaStructure.GrpMbrLsaList[Dest_Grp][j].LinkStateHeader.AdvRouter.NetNum,
		AreaStructure.GrpMbrLsaList[Dest_Grp][j].LinkStateHeader.AdvRouter.NodeNum);
		
	printf("qupdate->GrpmbrLinkStateAdv[%d].LinkStateHeader.AdvRouter%d.%d\n",i,
		qupdate->GrpmbrLinkStateAdv[i].LinkStateHeader.AdvRouter.NetNum,
		qupdate->GrpmbrLinkStateAdv[i].LinkStateHeader.AdvRouter.NodeNum);
}									
		    
					if( AreaStructure.GrpMbrLsaList[Dest_Grp][j].LinkStateHeader.AdvRouter.NetNum ==
						qupdate->GrpmbrLinkStateAdv[i].LinkStateHeader.AdvRouter.NetNum &&
						AreaStructure.GrpMbrLsaList[Dest_Grp][j].LinkStateHeader.AdvRouter.NodeNum ==
						qupdate->GrpmbrLinkStateAdv[i].LinkStateHeader.AdvRouter.NodeNum )
					{

						if( qupdate->GrpmbrLinkStateAdv[i].LinkStateHeader.LSSeqNo >=
							AreaStructure.GrpMbrLsaList[Dest_Grp][j].LinkStateHeader.LSSeqNo
						     && ( qupdate->GrpmbrLinkStateAdv[i].LinkStateHeader.ThisAreaId == AreaId
							|| IS_AREA_BORDER_ROUTER ) )
						{
							op_prg_mem_copy( &(qupdate->GrpmbrLinkStateAdv[i]), 
									 &AreaStructure.GrpMbrLsaList[Dest_Grp][j], 
									 sizeof(grpmbrlsa) );

/*			Added by DG, 11/8/97, Clear infValid when receiving a update LSA for that group */
							mc_table[Dest_Grp].infValid = -1;
						 }
						FoundFlag = ON;
						break;
					}
					j++; 
				}
				if(FoundFlag == OFF &&
				   ( qupdate->GrpmbrLinkStateAdv[i].LinkStateHeader.ThisAreaId == AreaId
					|| IS_AREA_BORDER_ROUTER ) )
				{
					int k=0;
					for(k=0; k<MAX_NO_LSAS; k++)  
					{
/*			commented by Jiemei */
/*						if(!(AreaStructure.GrpMbrLsaList[Dest_Grp][k].LinkStateHeader.AdvRouter.NetNum==0 || 
							AreaStructure.GrpMbrLsaList[Dest_Grp][k].LinkStateHeader.AdvRouter.NodeNum == 0) )
*/

/*			added by Jiemei */
						if( (AreaStructure.GrpMbrLsaList[Dest_Grp][k].LinkStateHeader.AdvRouter.NetNum  == 0 &&
						     AreaStructure.GrpMbrLsaList[Dest_Grp][k].LinkStateHeader.AdvRouter.NodeNum == 0) )
						{
								op_prg_mem_copy( &(qupdate->GrpmbrLinkStateAdv[i]), 
										 &AreaStructure.GrpMbrLsaList[Dest_Grp][k], 
										 sizeof(grpmbrlsa) );

/*			Added by DG, 11/8/97, Clear infValid when receiving a update LSA for that group */
								mc_table[Dest_Grp].infValid = -1;

								break;
						}
					}
				}
			}
		}
	}
	op_prg_mem_free( qupdate );
/*	mospfspfcalc = 1;*/

VL 	MPDL	
{
int i = 0;
int j = 0; 
int k = 0;

printf("Router %d.%d , its Links Information is \n", 
	RouterId.NetNum, RouterId.NodeNum );
printf("- - - - - - - - - - - - - - - - - - - - - - - - - \n");

for( i = 0; i < MAX_NO_GRPS; i++ )
{
	if( !(AreaStructure.GrpMbrLsaList[i][0].LinkStateHeader.AdvRouter.NetNum == 0 || 
		  AreaStructure.GrpMbrLsaList[i][0].LinkStateHeader.AdvRouter.NodeNum == 0) )
	{
		for( j = 0; j < MAX_NO_LSAS; j++ ) 
		{
			if( !(AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.AdvRouter.NetNum == 0 || 
				  AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.AdvRouter.NodeNum == 0) )
			{
				for( k = 0; k < NoOfLinks; k++ )
				{
					if( !(AreaStructure.GrpMbrLsaList[i][j].VertexInfo[k].VertexId.NetNum <= 0 || 
						  AreaStructure.GrpMbrLsaList[i][j].VertexInfo[k].VertexId.NodeNum <= 0) )
					{
						printf( "AreaStructure.GrpMbrLsaList[%d][%d].LinkStateHeader.AdvRouter = "
								"%d.%d\n", i, j, 
							AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.AdvRouter.NetNum,
							AreaStructure.GrpMbrLsaList[i][j].LinkStateHeader.AdvRouter.NodeNum);
						printf( "AreaStructure.GrpMbrLsaList[%d][%d].VertexInfo[%d].VertexId = "
							"%d.%d\n", i, j, k, 
							AreaStructure.GrpMbrLsaList[i][j].VertexInfo[k].VertexId.NetNum,
							AreaStructure.GrpMbrLsaList[i][j].VertexInfo[k].VertexId.NodeNum );
					}
				}
			} 
		}
	}
}
}

VL  MPDL	printf("--------- End Mospf Link State Adverisments On the Router -------- \n");
return;
}

void
OspfLsUpdatePk_pro( Packet *ReceivedQospfPkptr )
{
	int i = 0;
	int j = 0;
	int FoundFlag = OFF;
	OspfLSUpdatePk *qupdate;
	
	/*qupdate = (OspfLSUpdatePk *) op_prg_mem_alloc( sizeof(OspfLSUpdatePk) ); 
	memset( qupdate, 0, sizeof(QospfLSUpdatePk) );      commented by Jiemei */

	op_pk_nfd_get( ReceivedQospfPkptr, "Lsupdate" , &qupdate );
	if( qupdate->OspfPkHeader.Type != OSPFLSUPDATEPK )
	{
		op_prg_mem_free( qupdate );
		return;
	}
	else 
	{
		PROBE_ON
		{
			NoOfOspfLsaRecv = NoOfOspfLsaRecv + qupdate->NoOfRouterAdvs +
					  qupdate->NoOfNetworkAdvs + qupdate->NoOfSummaryAdvs;
			op_stat_local_write( CUMM_OSPF_LSA_RECV_STAT, NoOfOspfLsaRecv );
			NoOfTotalLsaRecv = NoOfTotalLsaRecv + qupdate->NoOfRouterAdvs +
                                          qupdate->NoOfNetworkAdvs + qupdate->NoOfSummaryAdvs;
			op_stat_local_write( CUMM_TOTAL_LSA_RECV_STAT, NoOfTotalLsaRecv );

/* Added by Jiemei and Dg 11/27 */
			NoOfInstOspfLsaRecv = NoOfInstOspfLsaRecv + qupdate->NoOfRouterAdvs +
                                          qupdate->NoOfNetworkAdvs + qupdate->NoOfSummaryAdvs;
			op_stat_local_write( INST_OSPF_LSA_RECV_STAT, NoOfInstOspfLsaRecv );

			if( op_sim_time() > OspfLsa_Recv_sim_time + OspfLsaRefreshTime )
			{
        			NoOfInstOspfLsaRecv = 0;
       		 		OspfLsa_Recv_sim_time = op_sim_time();
 			}
		}

		for( i = 0; i < qupdate->NoOfRouterAdvs; i++ )
		{
			FoundFlag = OFF;
			for(j = 0; j < MAX_NO_LSAS; j++ )
			{
				if( (AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NetNum ==
					 qupdate->RouterLinkStateAdv[i].LinkStateHeader.AdvRouter.NetNum &&
					 AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NodeNum ==
					 qupdate->RouterLinkStateAdv[i].LinkStateHeader.AdvRouter.NodeNum) )

				{
					if( IS_AREA_BORDER_ROUTER == 1 ||
					qupdate->RouterLinkStateAdv[i].LinkStateHeader.ThisAreaId == AreaId )
					{
						op_prg_mem_copy( &(qupdate->RouterLinkStateAdv[i]), 
								 &AreaStructure.RouterLsaList[j], sizeof(RouterLinkAdv) );
						FoundFlag = ON;
					}
					break;
				}

			}
			if( FoundFlag == OFF )
			{
				for(j = 0; j < MAX_NO_LSAS; j++ )
				{
					if( (AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NetNum == 0 &&
						 AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NodeNum == 0 &&
						 AreaStructure.RouterLsaList[j].LinkStateHeader.LSSeqNo ==  0) )
					{
						if( IS_AREA_BORDER_ROUTER == 1 ||
							qupdate->RouterLinkStateAdv[i].LinkStateHeader.ThisAreaId == AreaId )
							op_prg_mem_copy( &(qupdate->RouterLinkStateAdv[i]), 
											 &AreaStructure.RouterLsaList[j], sizeof(RouterLinkAdv) );
						break;
					}
				}
			}
		}
		
		if( IS_DESIGNATED_ROUTER == 1 )
		{
			for( i = 0; i < qupdate->NoOfNetworkAdvs; i++ )
			{
				FoundFlag = OFF;
				for(j = 0; j < MAX_NO_LSAS; j++ )
				{
					if( (AreaStructure.NetworkLsaList[j].LinkStateHeader.AdvRouter.NetNum ==
						 qupdate->NetworkLinkStateAdv[i].LinkStateHeader.AdvRouter.NetNum &&
						 AreaStructure.NetworkLsaList[j].LinkStateHeader.AdvRouter.NodeNum ==
						 qupdate->NetworkLinkStateAdv[i].LinkStateHeader.AdvRouter.NodeNum) )
					{
						if( IS_AREA_BORDER_ROUTER == 1 ||
							qupdate->NetworkLinkStateAdv[i].LinkStateHeader.ThisAreaId == AreaId )
						{
							op_prg_mem_copy( &(qupdate->NetworkLinkStateAdv[i]), 
											 &AreaStructure.NetworkLsaList[j], sizeof(NetworkLinkAdv) );
							FoundFlag = ON;
						}
						break;
					}

				}
				if( FoundFlag == OFF )
				{
					for(j = 0; j < MAX_NO_LSAS; j++ )
					{
						if( (AreaStructure.NetworkLsaList[j].LinkStateHeader.AdvRouter.NetNum == 0 &&
						 AreaStructure.NetworkLsaList[j].LinkStateHeader.AdvRouter.NodeNum == 0 &&
						 AreaStructure.NetworkLsaList[j].LinkStateHeader.LSSeqNo == 0) )
						{
							if( IS_AREA_BORDER_ROUTER == 1 ||
							  qupdate->NetworkLinkStateAdv[i].LinkStateHeader.ThisAreaId == AreaId )
								op_prg_mem_copy( &(qupdate->NetworkLinkStateAdv[i]), 
												 &AreaStructure.NetworkLsaList[j], sizeof(NetworkLinkAdv) );
							break;
						}
						   
					}
				}
			}
			if( IS_AREA_BORDER_ROUTER == 1 )
			{
				for( i = 0; i < qupdate->NoOfSummaryAdvs; i++ )
				{
					FoundFlag = OFF;
					for(j = 0; j < MAX_NO_LSAS; j++ )
					{
						if( (AreaStructure.SummaryLsaList[j].LinkStateHeader.AdvRouter.NetNum ==
				         	  qupdate->SummaryLinkStateAdv[i].LinkStateHeader.AdvRouter.NetNum &&
						  AreaStructure.SummaryLsaList[j].LinkStateHeader.AdvRouter.NodeNum ==
						  qupdate->SummaryLinkStateAdv[i].LinkStateHeader.AdvRouter.NodeNum) )
						{
							if( IS_AREA_BORDER_ROUTER == 1 ||
						       	  qupdate->SummaryLinkStateAdv[i].LinkStateHeader.ThisAreaId == AreaId )
							{
								op_prg_mem_copy( &(qupdate->SummaryLinkStateAdv[i]), 
										 &AreaStructure.SummaryLsaList[j], sizeof(SummaryLinkAdv) );
								FoundFlag = ON;
							}
							break;
						}

					}
					if( FoundFlag == OFF )
					{
						for(j = 0; j < MAX_NO_LSAS; j++ )
						{
							if( (AreaStructure.SummaryLsaList[j].LinkStateHeader.AdvRouter.NetNum == 0 &&
								 AreaStructure.SummaryLsaList[j].LinkStateHeader.AdvRouter.NodeNum == 0 &&
								 AreaStructure.SummaryLsaList[j].LinkStateHeader.LSSeqNo == 0) )
							{
								if( IS_AREA_BORDER_ROUTER == 1 ||
									qupdate->SummaryLinkStateAdv[i].LinkStateHeader.ThisAreaId == AreaId )
									op_prg_mem_copy( &(qupdate->SummaryLinkStateAdv[i]), 
													 &AreaStructure.SummaryLsaList[j], 
													 sizeof(SummaryLinkAdv) );
								break;
							}
							   
						}
					}
				}
			}
		}
	}
	op_prg_mem_free( qupdate );
	ospfspfcalc = 1;
	VL	OPDL 
	{
		printf("\n---------------------- (%d.%d) OSPF LSUPDATE PACKET ----------------------\n",
				RouterId.NetNum, RouterId.NodeNum );
		printf("ROUTER LSA'S : \n");
		for( i = 0; i < MAX_NO_LSAS; i++ )
		{
			if( !(AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 || 
				  AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 ) )
			{
				printf("Adv. Router is %d.%d and its Links Information is \n", 
				AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum,
				AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NodeNum );
				printf("- - - - - - - - - - - - - - - - - - - - - - - - - \n");
				printf("No Of Links are %d\n", AreaStructure.RouterLsaList[i].No_of_Links);

				for( j = 0; j < AreaStructure.RouterLsaList[i].No_of_Links; j++ )
				{
					printf( "Type of the Link is %d, ", 
							 AreaStructure.RouterLsaList[i].Link_Info[j].Type );
					printf( "Link Id is %d.%d, ", 
							 AreaStructure.RouterLsaList[i].Link_Info[j].LinkId.NetNum,
							 AreaStructure.RouterLsaList[i].Link_Info[j].LinkId.NodeNum );
					printf( "Link Data is %d.%d, ", 
							 AreaStructure.RouterLsaList[i].Link_Info[j].LinkData.NetNum,
							 AreaStructure.RouterLsaList[i].Link_Info[j].LinkData.NodeNum );
					printf( "TOS Metric is %d\n", 
							 AreaStructure.RouterLsaList[i].Link_Info[j].TOSMetric );
				}
				printf("\n");
			}
		}
		printf("\n");
		printf( "NETWORK LSA'S : \n" );
		for( i = 0; i < MAX_NO_LSAS; i++ )
		{
			if( !(AreaStructure.NetworkLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 || 
				  AreaStructure.NetworkLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 ) )
			{
				printf(" Adv. Router is %d.%d\n",
				AreaStructure.NetworkLsaList[i].LinkStateHeader.AdvRouter.NetNum,
				AreaStructure.NetworkLsaList[i].LinkStateHeader.AdvRouter.NodeNum );
				printf("- - - - - - - - - - - - - - - - - - - - - - - - - \n");
				for( j = 0; j < MAX_NO_LINKS; j++ )
				{
					if (!(AreaStructure.NetworkLsaList[i].AttachedRouter[j].NetNum==0 ||
					      AreaStructure.NetworkLsaList[i].AttachedRouter[j].NodeNum==0))
					{
						printf("AttachedRouter[%d]=%d.%d\n",j,
						   AreaStructure.NetworkLsaList[i].AttachedRouter[j].NetNum,
						   AreaStructure.NetworkLsaList[i].AttachedRouter[j].NodeNum);
						
					}
				}
			}
		}
		printf("\n");
		printf( "SUMMARY LSA'S : \n" );
		for( i = 0; i < MAX_NO_LSAS; i++ )
		{
			if( !(AreaStructure.SummaryLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 || 
				  AreaStructure.SummaryLsaList[i].LinkStateHeader.AdvRouter.NetNum == 0 ) )
			printf(" Adv. Router is %d.%d \n",
			AreaStructure.SummaryLsaList[i].LinkStateHeader.AdvRouter.NetNum,
			AreaStructure.SummaryLsaList[i].LinkStateHeader.AdvRouter.NodeNum );
		}
		printf("\n");
		printf("-------------------- (%d.%d)END OSPF LSUPDATE PACKET --------------------\n",
				RouterId.NetNum, RouterId.NodeNum );
	}
	return;
}

/**
*** Mospf Routing Table calculating functions
**/
void
CandListInt(int s_net, ListL* CandidatesList)
{
    int SrcNet;
    Vertex_Struct CandVertex;
 
    int SrcNet_Found_Flag = OFF;
 
    memset( &CandVertex, 0, sizeof(Vertex_Struct) );
 
    SrcNet = s_net;
    if( (SrcNet <= (AreaId*MaxNoNetsInArea) && SrcNet >= (AreaId * MaxNoNetsInArea - MaxNoNetsInArea )) ||
        (SrcNet <= (Router_Start_Index + AreaId*MaxNoRoutersInArea) && 
         SrcNet >= (Router_Start_Index + AreaId*MaxNoRoutersInArea - MaxNoNetsInArea) ) )
/* Add last two lower conditions by DG , 10/31 */

        SrcNet_Found_Flag = ON;
/* SourceNet is in this area, DG 10/17 ??? */
 
    if(SrcNet_Found_Flag == ON)
    {
        CandVertex.VertexId.NetNum = SrcNet;
        CandVertex.DistanceToRoot = 0;
        AddListEle( CandidatesList, (char *)&CandVertex, sizeof(Vertex_Struct) );
    }
    else
    {
        CandVertex.VertexId = AreaBorderRouters[AreaId];
        AddListEle( CandidatesList, (char *)&CandVertex, sizeof(Vertex_Struct) );
    }
}

void
MospfCandAddPro( int GrpId, Vertex_Struct* VertexV, ListL* MospfCandidatesList )
{
}

void
/* MospfSPFTreeCalc(int GrpId,  ListL *CandidateList ,Area_Struct *tmp_AreaStructure) */
MospfSPFTreeCalc(int GrpId,  ListL *CandidateList)
{
    int Inx, i;
    Vertex_Struct VertexW;

VL MCSPFDL printf("***********MospfSPFTreeCalc  Begin ************\n");

    if( !IsListEmpty( CandidateList ) )
    {
        int LeastDistance = 10000;
        int IsLessThanFlag = OFF;
        int FoundFlag = OFF;
        Vertex_Struct* tmpCandidateEle;
        
	tmpCandidateEle = (Vertex_Struct*)GetNextListEle( CandidateList, 1 );
        while( tmpCandidateEle )
        {
/* Add two lines, DG, 10/23 */
        IsLessThanFlag = OFF;
        FoundFlag = OFF;

VL MCSPFDL printf("VertexId=%d.%d DistanceToRoot=%d\n",tmpCandidateEle->VertexId.NetNum,
		  tmpCandidateEle->VertexId.NodeNum,tmpCandidateEle->DistanceToRoot);

            LeastDistance = tmpCandidateEle->DistanceToRoot;
            Inx = 0;

	    /* added by Jiemei */
	    FoundFlag=OFF;
	    IsLessThanFlag=OFF;
            while( AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum != 0 
			&& (Inx<MAX_NO_SPROUTERS) ) /* Inx<MAX_NO_SPROUTERS is added by Jiemei*/ 
	    {
VL MCSPFDL printf("AreaStructure.MospfShortestPathTree[%d].SPVertex[%d].VertexId=%d.%d\n",
        	  GrpId,Inx,AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum,
        	  AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NodeNum);
                	if( AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum ==
                    	tmpCandidateEle->VertexId.NetNum)
                	{
                	    FoundFlag = ON;
                	    break;
                	}
 
                	if( AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].DistanceToRoot
                	    > LeastDistance )
                	{
                	    IsLessThanFlag = ON;
                	    break;
                	}
                Inx++; 
            }
            if( FoundFlag == ON )
                if( AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].DistanceToRoot
                    > LeastDistance )
                    AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].DistanceToRoot
                                                                                    = LeastDistance;
 
            if( FoundFlag == OFF )
            {
                if(IsLessThanFlag == ON)
                {
                    Vertex_Struct tmpSPVertex;
                    Vertex_Struct sectmpSPVertex;
    
	            tmpSPVertex = AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx];
/*		op_prg_mem_copy(&AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx],&tmpSPVertex,
				    sizeof(Vertex_Struct)); changed by Jiemei*/
                    op_prg_mem_copy( tmpCandidateEle,
                                     &AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx],
                                     sizeof(Vertex_Struct) );

                    while( (AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum !=0)
		        && (Inx<MAX_NO_SPROUTERS) ) /* Inx<MAX_NO_SPROUTERS is added by Jiemei*/ 
                   {
			
                        sectmpSPVertex = AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i+1];
                        AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i+1] = tmpSPVertex;
                        tmpSPVertex = sectmpSPVertex; 
                        Inx++;
                    }
                }
                if( IsLessThanFlag == OFF )
                    op_prg_mem_copy( tmpCandidateEle,
                                     &AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx],
                                     sizeof(Vertex_Struct) );
                op_prg_mem_copy( tmpCandidateEle, &VertexW, sizeof(Vertex_Struct) );
                MospfCandidateAddPro( GrpId, &VertexW , CandidateList );
            }

VL MCSPFDL 
{
	printf("-----------------------------------------------------------------------\n");
	for (Inx=0; Inx < MAX_NO_SPROUTERS; Inx++)
	{
		if(AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum!=0)
		{ 
			printf("AreaStructure.MospfShortestPathTree[%d].SPVertex[%d].VertexId=%d.%d\n",GrpId,Inx,
				AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum, 
				AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NodeNum);
			printf("                                                    .DistanceToRoot=%d\n",
				AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].DistanceToRoot);
			for (i=0; i< MAX_NO_NEIGHBORS; i++)
			{
			        if (AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].NextHopsList[i].NetNum!=0) 
				{
					printf("                                                    .NextHopsList[%d]=%d.%d\n",i,
					AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].NextHopsList[i].NetNum,
					AreaStructure.MospfShortestPathTree[GrpId].SPVertex[Inx].NextHopsList[i].NodeNum);
				}
			}
			printf("\n");
		}
	}
} 
            DeleteListEle(CandidateList, (char*)tmpCandidateEle, sizeof(Vertex_Struct) );
            tmpCandidateEle = (Vertex_Struct*)GetNextListEle( CandidateList, 1 );
	}
    }
VL MCSPFDL printf("***********MospfSPFTreeCalc  End ************\n");
}


void
/* MCRoutetableCalc(int GrpId, int NoLinks, Area_Struct *tmp_AreaStructure, int *rte_table, mc_id *mc_table) */
MMCRoutetableCalc(int GrpId, int s_net)	/* Last parameter added by DG, 11/8, Source of the tree. */
{
    int i = 0;   /* changed by Jiemei from 1 to 0 */
    int j;
    int RouteFoundFlag;
    int *rte_table;
    mc_id *mc_table;


    rte_table = (int *)op_ima_obj_svar_get(ip_objid, "rte_tab" );
    mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");

VL MCSPFDL printf("***********Mospf McRoutetableCalc  Begin ************\n");
 
    while(AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum != 0) 
    {

VL MCSPFDL printf("AreaStructure.MospfShortestPathTree[%d].SPVertex[%d].VertexId=%d.%d\n",
        GrpId,i,AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum,
        AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NodeNum);

	j=0;
	RouteFoundFlag= 0;
	while( AreaStructure.GrpMbrLsaList[GrpId][j].LinkStateHeader.AdvRouter.NetNum != 0 &&
		AreaStructure.GrpMbrLsaList[GrpId][j].LinkStateHeader.AdvRouter.NodeNum != 0 )
	{
		if ( AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum ==
			AreaStructure.GrpMbrLsaList[GrpId][j].LinkStateHeader.AdvRouter.NetNum )

		{
VL MCSPFDL printf("rte_table[%d]=%d\n",
	AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum,
	rte_table[AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum]);
			

mc_table[GrpId].interface[rte_table[AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum]] = 1;

VL MCSPFDL printf("mc_table[%d].interface[rte_table]=%d\n",GrpId, 1 );

			RouteFoundFlag = 1;
			break;
		}
		j++;
	} 

/*
        int j = 0;
        int k = 0;
        int RouteFoundFlag = OFF;
 
VL MCSPFDL printf("AreaStructure.MospfShortestPathTree[%d].SPVertex[%d].VertexId=%d.%d\n",
        GrpId,i,AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum,
        AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NodeNum);
       
        for( j = 0; j < NoLinks; j++ )
        {
VL MCSPFDL printf("rte_table[%d]=%d\n",
	AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum,
	rte_table[AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum]);

	    mc_table[GrpId].interface[rte_table[AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum]] = 1;
	    
VL MCSPFDL printf("mc_table[%d].interface[rte_table]=%d\n",GrpId,
		mc_table[GrpId].interface[rte_table[AreaStructure.MospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum]]);

            RouteFoundFlag = ON;
        }
*/
        if( RouteFoundFlag == OFF )
        {
            if( !IS_AREA_BORDER_ROUTER )
                /*mc_rte_table[GrpId].interface[rte_table[DEF_STREAM]] = 1; commented by Jiemei*/
                mc_table[GrpId].interface[rte_table[DEF_STREAM]] = 1;

VL MCSPFDL printf("Default: mc_table[%d].interface[%d]=%d\n",GrpId,rte_table[DEF_STREAM],mc_table[GrpId].interface[rte_table[DEF_STREAM]]);
        }

       i++;
    }


/* Added by DG, 11/8/97, set infValid of the group to the NetNum of source. */
    mc_table[GrpId].infValid = s_net;


VL MCSPFDL
{
	int i,j;

	printf("---------------(%d.%d) mc_table-------------\n",net_num0,node_num0);
	for (i=0; i< MAX_NO_GRPS; i++)
	{
		for (j=0; j< MAX_NO_LINKS; j++)
		{
			if (mc_table[i].interface[j] > 0 )
				printf("Group[%d].interface[%d]=%d\n",i,j,mc_table[i].interface[j]); 
		}
	}	
}
VL printf("*************** Mospf McRoutetableCalc End ****************\n");
}


/**
*** Add by DG, 11/22, Just copy MMCRoutetableCalc, but using QospfShortestPathTree
**/
void
QMCRoutetableCalc(int GrpId, int s_net)	
{
    int i = 0;   
    int j;
    int RouteFoundFlag;
    int *rte_table;
    mc_id *mc_table;


    rte_table = (int *)op_ima_obj_svar_get(ip_objid, "rte_tab" );
    mc_table = (mc_id *) op_ima_obj_svar_get(ip_objid,"mc_table");

VL printf("*********** Qospf McRoutetableCalc  Begin ************\n");
 
    while(AreaStructure.QospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum != 0) 
    {

VL QSPFDL  printf("AreaStructure.QospfShortestPathTree[%d].SPVertex[%d].VertexId=%d.%d\n",
        GrpId,i,AreaStructure.QospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum,
        AreaStructure.QospfShortestPathTree[GrpId].SPVertex[i].VertexId.NodeNum);

	j=0;
	RouteFoundFlag= 0;
	while( AreaStructure.GrpMbrLsaList[GrpId][j].LinkStateHeader.AdvRouter.NetNum != 0 &&
		AreaStructure.GrpMbrLsaList[GrpId][j].LinkStateHeader.AdvRouter.NodeNum != 0 )
	{
		if ( AreaStructure.QospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum ==
			AreaStructure.GrpMbrLsaList[GrpId][j].LinkStateHeader.AdvRouter.NetNum )

		{

			mc_table[GrpId].interface[rte_table[AreaStructure.QospfShortestPathTree[GrpId].SPVertex[i].VertexId.NetNum]] = 1;

VL QSPFDL  printf("mc_table[%d].interface[rte_table]=%d\n",GrpId, 1 );

			RouteFoundFlag = 1;
			break;
		}
		j++;
	} 

        if( RouteFoundFlag == OFF )
        {
            if( !IS_AREA_BORDER_ROUTER )
                /*mc_rte_table[GrpId].interface[rte_table[DEF_STREAM]] = 1; commented by Jiemei*/
                mc_table[GrpId].interface[rte_table[DEF_STREAM]] = 1;

VL QSPFDL printf("Default: mc_table[%d].interface[%d]=%d\n",
		GrpId,rte_table[DEF_STREAM],mc_table[GrpId].interface[rte_table[DEF_STREAM]]);
        }

       i++;
    }



VL 
{
	int i,j;

	printf("---------------(%d.%d) mc_table-------------\n",net_num0,node_num0);
	for (i=0; i< MAX_NO_GRPS; i++)
	{
		for (j=0; j< MAX_NO_LINKS; j++)
		{
			if (mc_table[i].interface[j] > 0 )
				printf("Group[%d].interface[%d]=%d\n",i,j,mc_table[i].interface[j]); 
		}
	}	
}
VL printf("*************** Qospf McRoutetableCalc End ****************\n");
}


void
QospfCandidateListInit(int SrcNet, ListL* QospfCandidatesList)
{
    Vertex_Struct CandVertex;
 
    int SrcNet_Found_Flag = OFF;
 
    memset( &CandVertex, 0, sizeof(Vertex_Struct) );
 
    if( SrcNet <= (AreaId*MaxNoNetsInArea) ||
        SrcNet <= (Router_Start_Index + AreaId*MaxNoRoutersInArea) )
        SrcNet_Found_Flag = ON;
 
    if(SrcNet_Found_Flag == ON)
    {
        CandVertex.VertexId.NetNum = SrcNet;
        CandVertex.DistanceToRoot = 0;
        AddListEle( QospfCandidatesList, (char *)&CandVertex, sizeof(Vertex_Struct) );
    }
    else
    {
        CandVertex.VertexId = AreaBorderRouters[AreaId];
        AddListEle( QospfCandidatesList, (char *)&CandVertex, sizeof(Vertex_Struct) );
    }
}
 
void
QospfCandAddPro( int GrpId, Vertex_Struct* VertexV, ListL* QospfCandidatesList )
{
    int i=0;
    int k=0;
 
    Vertex_Struct VertexW;
    Vertex_Struct tmpVertexW;
 
    reslsa  WResLsa;
    rralsa  WRraLsa;
    reslsa  VResLsa;
    rralsa  VRraLsa;
 
    Vertex_Struct* CandidateEle;
    Vertex_Struct* tmpCandEle;
    int CandidateFoundFlag = OFF;
    int Rra_Found_Flag = OFF;
 
    memset( &VertexW, 0, sizeof(Vertex_Struct) );
    memset( &tmpVertexW, 0, sizeof(Vertex_Struct) );
    memset( &WResLsa, 0, sizeof(reslsa) );
    memset( &VResLsa, 0, sizeof(reslsa) );
    memset( &WRraLsa, 0, sizeof(rralsa) );
    memset( &VRraLsa, 0, sizeof(rralsa) );
 
VL printf( " ************ QospfCandAddPro Begin VertexV(%d.%d) *************\n",
	VertexV->VertexId.NetNum, VertexV->VertexId.NodeNum );
    for(i = 0; i<MAX_NO_LSAS; i++ ) /* found by Sitaram */
    {
        if( AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NetNum == VertexV->VertexId.NetNum )
/*	Commented by DG, 11/25/97, VertexV.NodeNum is 0, doesn't use that.
          &&  AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NodeNum == VertexV->VertexId.NodeNum  )
*/
        {
            op_prg_mem_copy( &AreaStructure.ResourceLsaList[i], &VResLsa, sizeof(reslsa) );
            break;
        }
    }
    for(i = 0; i<MAX_NO_RRA_LSAS; i++ ) /* found by Sitaram */
    {
        if( AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NetNum == VertexV->VertexId.NetNum &&
/*	Commented by DG, 11/25/97, VertexV.NodeNum is 0, doesn't use that.
            AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NodeNum == VertexV->VertexId.NodeNum  &&
*/
            AreaStructure.RRALsaList[i].DestAddr.NodeNum == GrpId )
        {
            op_prg_mem_copy( &AreaStructure.RRALsaList[i], &VRraLsa, sizeof(reslsa) );
            Rra_Found_Flag = ON;
            break;
        }
    }

    for(i = 0; i < VResLsa.ResLsaHeader.No_of_Links; i++ ) /* found by sitaram */
    {
        int Link_Eligible_Flag = OFF;
 
 
        if( Rra_Found_Flag == ON )
        {
            if( (VRraLsa.LinkResvInfo[i].ReservedBandwidth >= SessionDirectory[GrpId].Bandwidth) ||
                (VResLsa.ResLinkInfo[i].Bandwidth >= (SessionDirectory[GrpId].Bandwidth -
                VRraLsa.LinkResvInfo[i].ReservedBandwidth)) )
 
                Link_Eligible_Flag = ON;
        }
        else
        {
            if( VResLsa.ResLinkInfo[i].Bandwidth >= SessionDirectory[GrpId].Bandwidth )
            {
                Link_Eligible_Flag = ON;
            }
        }

        if( Link_Eligible_Flag == ON)
        {
            int FoundFlag = OFF;
            int j = 0;
            op_prg_mem_copy( &VResLsa.ResLinkInfo[i].LinkId, &VertexW.VertexId, sizeof(IPAddress) );

            j = 0;
            while( !(AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NodeNum == 0 || 
                     AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NetNum == 0 ) )
            {
                if( AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NetNum
                    == VertexW.VertexId.NetNum &&
                    AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NodeNum
                    == VertexW.VertexId.NodeNum )
                {
                    FoundFlag = ON;
                    op_prg_mem_copy( &AreaStructure.ResourceLsaList[j], &WResLsa, sizeof(reslsa) );
                    break;
                }
                j++;
            }
            if( FoundFlag = ON )
            {
                int WonSPF_Flag = OFF;
                j = 0;
 
                for( k = 0; k < WResLsa.ResLsaHeader.No_of_Links; k++ )
                {
/* Added by DG, 12/1/97 */
		    j=0;


/* Add the first condition by DG, 12/1/97 */
                   while( j < MAX_NO_SPROUTERS && 
			 !(AreaStructure.QospfShortestPathTree[GrpId].SPVertex[j].VertexId.NetNum == 0 || 
                            AreaStructure.QospfShortestPathTree[GrpId].SPVertex[j].VertexId.NodeNum == 0 ) )
                   {
                        if( AreaStructure.QospfShortestPathTree[GrpId].SPVertex[j].VertexId.NetNum
                            == WResLsa.ResLinkInfo[k].LinkData.NetNum &&
 
                            AreaStructure.QospfShortestPathTree[GrpId].SPVertex[j].VertexId.NodeNum
                            == WResLsa.ResLinkInfo[k].LinkData.NodeNum )
                        {
                            WonSPF_Flag = ON;
                            break;
                        }
                        j++;
                    }
                    if( WonSPF_Flag == OFF && 
			!(WResLsa.ResLinkInfo[k].LinkData.NetNum == 0 || 
                          WResLsa.ResLinkInfo[k].LinkData.NodeNum == 0 ) )
                    {
                        op_prg_mem_copy(&WResLsa.ResLinkInfo[k].LinkData, &tmpVertexW.VertexId,
                                        sizeof(IPAddress) );
                        if( VResLsa.ResLinkInfo[i].TOSMetric != LSInfinity )
                        {
                            tmpVertexW.DistanceToRoot =
                                        VertexV->DistanceToRoot + VResLsa.ResLinkInfo[i].TOSMetric;
                        }
 
			CandidateFoundFlag = OFF;	/* Added by DG, 12/1/97 */
                        if( !IsListEmpty(QospfCandidatesList) )
                        {
                            CandidateEle = (Vertex_Struct*)GetNextListEle( QospfCandidatesList, 1 );
                            while( CandidateEle )
                            {
                                if( CandidateEle->VertexId.NetNum == tmpVertexW.VertexId.NetNum &&
                                    CandidateEle->VertexId.NodeNum == tmpVertexW.VertexId.NodeNum )
                                {
                                    CandidateFoundFlag = ON;
                                    break;
                                }
				else
                                    CandidateEle = (Vertex_Struct*)GetNextListEle( QospfCandidatesList, 0 );
                            }
                        }

                        if( CandidateFoundFlag == ON )
                        {
                            if( CandidateEle->DistanceToRoot > tmpVertexW.DistanceToRoot )
                            {
                                CandidateEle->DistanceToRoot = tmpVertexW.DistanceToRoot;
                            }
                        }
                        if( CandidateFoundFlag == OFF )
                        {
                            AddListEle(QospfCandidatesList, (char *)&tmpVertexW, sizeof(Vertex_Struct) );
                        }
                    }
                }
            }
 
        }
    }
}

/**
*** Rewriten by Dg, 12/3, according to OspfCandidateAddPro.
**/

void
QospfCandidateAddPro( int GrpId, Vertex_Struct* VertexV, ListL* QospfCandidatesList )
{
    int i=0;
    int k=0;
 
    Vertex_Struct VertexW;
 
    reslsa  WResLsa;
    rralsa  WRraLsa;
    reslsa  VResLsa;
    rralsa  VRraLsa;
 
    Vertex_Struct* CandidateEle;
    Vertex_Struct* tmpCandEle;
    int CandidateFoundFlag = OFF;
    int Rra_Found_Flag = OFF;
 
    memset( &VertexW, 0, sizeof(Vertex_Struct) );
    memset( &WResLsa, 0, sizeof(reslsa) );
    memset( &VResLsa, 0, sizeof(reslsa) );
    memset( &WRraLsa, 0, sizeof(rralsa) );
    memset( &VRraLsa, 0, sizeof(rralsa) );
 
VL printf( " ************ QospfCandAddPro Begin VertexV(%d.%d) T=%fs*************\n",
	VertexV->VertexId.NetNum, VertexV->VertexId.NodeNum, op_sim_time() );
    for(i = 0; i<MAX_NO_LSAS; i++ ) /* found by Sitaram */
    {
        if( AreaStructure.ResourceLsaList[i].ResLsaHeader.AdvRouter.NetNum == VertexV->VertexId.NetNum )
        {
            op_prg_mem_copy( &AreaStructure.ResourceLsaList[i], &VResLsa, sizeof(reslsa) );
            break;
        }
    }
VL
{
/*	if ( op_sim_time() > 35 ) */
	{
		printf( "VertexV = (%d.%d) Group=%d\n",
			VertexV->VertexId.NetNum,VertexV->VertexId.NodeNum, GrpId );
/*		getchar(); */
	}
}
    for(i = 0; i<MAX_NO_RRA_LSAS; i++ ) /* found by Sitaram */
    {
        if( AreaStructure.RRALsaList[i].OpaqueLsaHeader.AdvRouter.NetNum == VertexV->VertexId.NetNum &&
            AreaStructure.RRALsaList[i].DestAddr.NodeNum == GrpId )
        {
            op_prg_mem_copy( &AreaStructure.RRALsaList[i], &VRraLsa, sizeof(reslsa) );
            Rra_Found_Flag = ON;
            break;
        }
    }
VL
{
/*	if ( op_sim_time() > 35 ) */
	{
		printf( "Rra_Found_Flag=%d\n", Rra_Found_Flag );
/*		getchar(); */
	}
}

    for(i = 0; i < VResLsa.ResLsaHeader.No_of_Links; i++ ) /* found by sitaram */
    {
        int Link_Eligible_Flag = OFF;

/* Add condition by DG, 12/3 follow OSPF */
 	if ( VResLsa.ResLinkInfo[i].Type == ROUTER || VResLsa.ResLinkInfo[i].Type == TRANSITNET )
	{
        	if( Rra_Found_Flag == ON )
        	{
            		if( (VRraLsa.LinkResvInfo[i].ReservedBandwidth >= SessionDirectory[GrpId].Bandwidth) ||
                		(VResLsa.ResLinkInfo[i].Bandwidth >= 
				   (SessionDirectory[GrpId].Bandwidth - VRraLsa.LinkResvInfo[i].ReservedBandwidth)) )
 
                		Link_Eligible_Flag = ON;
        	}
        	else
        	{
            		if( VResLsa.ResLinkInfo[i].Bandwidth >= SessionDirectory[GrpId].Bandwidth )
				Link_Eligible_Flag = ON;
		}
	}
VL
{
/*	if ( op_sim_time() > 35 ) */
	{
		printf( "Link_Eligible_Flag=%d\n", Link_Eligible_Flag );
/*		getchar(); */
	}
}

        if( Link_Eligible_Flag == ON)
        {
            int FoundFlag = OFF;
            int j = 0;
            op_prg_mem_copy( &VResLsa.ResLinkInfo[i].LinkId, &VertexW.VertexId, sizeof(IPAddress) );
VL
{
/*	if ( op_sim_time() > 35 ) */
	{
		printf( "VertexW=(%d.%d)\n", VertexW.VertexId.NetNum, VertexW.VertexId.NodeNum );
	}
}

            j = 0;
            while( AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NetNum != 0 )
            {
/* Changed by DG, 12/3, no node conditon */
                if( AreaStructure.ResourceLsaList[j].ResLsaHeader.AdvRouter.NetNum
                    == VertexW.VertexId.NetNum )
                {
                    FoundFlag = ON;
                    op_prg_mem_copy( &AreaStructure.ResourceLsaList[j], &WResLsa, sizeof(reslsa) );
                    break;
                }
                j++;
            }
VL
{
/*	if ( op_sim_time() > 35 ) */
	{
		printf( "FoundFlag=%d\n", FoundFlag );
/*		getchar(); */
	}
}
            if( FoundFlag = ON )
            {
                int WonSPF_Flag = OFF;
                j = 0;
 
VL
{
printf( "WResLsa.ResLsaHeader.link = %d begin\n", k );
}

/* Add the first condition by DG, 12/1/97 */
                while( j < MAX_NO_SPROUTERS && 
			AreaStructure.QospfShortestPathTree[GrpId].SPVertex[j].VertexId.NetNum != 0 ) 
		{
                        if( AreaStructure.QospfShortestPathTree[GrpId].SPVertex[j].VertexId.NetNum
                            == VertexW.VertexId.NetNum &&
                            AreaStructure.QospfShortestPathTree[GrpId].SPVertex[j].VertexId.NodeNum
                            == VertexW.VertexId.NodeNum )
                        {
                            WonSPF_Flag = ON;
                            break;
                        }
                        j++;
                }
VL
{
/*	if ( op_sim_time() > 35 ) */
	{
		printf( "WonSPF_Flag=%d\n", WonSPF_Flag );
/*		getchar(); */
	}
}
                    if( WonSPF_Flag == OFF && 
			WResLsa.ResLinkInfo[k].LinkData.NetNum != 0 )
                    {
                        if( VResLsa.ResLinkInfo[i].TOSMetric != LSInfinity )
                        {
                            VertexW.DistanceToRoot =
                                        VertexV->DistanceToRoot + VResLsa.ResLinkInfo[i].TOSMetric;
                        }
 
			CandidateFoundFlag = OFF;	/* Added by DG, 12/1/97 */
                        if( !IsListEmpty(QospfCandidatesList) )
                        {
                            CandidateEle = (Vertex_Struct*)GetNextListEle( QospfCandidatesList, 1 );
                            while( CandidateEle )
                            {
                                if( CandidateEle->VertexId.NetNum == VertexW.VertexId.NetNum )
                                {
                                    CandidateFoundFlag = ON;
                                    break;
                                }
				else
                                    CandidateEle = (Vertex_Struct*)GetNextListEle( QospfCandidatesList, 0 );
                            }
                        }

VL
{
/*	if ( op_sim_time() > 35 ) */
	{
		printf( "CandidateFoundFlag=%d\n", CandidateFoundFlag );
/*		getchar(); */
	}
}

                        if( CandidateFoundFlag == ON )
                        {
                            if( CandidateEle->DistanceToRoot == VertexW.DistanceToRoot )
			    {
/*				CalculateNextHops( &VertexW, VertexV, &VResLsa ); */
			    }
                            else if( CandidateEle->DistanceToRoot > VertexW.DistanceToRoot )
                            {
				CandidateEle->VertexId = VertexW.VertexId;
				CandidateEle->DistanceToRoot= VertexW.DistanceToRoot;
/*				CalculateNextHops( &VertexW, VertexV, &VResLsa ); */
                            }
                        }
                        if( CandidateFoundFlag == OFF )
                        {
				int Hop_Count = 0;
				int hopfound = OFF;
				int k;
				for( k = 0; k < VResLsa.ResLsaHeader.No_of_Links; k++ )
				{
					if( VertexW.VertexId.NetNum==VResLsa.ResLinkInfo[i].LinkId.NetNum )
					{
						op_prg_mem_copy(&VResLsa.ResLinkInfo[k].LinkData,
								&VertexW.NextHopsList[Hop_Count],
								sizeof(IPAddress));
						Hop_Count++;
						hopfound = ON;
					}
				}
				if(hopfound == OFF )
				{
					op_prg_mem_copy(&VertexV->VertexId, 
							&VertexW.NextHopsList[Hop_Count],
							sizeof(IPAddress));
				}
				AddListEle(QospfCandidatesList, (char *)&VertexW, sizeof(Vertex_Struct) );
                        }
                    }
VL
{
printf( "WResLsa.ResLsaHeader.link = %d end\n", k );
}
            }
        }
    }
}

 
void
QospfSPFTreeCalc(int GrpId,  ListL* CandidateList )
{
    int Inx, i;
    Vertex_Struct VertexW;
    if( !IsListEmpty( CandidateList ) )
    {
/* Add two link by DG, 12/3 */
        int LeastDistance = 10000;
        int IsLessThanFlag = OFF;
        int FoundFlag = OFF;
        Vertex_Struct* tmpCandidateEle;
        tmpCandidateEle = (Vertex_Struct*)GetNextListEle( CandidateList, 1 );
        while( tmpCandidateEle )
        {
       	    IsLessThanFlag = OFF;
            FoundFlag = OFF;
            LeastDistance = tmpCandidateEle->DistanceToRoot;
            Inx = 0;
/* Changed by DG, 12/3 */
            while( AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum != 0 && 
			Inx < MAX_NO_SPROUTERS )
            {
                if( AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum ==
                    tmpCandidateEle->VertexId.NetNum )
                {
                    FoundFlag = ON;
                    break;
                }
 
                if( AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].DistanceToRoot
                    > LeastDistance )
                {
                    IsLessThanFlag = ON;
                    break;
                }
                Inx++;
            }
            if( FoundFlag == ON )
                if( AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].DistanceToRoot
                    > LeastDistance )
                    AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].DistanceToRoot
                                                                                    = LeastDistance;
            if( FoundFlag == OFF )
            {
                if(IsLessThanFlag == ON)
                {
                    Vertex_Struct tmpSPVertex;
                    Vertex_Struct sectmpSPVertex;
                    tmpSPVertex = AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx];
                    op_prg_mem_copy( tmpCandidateEle,
                                     &(AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx]),
                                     sizeof(Vertex_Struct) );
/* Changed by DG , 12/3 */
                    while(AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum != 0 && 
                          Inx < MAX_NO_SPROUTERS )
                   {
                        sectmpSPVertex = AreaStructure.QospfShortestPathTree[GrpId].SPVertex[i+1];
                        AreaStructure.QospfShortestPathTree[GrpId].SPVertex[i+1] = tmpSPVertex;
                        tmpSPVertex = sectmpSPVertex;
                        Inx++;
                    }
                }
                if( IsLessThanFlag == OFF )
                    op_prg_mem_copy( tmpCandidateEle,
                                     &(AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx]),
                                     sizeof(Vertex_Struct) );
 
                op_prg_mem_copy( tmpCandidateEle, &VertexW, sizeof(Vertex_Struct) );
/*                OspfCandidateAddPro( &VertexW , CandidateList ); */
/* Changed by DG 11/18/97 */
		QospfCandidateAddPro( GrpId, &VertexW , CandidateList );
            }
            DeleteListEle(CandidateList, (char*)tmpCandidateEle, sizeof(Vertex_Struct) );
            tmpCandidateEle = (Vertex_Struct*)GetNextListEle( CandidateList, 1 );
        }
VL QSPFDL
{
	printf("-----------------------------------------------------------------------\n");
	for (Inx=0; Inx < MAX_NO_SPROUTERS; Inx++)
	{
		if(AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum!=0)
		{ 
			printf("AreaStructure.QospfShortestPathTree[%d].SPVertex[%d].VertexId=%d.%d\n",GrpId,Inx,
				AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NetNum, 
				AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].VertexId.NodeNum);
			printf("                                                    .DistanceToRoot=%d\n",
				AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].DistanceToRoot);
			for (i=0; i< MAX_NO_NEIGHBORS; i++)
			{
			        if (AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].NextHopsList[i].NetNum!=0) 
				{
					printf("                                                    .NextHopsList[%d]=%d.%d\n",i,
					AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].NextHopsList[i].NetNum,
					AreaStructure.QospfShortestPathTree[GrpId].SPVertex[Inx].NextHopsList[i].NodeNum);
				}
			}
			printf("\n");
		}
	}
} 

    }
}

/**
*** Function : QospfPkPro.
*** Input    : A pointer to Qospf packet that is obtained from the IP stream
*** Output   : This function will ckeck for the Qospf/Ospf packet type that it has recived and call the
***            appropriate functions accordingly.
	*/
void
QospfPkPro( Packet *ReceivedQospfPkptr )
{
	int tmp_msg_type;

    op_pk_nfd_get(ReceivedQospfPkptr, "Msg_type", &tmp_msg_type);
 
    switch(tmp_msg_type) 
	{
        case HELLOPK  : 
		HelloPk_pro(ReceivedQospfPkptr);
		break;
        case OSPFLSUPDATEPK  : 
		VL	OPDL	printf("(%d.%d) Received OspfLsUpdate Packet \n" ,
					RouterId.NetNum, RouterId.NodeNum );
		InstNoLsa++;
		OspfLsUpdatePk_pro(ReceivedQospfPkptr);
		break;
        case QOSPFLSUPDATEPK  : 
TRACE		VL QPDL printf("Time: %f ", op_sim_time() ); 
TRACE        	VL QPDL printf("ROUTER ID(%d.%d) : Qospf LS Update packet is obtained from (%d.%d)\n",
					   RouterId.NetNum, RouterId.NodeNum, 
					   obt_pk_srcnet, obt_pk_srcnode );
		InstNoLsa++;
		QospfLsUpdatePk_pro(ReceivedQospfPkptr);
		break;
        case MOSPFLSUPDATEPK  : 
		VL MPDL printf("Time: %f ", op_sim_time() ); 
        	VL MPDL printf("ROUTER ID(%d.%d) : Mospf LS Update packet is obtained from (%d.%d)\n",
					   RouterId.NetNum, RouterId.NodeNum, 
					   obt_pk_srcnet, obt_pk_srcnode );
		InstNoLsa++;
		MospfLsUpdatePk_pro(ReceivedQospfPkptr);
		break;
		/*
        case DDPK  : 
		DataDescPk_pro(ReceivedQospfPkptr);
                break;
        case LSREQPK  : 
		LsReqPk_pro(ReceivedQospfPkptr);
                break;
        case LSACKPK  : 
		LsAckPk_pro(ReceivedQospfPkptr);
                break;
        default : break;
        */
    }
	op_pk_destroy( ReceivedQospfPkptr );
}



/**
*** Function : Path_Msg_Pro.
*** Input    : A pointer to PathMessage packet that is obtained from the Rsvp stream
*** Output   : This function gets the source, destination and sender specs information,
***            but doesn't invoke qospfcalc.
*** Author   : Gang Duan 11/18
**/
void
Path_Msg_Pro( Packet *RsvpPkptr )
{
    PathMessage *pm;

    int sid;
    
/*    pm = (PathMessage *)op_prg_mem_alloc( sizeof(PathMessage) ); */
/*    memset( pm, 0, sizeof(PathMessage) ); */
    op_pk_nfd_get( RsvpPkptr, "PathMsg", &pm );

    sid = pm -> SessObj.DestAddr.NodeNum;
    SessionDirectory[sid].SenderAddr = pm->SenderDes.SenderTempObj.SrcAddr;
    SessionDirectory[sid].DestAddr = pm->SessObj.DestAddr;
    SessionDirectory[sid].SenderPort = pm->SenderDes.SenderTempObj.SrcPort;
    SessionDirectory[sid].Bandwidth = pm->SenderDes.SenderTspecObj.Bandwidth;
    SessionDirectory[sid].Delay = pm->SenderDes.SenderTspecObj.LinkDelay;

    CurrSess = sid;

VL QPDL printf(" &&&&& (%d.%d) Path_Msg_Pro Time=%fs Session=%d Sender=(%d.%d) Recver=(%d.%d)\n",
                RouterId.NetNum,RouterId.NodeNum, op_sim_time(), sid,
		SessionDirectory[sid].SenderAddr.NetNum, SessionDirectory[sid].SenderAddr.NodeNum, 
		SessionDirectory[sid].DestAddr.NetNum, SessionDirectory[sid].DestAddr.NodeNum );


    op_prg_mem_free( pm );

/* Invoke LSA update */
 op_intrpt_schedule_self( op_sim_time(), 5 );
}


/**
*** Function : Resv_Msg_Pro.
*** Input    : A pointer to PathMessage packet that is obtained from the Rsvp stream
*** Output   : This function gets the resources reserved infomation from the TCSB 
*** Author   : Gang Duan 11/18
**/
void
Resv_Msg_Pro( Packet *RsvpPkptr )
{
    ResvMessage *rm;
 
    int sid;
 
/*    rm = (ResvMessage *)op_prg_mem_alloc( sizeof(ResvMessage) ); */
/*    memset( rm, 0, sizeof(ResvMessage) ); */
    op_pk_nfd_get( RsvpPkptr, "ResvMsg", &rm );
 
    sid = rm -> SessObj.DestAddr.NodeNum;
    SessionDirectory[sid].SenderAddr = rm->FlowDesList[0].FilterSpecObj.SrcAddr;
    SessionDirectory[sid].DestAddr = rm->SessObj.DestAddr;
    SessionDirectory[sid].SenderPort = rm->FlowDesList[0].FilterSpecObj.SrcPort;
    SessionDirectory[sid].Bandwidth = rm->FlowDesList[0].FlowSpecObj.ReqBandwidth;
    SessionDirectory[sid].Delay = rm->FlowDesList[0].FlowSpecObj.ReqLinkDelay;

    CurrSess = sid;

VL QPDL printf(" %%%%% (%d.%d) Resv_Msg_Pro Time=%fs Session=%d Sender=(%d.%d) Recver=(%d.%d)\n",
                RouterId.NetNum,RouterId.NodeNum, op_sim_time(), sid,
		SessionDirectory[sid].SenderAddr.NetNum, SessionDirectory[sid].SenderAddr.NodeNum, 
		SessionDirectory[sid].DestAddr.NetNum, SessionDirectory[sid].DestAddr.NodeNum );
 
    op_prg_mem_free( rm );

/* invoke LSA update */
 op_intrpt_schedule_self( op_sim_time(), 5 );
}


/**
*** Function : RsvpPkPro.
*** Input    : A pointer to Rsvp packet that is obtained from the Rsvp stream
*** Output   : This function will ckeck for the Rsvp packet type that it has recived and call the
***            appropriate functions accordingly.
**/
void
RsvpPkPro( Packet * ReceivedRsvpPkptr )
{
    int tmpMsgType;

    op_pk_nfd_get(ReceivedRsvpPkptr, "Msg_Type", &tmpMsgType);
 
    switch(tmpMsgType) {
         case 1  : 	/* PATH_MESSAGE */
VL QPDL		printf("%d.%d Received a Path Message from Local Rsvp Deamon \n",
			RouterId.NetNum, RouterId.NodeNum );
		Path_Msg_Pro(ReceivedRsvpPkptr);
		break;
         case 2  : 	/* RESV_MESSAGE */
VL QPDL		printf("%d.%d Received a Resv Message from Local Rsvp Deamon \n",
			RouterId.NetNum, RouterId.NodeNum );
		Resv_Msg_Pro(ReceivedRsvpPkptr);
                break;
         case 7  : 
		 /*  ResvConf_Msg_pro(ReceivedRsvpPkptr);  */
                   break;
         default : break;
    }
	op_pk_destroy(ReceivedRsvpPkptr);
}

/*
void
CalculateNextHops( Vertex_Struct* VertexW, Vertex_Struct* VertexV , RouterLinkAdv* WRouterLsa )
{

    int Hop_Count = 0;
	int j;
    if( !(VertexW->VertexId.NetNum == RouterId.NetNum || 
          VertexW->VertexId.NodeNum == RouterId.NodeNum) ) 
        if( IsNetwork(VertexW.VertexId) ) 
            for( j = 0; j < WRouterLsa.No_of_Links; j++ )
                if( WRouterLsa->Link_Info[j].LinkId.NetNum ==
                                                VertexV->VertexId.NetNum &&
                    WRouterLsa->Link_Info[j].LinkId.NodeNum ==
                                                VertexV->VertexId.NodeNum )
                {
                    VertexW->NextHopsList[Hop_Count] =
                                    WRouterLsa->Link_Info[j].LinkData;
                    Hop_Count++;
                }
	VL	SPFDL	
	{
		printf("My Vertes Id is : %d.%d and my parent Vertex Id is : %d.%d\n",
				VertexW->VertexId.NetNum, VertexW->VertexId.NodeNum, 
				VertexV->VertexId.NetNum, VertexV->VertexId.NodeNum );
		printf("NextHopsList is : \n");
		for( j = 0; j < Hop_Count; j++ )
		{
			printf( "Hop Count is %d and the Next Hop is %d.%d \n", j, 
					 VertexW->NextHopsList[j].NetNum, VertexW->NextHopsList[j].NodeNum );
		}
	}
	printf( "!!!!!!!!!!!!!!!   End CalculateNextHops   !!!!!!!!!!!!\n" );
}
*/

/* changed by Jiemei: the parameter should be pointer*/
void
CalculateNextHops( Vertex_Struct* VertexW, Vertex_Struct* VertexV , RouterLinkAdv* VRouterLsa )
{

    int Hop_Count = 0;
	int j;
	int FoundFlag = OFF;
    /*  if( !(VertexW->VertexId.NetNum == RouterId.NetNum ||  */
          /*  VertexW->VertexId.NodeNum == RouterId.NodeNum) )  */
        /*  if( IsNetwork(VertexW.VertexId) )  */
            for( j = 0; j < VRouterLsa->No_of_Links; j++ )
			{
                if( VRouterLsa->Link_Info[j].LinkId.NetNum ==
                                                VertexW->VertexId.NetNum &&
                    VRouterLsa->Link_Info[j].LinkId.NodeNum ==
                                                VertexW->VertexId.NodeNum )
                {
                    VertexW->NextHopsList[Hop_Count] =
                                   VRouterLsa->Link_Info[j].LinkData;
                    Hop_Count++;
					FoundFlag = ON;
                }
			}
			if( FoundFlag = OFF )
			{
				VertexW->NextHopsList[Hop_Count] = VertexV->VertexId;
				Hop_Count++;
			}
	VL	SPFDL	
	{
		printf("My Vertes Id is : %d.%d and my parent Vertex Id : %d.%d\n",
				VertexW->VertexId.NetNum, VertexW->VertexId.NodeNum, 
				VertexV->VertexId.NetNum, VertexV->VertexId.NodeNum );
		printf("NextHopsList is : \n");
		for( j = 0; j < Hop_Count; j++ )
		{
			printf( "Hop Count is %d and the Next Hop is %d.%d \n", j, 
					 VertexW->NextHopsList[j].NetNum, VertexW->NextHopsList[j].NodeNum );
		}
	}
}
int 
IsNetwork( IPAddress ipa )
{
	if(ipa.NetNum < 60)
		return 1;
	else 
		return 0;
}

void
OspfCandidateAddPro( Vertex_Struct* VertexV, ListL* CandidatesList )
{
	int i = 0;
	Vertex_Struct VertexW;  
	RouterLinkAdv WRouterLsa;
	RouterLinkAdv VRouterLsa;
	Vertex_Struct* CandidateEle;
	Vertex_Struct* tmpCandEle;
	int CandidateFoundFlag = OFF;

DL printf("*******************OspfCandidateAddPro begin***********************\n");

	memset( &VRouterLsa, 0, sizeof(RouterLinkAdv) );

	for(i = 0; i< MAX_NO_LSAS; i++ ) /* found by Sitaram */
	{

DL printf("AreaStructure.RouterLsaList[%d].LinkStateHeader.AdvRouter=%d.%d\n",i,
       		AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum,
        	AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NodeNum);
DL printf("VertexV->VertexId=%d.%d\n",VertexV->VertexId.NetNum,VertexV->VertexId.NodeNum);

		/*if( AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum ==
		VertexV->VertexId.NetNum &&
		AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NodeNum ==
		VertexV->VertexId.NodeNum  )    commented by Jiemei*/

		/* added by Jiemei*/
		if( AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum ==
		VertexV->VertexId.NetNum) 
		{
			op_prg_mem_copy( &AreaStructure.RouterLsaList[i], &VRouterLsa, sizeof(RouterLinkAdv) );
			break;
		}
	}

	for(i = 0; i < VRouterLsa.No_of_Links; i++ )
	{

DL printf("VRouterLsa.Link_Info[%d].Type=%d  LinkId=%d.%d  LinkData=%d.%d\n",i,
		VRouterLsa.Link_Info[i].Type,
		VRouterLsa.Link_Info[i].LinkId.NetNum,VRouterLsa.Link_Info[i].LinkId.NodeNum,
		VRouterLsa.Link_Info[i].LinkData.NetNum,VRouterLsa.Link_Info[i].LinkData.NodeNum);

		if( VRouterLsa.Link_Info[i].Type == ROUTER ||
		VRouterLsa.Link_Info[i].Type == TRANSITNET )
		{
			int FoundFlag = OFF;
			int j = 0;
			op_prg_mem_copy( &VRouterLsa.Link_Info[i].LinkId, &VertexW.VertexId,
							  sizeof(IPAddress) );
			j = 0;
			while(!(AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NetNum == 0 || 
				AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NodeNum == 0) )
			{
				/*if( AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NetNum ==
					VertexW.VertexId.NetNum &&
					AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NodeNum ==
					VertexW.VertexId.NodeNum  )   commented by Jiemei*/
				if( AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NetNum ==
					VertexW.VertexId.NetNum)
				{
					FoundFlag = ON;
					op_prg_mem_copy( &AreaStructure.RouterLsaList[j], &WRouterLsa,
									 sizeof(RouterLinkAdv) );
					break;
				}
				j++;
			}
			if(FoundFlag == ON)
			{
				int WonSPF_Flag = OFF;
				j = 0;
				while( !(AreaStructure.ShortestPathTree.SPVertex[j].VertexId.NetNum == 0 || 
					 AreaStructure.ShortestPathTree.SPVertex[j].VertexId.NodeNum == 0 ) )
				{
					if( AreaStructure.ShortestPathTree.SPVertex[j].VertexId.NetNum ==
						VertexW.VertexId.NetNum&&
						AreaStructure.ShortestPathTree.SPVertex[j].VertexId.NodeNum ==
						VertexW.VertexId.NodeNum )
					{
						WonSPF_Flag = ON;
						break;
					}
					j++;
				}
				if( WonSPF_Flag == OFF )
				{
					if( VRouterLsa.Link_Info[i].TOSMetric != LSInfinity )
					{
						VertexW.DistanceToRoot = VertexV->DistanceToRoot +
									 VRouterLsa.Link_Info[i].TOSMetric;
					}
	
					/* added by Jiemei*/
					CandidateFoundFlag = OFF;

					if( !IsListEmpty(CandidatesList) )
					{
						CandidateEle = (Vertex_Struct*)GetNextListEle( CandidatesList, 1 );
						while( CandidateEle )
						{
							if( CandidateEle->VertexId.NetNum == VertexW.VertexId.NetNum &&
							    CandidateEle->VertexId.NodeNum == VertexW.VertexId.NodeNum )
							{
								CandidateFoundFlag = ON;
								break;
							}
							CandidateEle = (Vertex_Struct*)GetNextListEle( CandidatesList, 0 );
						}
					}
					if ( CandidateFoundFlag == ON )
					{
						if( CandidateEle->DistanceToRoot == VertexW.DistanceToRoot )
						{
							/*  CalculateNextHops( &VertexW, VertexV, &WRouterLsa );  */
							CalculateNextHops( &VertexW, VertexV, &VRouterLsa ); 
						}
						else if( CandidateEle->DistanceToRoot > VertexW.DistanceToRoot  )
						{
							CandidateEle->VertexId = VertexW.VertexId;
							CandidateEle->DistanceToRoot= VertexW.DistanceToRoot;
							/*  CalculateNextHops( &VertexW, VertexV, &WRouterLsa );  */
							CalculateNextHops( &VertexW, VertexV, &VRouterLsa ); 
						}
					}
					if( CandidateFoundFlag == OFF )
					{
						/*  CalculateNextHops( &VertexW, VertexV, &VRouterLsa );   */
						int Hop_Count = 0;
						int hopfound = OFF;
						int k;
						for( k = 0; k < VRouterLsa.No_of_Links; k++ )
						{
							/* commented by Jiemei
							if(VertexW.VertexId.NetNum==VRouterLsa.Link_Info[i].LinkId.NetNum &&
							   VertexW.VertexId.NetNum == VRouterLsa.Link_Info[i].LinkId.NodeNum )*/
							/*added by Jiemei*/
							if(VertexW.VertexId.NetNum==VRouterLsa.Link_Info[k].LinkId.NetNum &&
							   VertexW.VertexId.NodeNum==VRouterLsa.Link_Info[k].LinkId.NodeNum)
							{
								op_prg_mem_copy(&VRouterLsa.Link_Info[k].LinkData,
										&VertexW.NextHopsList[Hop_Count],
										sizeof(IPAddress));
								Hop_Count++;
								hopfound = ON;
							}
						}
						if(hopfound == OFF )
						{
							op_prg_mem_copy(&VertexV->VertexId, 
									&VertexW.NextHopsList[Hop_Count],
									sizeof(IPAddress));
						}
						VL	SPFDL	printf("VertexW.NextHopsList[%d] = %d.%d DistanceToRoot=%d\n",
									Hop_Count-1,
									VertexW.NextHopsList[Hop_Count-1].NetNum,
									VertexW.NextHopsList[Hop_Count-1].NodeNum, 
									VertexW.DistanceToRoot);
						AddListEle( CandidatesList, (char *)&VertexW, sizeof(Vertex_Struct) );
						/*  CalculateNextHops( &VertexW, VertexV, &WRouterLsa );  */
					}
				}
			}
		}
	}
DL printf("*******************OspfCandidateAddPro end***********************\n");
}

void
MospfCandidateAddPro(int GrpId, Vertex_Struct* VertexV, ListL* CandidatesList )
{
	int i = 0;
	Vertex_Struct VertexW;  
	RouterLinkAdv WRouterLsa;
	RouterLinkAdv VRouterLsa;
	Vertex_Struct* CandidateEle;
	Vertex_Struct* tmpCandEle;
	int CandidateFoundFlag = OFF;

MDL printf("*******************MOspfCandidateAddPro begin***********************\n");

	memset( &VRouterLsa, 0, sizeof(RouterLinkAdv) );

	for(i = 0; i< MAX_NO_LSAS; i++ ) /* found by Sitaram */
	{

MDL printf("AreaStructure.RouterLsaList[%d].LinkStateHeader.AdvRouter=%d.%d\n",i,
       		AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum,
        	AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NodeNum);
MDL printf("VertexV->VertexId=%d.%d\n",VertexV->VertexId.NetNum,VertexV->VertexId.NodeNum);

		/* added by Jiemei*/
		if( AreaStructure.RouterLsaList[i].LinkStateHeader.AdvRouter.NetNum ==
		VertexV->VertexId.NetNum) 
		{
			op_prg_mem_copy( &AreaStructure.RouterLsaList[i], &VRouterLsa, sizeof(RouterLinkAdv) );
			break;
		}
	}

	for(i = 0; i < VRouterLsa.No_of_Links; i++ )
	{

MDL printf("VRouterLsa.Link_Info[%d].Type=%d  LinkId=%d.%d  LinkData=%d.%d\n",i,
		VRouterLsa.Link_Info[i].Type,
		VRouterLsa.Link_Info[i].LinkId.NetNum,VRouterLsa.Link_Info[i].LinkId.NodeNum,
		VRouterLsa.Link_Info[i].LinkData.NetNum,VRouterLsa.Link_Info[i].LinkData.NodeNum);

		if( VRouterLsa.Link_Info[i].Type == ROUTER ||
		VRouterLsa.Link_Info[i].Type == TRANSITNET )
		{
			int FoundFlag = OFF;
			int j = 0;
			op_prg_mem_copy( &VRouterLsa.Link_Info[i].LinkId, &VertexW.VertexId,
							  sizeof(IPAddress) );
			j = 0;
			while(!(AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NetNum == 0 ||
				AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NodeNum == 0 ))
			{
				if( AreaStructure.RouterLsaList[j].LinkStateHeader.AdvRouter.NetNum ==
					VertexW.VertexId.NetNum)
				{
					FoundFlag = ON;

					op_prg_mem_copy(&AreaStructure.RouterLsaList[j], 
								&WRouterLsa,
								sizeof(RouterLinkAdv) );
					break;
				}
				j++;
			}
MDL printf("FoundFlag=%d\n",FoundFlag);
			if(FoundFlag == ON)
			{
				int WonSPF_Flag = OFF;
				j = 0;
				while(!(AreaStructure.MospfShortestPathTree[GrpId].SPVertex[j].VertexId.NetNum == 0 ||
					AreaStructure.MospfShortestPathTree[GrpId].SPVertex[j].VertexId.NodeNum == 0))
				{
/* if SPT.vertext == RouterLSA.link_info.linkid */
					if( AreaStructure.MospfShortestPathTree[GrpId].SPVertex[j].VertexId.NetNum ==
						VertexW.VertexId.NetNum &&
					    AreaStructure.MospfShortestPathTree[GrpId].SPVertex[j].VertexId.NodeNum ==
                                                VertexW.VertexId.NodeNum )
					{
						WonSPF_Flag = ON;
						break;
					}
					j++;
				}
MDL printf("WonSPF_Flag=%d\n",WonSPF_Flag);
				if( WonSPF_Flag == OFF )
				{
					if( VRouterLsa.Link_Info[i].TOSMetric != LSInfinity )
					{
						VertexW.DistanceToRoot = VertexV->DistanceToRoot +
									 VRouterLsa.Link_Info[i].TOSMetric;
					}
	
					/* added by Jiemei*/
					CandidateFoundFlag = OFF;

					if( !IsListEmpty(CandidatesList) )
					{
						CandidateEle = (Vertex_Struct*)GetNextListEle( CandidatesList, 1 );
						while( CandidateEle )
						{
							if( CandidateEle->VertexId.NetNum == VertexW.VertexId.NetNum)
							{
								CandidateFoundFlag = ON;
								break;
							}
							CandidateEle = (Vertex_Struct*)GetNextListEle( CandidatesList, 0 );
						}
MDL printf("CandidateFoundFlag=%d\n",CandidateFoundFlag);
					}
					if ( CandidateFoundFlag == ON )
					{
						if( CandidateEle->DistanceToRoot == VertexW.DistanceToRoot )
						{
							/*  CalculateNextHops( &VertexW, VertexV, &WRouterLsa );  */
							CalculateNextHops( &VertexW, VertexV, &VRouterLsa ); 
						}
						else if( CandidateEle->DistanceToRoot > VertexW.DistanceToRoot  )
						{
							CandidateEle->VertexId = VertexW.VertexId;
							CandidateEle->DistanceToRoot= VertexW.DistanceToRoot;
							/*  CalculateNextHops( &VertexW, VertexV, &WRouterLsa );  */
							CalculateNextHops( &VertexW, VertexV, &VRouterLsa ); 
						}
					}
					if( CandidateFoundFlag == OFF )
					{
						/*  CalculateNextHops( &VertexW, VertexV, &VRouterLsa );   */
						int Hop_Count = 0;
						int hopfound = OFF;
						int k;
						for( k = 0; k < VRouterLsa.No_of_Links; k++ )
						{
							if(VertexW.VertexId.NetNum==VRouterLsa.Link_Info[i].LinkId.NetNum )
						/*	VertexW.VertexId.NodeNum==VRouterLsa.Link_Info[i].LinkId.NodeNum ) DG 10/23 */
							{
								op_prg_mem_copy(&VRouterLsa.Link_Info[k].LinkData,
										&VertexW.NextHopsList[Hop_Count],
										sizeof(IPAddress));
								Hop_Count++;
								hopfound = ON;
							}
						}
MDL printf("hopfound=%d\n",hopfound);
						if(hopfound == OFF )
						{
							op_prg_mem_copy(&VertexV->VertexId, 
									&VertexW.NextHopsList[Hop_Count],
									sizeof(IPAddress));
						}
						VL	SPFDL	printf("VertexW.NextHopsList[%d] = %d.%d DistanceToRoot=%d\n",
									Hop_Count-1,
									VertexW.NextHopsList[Hop_Count-1].NetNum,
									VertexW.NextHopsList[Hop_Count-1].NodeNum, 
									VertexW.DistanceToRoot);
						AddListEle( CandidatesList, (char *)&VertexW, sizeof(Vertex_Struct) );
						/*  CalculateNextHops( &VertexW, VertexV, &WRouterLsa );  */
					}
				}
			}
		}
	}
MDL printf("*******************MOspfCandidateAddPro end***********************\n");
}

void 
OspfSPTreeBuild( ListL* CandidateList )
{
	int Inx, i;
	Vertex_Struct VertexW;

    if( !IsListEmpty( CandidateList ) )
    {
        int LeastDistance = 10000;
        int IsLessThanFlag = OFF;
        int FoundFlag = OFF;
        Vertex_Struct* tmpCandidateEle;

        tmpCandidateEle = (Vertex_Struct*)GetNextListEle( CandidateList, 1 );
        while( tmpCandidateEle )
        {
/* Added, 10/23 */
            IsLessThanFlag = OFF;
            FoundFlag = OFF;


VL OPDL printf("tmpCandidateEle->VertexId=%d.%d  DistanceToRoot=%d\n",
		tmpCandidateEle->VertexId.NetNum,tmpCandidateEle->VertexId.NodeNum,
		tmpCandidateEle->DistanceToRoot);
            LeastDistance = tmpCandidateEle->DistanceToRoot;
            Inx = 0;
            while( !(AreaStructure.ShortestPathTree.SPVertex[Inx].VertexId.NetNum == 0 || 
                     AreaStructure.ShortestPathTree.SPVertex[Inx].VertexId.NodeNum == 0) )
            {
VL OPDL printf("AreaStructure.ShortestPathTree.SPVertex[%d].VertexId=%d,%d\n",Inx,
		AreaStructure.ShortestPathTree.SPVertex[Inx].VertexId.NetNum,
		AreaStructure.ShortestPathTree.SPVertex[Inx].VertexId.NodeNum);
	

		/* added by Jiemei */
		FoundFlag = OFF;
		IsLessThanFlag=OFF;

                if( AreaStructure.ShortestPathTree.SPVertex[Inx].VertexId.NetNum ==
                    tmpCandidateEle->VertexId.NetNum &&
                    AreaStructure.ShortestPathTree.SPVertex[Inx].VertexId.NodeNum ==
                    tmpCandidateEle->VertexId.NodeNum )
                {
                    FoundFlag = ON;
                    break;
                }
 
                if( AreaStructure.ShortestPathTree.SPVertex[Inx].DistanceToRoot > LeastDistance )
                {
                    IsLessThanFlag = ON;
                    break;
                }
                Inx++;
            }
VL OPDL printf("FoundFlag=%d     IsLessThanFlag=%d\n",FoundFlag,IsLessThanFlag);
            if( FoundFlag == ON )
                if( AreaStructure.ShortestPathTree.SPVertex[Inx].DistanceToRoot > LeastDistance )
                    AreaStructure.ShortestPathTree.SPVertex[Inx].DistanceToRoot = LeastDistance;
            if( FoundFlag == OFF )
            {
                if(IsLessThanFlag == ON)
                {
                    Vertex_Struct tmpSPVertex;
                    Vertex_Struct sectmpSPVertex;
                    
		    tmpSPVertex = AreaStructure.ShortestPathTree.SPVertex[Inx];
                    op_prg_mem_copy( tmpCandidateEle, &(AreaStructure.ShortestPathTree.SPVertex[Inx]),
                                     sizeof(Vertex_Struct) );
	       	    while(!(AreaStructure.ShortestPathTree.SPVertex[Inx].VertexId.NetNum == 0 || 
			    AreaStructure.ShortestPathTree.SPVertex[Inx].VertexId.NodeNum == 0) )
                    /*  for(i = Inx; i < MAX_NO_SPROUTERS; i++)  */
                    {
                        sectmpSPVertex = AreaStructure.ShortestPathTree.SPVertex[i+1];
                        AreaStructure.ShortestPathTree.SPVertex[i+1] = tmpSPVertex;
                        tmpSPVertex = sectmpSPVertex;
			Inx++;
                    }
                }
                if( IsLessThanFlag == OFF )
                    op_prg_mem_copy( tmpCandidateEle, &(AreaStructure.ShortestPathTree.SPVertex[Inx]),
                                     sizeof(Vertex_Struct) );
                op_prg_mem_copy( tmpCandidateEle, &VertexW, sizeof(Vertex_Struct) );

VL OPDL printf("copy %d.%d to AreaStructure\n",VertexW.VertexId.NetNum,VertexW.VertexId.NodeNum); 

                OspfCandidateAddPro( &VertexW , CandidateList );
            }
            DeleteListEle(CandidateList, (char*)tmpCandidateEle, sizeof(Vertex_Struct) );
            tmpCandidateEle = (Vertex_Struct*)GetNextListEle( CandidateList, 1 );
        }
    }
}

void
RoutingTableCalc(Area_Struct* tmpAreaStructure, int* rte_table )
{
	int i = 1;
	while(!(tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum == 0 || 
		tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NodeNum == 0) )
	{ 	
		int j = 0;
		if(!(tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NetNum == 0 || 
		     tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NodeNum == 0) )
		{
			int k = 0;
			int RouteFoundFlag = OFF;
			for( k = 0; k < NoOfLinks; k++ )
			{
				/*
				if( tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NetNum ==  
					LinksInfo[k].IntfIPAddr.NetNum &&
					tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NodeNum ==
					LinksInfo[k].IntfIPAddr.NodeNum )
				*/
				if( tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum ==  
					LinksInfo[k].NeighborRtrId.NetNum) 
/*					|| tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum ==
					LinksInfo[k].NghborNetNum )   commented by Jiemei */
				{
					rte_table[tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum] = k+1;
					RouteFoundFlag = ON;
					break;
				}
			}
			if( RouteFoundFlag == OFF )
			{
				if( rte_table[tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NetNum] != 0 )
				{
					rte_table[tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum] = 
						 rte_table[tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NetNum];
				}
			}
		}
		i++;
	}

	if( !IS_AREA_BORDER_ROUTER )
		rte_table[DEF_STREAM] = rte_table[AreaBorderRouters[AreaId].NetNum];
	else 
		rte_table[DEF_STREAM] = UNDEFINED;

	rte_table[net_num0] = 0;

	VL	SPFDL
	{
		printf("---------- (%d.%d) Routing Table -------------- \n", 
				RouterId.NetNum, RouterId.NodeNum );
		for(i = 0; i < 200; i++ )
		{
			if( rte_table[i] != -1 )
			{
				printf("Route to the net %d is through the stream %d \n", i, rte_table[i] );
			}
		}
	}
}
void
QospfRouterInitialize(int ini)
{
	int i,j;

	VL	ILDL	printf("Router Id is : %d.%d\n", RouterId.NetNum, RouterId.NodeNum );

	for(i = ini; i < NoOfLinks; i++)
	{
		q_id = op_topo_assoc(ip_objid, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_QUEUE, i);
		tx_id = op_topo_assoc( q_id, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_PTTX, 0 );
		dup_lk_id = op_topo_assoc( tx_id, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_LKSIMP, 0 );
		rx_id = op_topo_assoc( dup_lk_id, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_PTRX, 0 );
		ip_id = op_topo_assoc( rx_id, OPC_TOPO_ASSOC_OUT, OPC_OBJTYPE_QUEUE, 0 );
		qospf_id = op_topo_assoc( ip_id, OPC_TOPO_ASSOC_OUT, OPC_OBJMTYPE_MODULE, 13 );
		/*  op_ima_obj_attr_get ( ip_id, "Indep_Rou", &LinksInfo[i].IsNghborRou );  */
		/*  op_ima_obj_attr_get ( qospf_id, "AreaBorderRouter", &LinksInfo[i].IsNghborRou );  */
		/*  op_ima_obj_attr_get ( qospf_id, "IsRouter", &LinksInfo[i].IsNghborRou );   */
		LinksInfo[i].IsNghborRou = 1;
		if( LinksInfo[i].IsNghborRou == 1 )
		{
			op_ima_obj_attr_get(ip_id,"net_num0",&Network_Routers[i].NetNum);
			op_ima_obj_attr_get(ip_id,"node_num0",&Network_Routers[i].NodeNum);

			LinksInfo[i].Types = ROUTER;
			op_ima_obj_attr_get(ip_id,"net_num0",&LinksInfo[i].NeighborRtrId.NetNum);
			op_ima_obj_attr_get(ip_id,"node_num0",&LinksInfo[i].NeighborRtrId.NodeNum);
		}
		else
		{
			LinksInfo[i].Types = STUBNET;
			op_ima_obj_attr_get( ip_id, "net_num0", &LinksInfo[i].NeighborRtrId.NetNum );
			IS_DESIGNATED_ROUTER = 1;
		}
	 
		LinksInfo[i].Bandwidth = LINKBANDWIDTH;
		LinksInfo[i].Delay = LINKDELAY;
		LinksInfo[i].BufferSize = LINKBUFFERSIZE;

/* Add 3 line by DG 11/19 */
		LinksInfo[i].ReservedBandwidth = LINKBANDWIDTH;
		LinksInfo[i].ReservedDelay = LINKDELAY;
		LinksInfo[i].ReservedBufferSize = LINKBUFFERSIZE;

		switch(i)
		{
			case 0  : LinksInfo[i].IntfIPAddr.NetNum = net_num0;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num0;
				  op_ima_obj_attr_get( my_id, "Link1Cost", &LinksInfo[i].CostMetric );
				  break;
			case 1  : LinksInfo[i].IntfIPAddr.NetNum = net_num1;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num1;
				  op_ima_obj_attr_get( my_id, "Link2Cost", &LinksInfo[i].CostMetric );
				  break;
			case 2  : LinksInfo[i].IntfIPAddr.NetNum = net_num2;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num2;
				  op_ima_obj_attr_get( my_id, "Link3Cost", &LinksInfo[i].CostMetric );
				  break;
			case 3  : LinksInfo[i].IntfIPAddr.NetNum = net_num3;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num3;
				  op_ima_obj_attr_get( my_id, "Link4Cost", &LinksInfo[i].CostMetric );
				  break;
			case 4  : LinksInfo[i].IntfIPAddr.NetNum = net_num4;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num4;
				  op_ima_obj_attr_get( my_id, "Link4Cost", &LinksInfo[i].CostMetric );
				  break;
			case 5  : LinksInfo[i].IntfIPAddr.NetNum = net_num5;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num5;
				  op_ima_obj_attr_get( my_id, "Link5Cost", &LinksInfo[i].CostMetric );
				  break;
			case 6  : LinksInfo[i].IntfIPAddr.NetNum = net_num6;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num6;
				  op_ima_obj_attr_get( my_id, "Link6Cost", &LinksInfo[i].CostMetric );
				  break;
			case 7  : LinksInfo[i].IntfIPAddr.NetNum = net_num7;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num7;
				  op_ima_obj_attr_get( my_id, "Link7Cost", &LinksInfo[i].CostMetric );
				  break;
			case 8  : LinksInfo[i].IntfIPAddr.NetNum = net_num8;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num8;
				  op_ima_obj_attr_get( my_id, "Link8Cost", &LinksInfo[i].CostMetric );
				  break;
			case 9  : LinksInfo[i].IntfIPAddr.NetNum = net_num9;
				  LinksInfo[i].IntfIPAddr.NodeNum = node_num9;
				  op_ima_obj_attr_get( my_id, "Link9Cost", &LinksInfo[i].CostMetric );
				  break;
			default:  printf("QospfRouterInitialize link index bad %d\n",i);
				  i=0;
				  break;
		}
	}
}

/**
*** 	Qospf Routing Table calculation functions
**/
/*
void 
QOSPFCalcTree(void )
{
	if( IsListEmpty(&CandidateList) ) return;
 
    else 
	{
        MoveClosestVertexToSPTList();
 
        AddVertexVNeighborCandList();
 
        if(VertexV->VertexType == ROUTER)
		{
            VertexVResLsa = VertexV->VertexResLsa;
            VertexVRraLsa = VertexV->VertexRraLsa;
            if(!VertexVResLsa) 
			{
                MoveClosestVertexToSPTList();
            }
            else 
			{
                for(i = 0; i < NoOfLinks; i++)
				{
                    AvailableBWOnLink =
                         VertexVResLsa->ResLinkInfo.Bandwidth - ReservedLinkBandwidth[i];
                    if(AvailableBWOnLink >= ReqiuredBWOnLink)
					{
                        LinkSPFPro( &VertexVResLsa, i );
					}
				}
			}
        QOSPFTreeCalc();
    }
	}
}
void
MoveClosestVertexToSPTList(void)
{
	VertexEle = (Vertex_Struct *) GetNextListEle( &CandidateList, 1)
	VertexV = VertexEle;
	while( VertexEle )
	{
		#ifdef ROUTE_PINNING_IS_USED
			if( VertexEle-> Reserved == TRUE )
			{
				VertexV = VertexEle;
				return;
			}
		#endif
		if( VertexEle->Cost < VertexV->Cost )
			VertexV = VertexEle;
		elseif( VertexEle->Cost = VertexV->Cost )
		{
			if( (VertexEle->VertexType == TRANSITNETWORK &&
				 VertexV->VertexType != TRANSITNETWORK) )
				VertexV = VertexEle;
			elseif( VertexEle->VertexType == TRANSITNETWORK &&
				    VertexV->VertexType == TRANSITNETWORK) &&
				    VertexEle->VertexId > VertexV->VertexId) )
				    VertexV = VertexEle;
		}
		VertexEle = (Vertex_Struct *) GetNextListEle( &CandidateList, 1)
	}

	DeleteListEle( &CandidateList, &VertexV, sizeof(Vertex_Struct) );
	AddListEle( &ShortestPathTreeList, &VertexV, sizeof(Vertex_Struct) );
}
void 
LinkSPFPro( &VertexVResLsa, i )
{
	reslsa* VertexVResLsa;
	Vertex_Struct* VertexW
	VertexW = getneighbor(LinkId);
	if( (VertexW->ResLsa.LinkId == VertexV->VertexId)||
	    (VertexW->ResLsa.ResLsaHeader.LSAge == LSMaxAge)
		(VertexW->ResLsa.ResLsaHeader.Options.MCBit == 0)
		(VertexW->ResLsa.ResLsaHeader.Options.QospfBit == 0) )
		return;
	elseif( VertexV.VertexType == ROUTER )
		DelayAssocLink[i] = VertexV->ResLsa.ResLinkInfo[i].LinkDelay;
	elseif( VertexV.VertexType == TRANSITNETWORK )
		DelayAssocLink[i] = 0;
	
	DelayToVertexW = DelayToVertexV + DelayAssocLink[i];

	if(IsVertexWInCandList()){
		if(DelayToVertexW < VertexW->Delay)
		{
			 Do the processing of the specified in mospf rfc  

		}
		elseif(DelayToVertexW == VertexW->Delay)
		{
			 Do the processing of the specified in mospf rfc  
		}
		elseif(DelayToVertexW > VertexW->Delay){
			return;
		}
	}
	else{
		AddListEle( &CandidateList, &VertexW, sizeof(Vertex_Struct) );
		Do the processing of the specified in mospf rfc  
	}
}

void
MospfRouteCalc( &ShortestPathTreeList ) 
{
	int i = 1;
	while(!(tmpAreaStructure->ShortestPathTreeList.SPVertex[i].VertexId.NetNum == 0 || 
		tmpAreaStructure->ShortestPathTreeList.SPVertex[i].VertexId.NodeNum == 0) )
	{ 	
		int j = 0;
		if(!(tmpAreaStructure->ShortestPathTreeList.SPVertex[i].NextHopsList[j].NetNum == 0 || 
			 tmpAreaStructure->ShortestPathTreeList.SPVertex[i].NextHopsList[j].NodeNum == 0) )
		{
			int k = 0;
			int RouteFoundFlag = OFF;
			for( k = 0; k < NoOfLinks; k++ )
			{
				if( tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NetNum ==  
					LinksInfo[k].IntfIPAddr.NetNum &&
					tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NodeNum ==
					LinksInfo[k].IntfIPAddr.NodeNum )

				if( tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum ==
					LinksInfo[k].NeighborRtrId.NetNum)
					|| tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum ==
					LinksInfo[k].NghborNetNum ) 
				{
					rte_table[tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum] = k+1;
					RouteFoundFlag = ON;
					break;
				}
			}
			if( RouteFoundFlag == OFF )
			{
				if( rte_table[tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NetNum] != 0 )
				{
					rte_table[tmpAreaStructure->ShortestPathTree.SPVertex[i].VertexId.NetNum] = 
							 rte_table[tmpAreaStructure->ShortestPathTree.SPVertex[i].NextHopsList[j].NetNum];
				}
				else printf("No Route to this host\n");
			}
		}
		i++;
	}
}
*/
	
void 
PrintLinkStateHdr( linkstateheader lshdr )
{
	printf("Link State Header is : \n");
	printf("LsAge        : %d\n", lshdr.LsAge );
	printf("LsType       : %d\n", lshdr.LsType );
	printf("LinkStateId  : %d.%d\n", 
			lshdr.LinkStateId.NetNum, lshdr.LinkStateId.NodeNum );
	printf("AdvRouter    : %d.%d\n", 
			lshdr.AdvRouter.NetNum, lshdr.AdvRouter.NodeNum );
	printf("ThisAreaId   : %d\n", lshdr.ThisAreaId );
	printf("LSSeqNo      : %f\n", lshdr.LSSeqNo );
	printf("LSChecksum   : %f\n", lshdr.LSChecksum );
	printf("Length       : %f\n", lshdr.Length );
}
void 
PrintLinkInfo( linkinfo* LI, int Nlinks )
{
	int i;
	printf("Router Link Information \n");
	for(i = 0; i < Nlinks; i++ )
	{
		printf("LinkId       : %d.%d\n", 
				LI[i].LinkId.NetNum, LI[i].LinkId.NodeNum );
		printf("LinkData     : %d.%d\n", 
				LI[i].LinkData.NetNum, LI[i].LinkData.NodeNum );
		printf("Type         : %d\n", LI[i].Type );
		printf("NoOfTOS      : %d\n", LI[i].NoOfTOS );
		printf("TOSMetric    : %d\n", LI[i].TOSMetric );
	}
}
void 
PrintAttRtrs( IPAddress* ip )
{
	int i = 0;
	printf("Attached Routers are : " );
	while( !(ip[i].NetNum == 0 && ip[i].NodeNum == 0) )
	{
		printf("%d.%d, ", ip[i].NetNum, ip[i].NodeNum );
		i++;
	}
}
void 
PrintTosInfo( tosinfo* TI )
{
	int i = 0;
	printf("TOS Info is : \n");
	while( !(TI[i].TOS == 0 && TI[i].metric == 0) )
	{
		printf("TI[i].TOS		: %d\n", TI[i].TOS );
		printf("TI[i].Metric	: %f\n", TI[i].metric );
		i++;
	}
}
void
PrintResLsaHdr( reslsaheader rlshdr )
{
	printf(" Resource Link State Header is : \n");
	printf("LsAge        : %d\n", rlshdr.LsAge );
	printf("Options      : %d\n", rlshdr.Options );
	printf("LSA_type     : %d\n", rlshdr.LSA_type );
	printf("LinkStateId  : %d.%d\n", 
			rlshdr.LinkStateId.NetNum, rlshdr.LinkStateId.NodeNum );
	printf("AdvRouter    : %d.%d\n", 
			rlshdr.AdvRouter.NetNum, rlshdr.AdvRouter.NodeNum );
	printf("ThisAreaId   : %d\n", rlshdr.ThisAreaId );
	printf("LSSeqNo      : %f\n", rlshdr.LSSeqNo );
	printf("LSChecksum   : %f\n", rlshdr.LSChecksum );
	printf("Length       : %f\n", rlshdr.Length );
	printf("rtype        : %f\n", rlshdr.rtype );
	printf("No_of_Links  : %f\n", rlshdr.No_of_Links );
}
void
PrintResLinkInfo( reslinkinfo* RLI )
{
	int i = 0;
	printf("Resource Link Information \n");
	while( !(RLI[i].LinkId.NetNum == 0 || RLI[i].LinkId.NodeNum == 0) )
	{
		printf("LinkId       : %d.%d\n", 
				RLI[i].LinkId.NetNum, RLI[i].LinkId.NodeNum );
		printf("LinkData     : %d.%d\n", 
				RLI[i].LinkData.NetNum, RLI[i].LinkData.NodeNum );
		printf("Type         : %d\n", RLI[i].Type );
		printf("TOSMetric    : %d\n", RLI[i].TOSMetric );
		printf("LinkDelay    : %f\n", RLI[i].LinkDelay );
		printf("Bandwidth    : %f\n", RLI[i].Bandwidth );
		printf("BufferSize   : %f\n", RLI[i].BufferSize );
		i++;
	}
}
void
PrintOpaqueLsaHdr( opaquelsaheader olshdr )
{
	printf(" Reservation Link State Header is : \n");
	printf("LsAge        : %d\n", olshdr.LsAge );
	printf("Options      : %d\n", olshdr.Options );
	printf("LSA_type     : %d\n", olshdr.LSA_type );
	printf("OpaqueType   : %d\n", olshdr.OpaqueType );
	printf("OpaqueId     : %d\n", olshdr.OpaqueId );
	printf("AdvRouter    : %d.%d\n", 
			olshdr.AdvRouter.NetNum, olshdr.AdvRouter.NodeNum );
	printf("ThisAreaId   : %d\n", olshdr.ThisAreaId );
	printf("LSSeqNo      : %f\n", olshdr.LSSeqNo );
	printf("LSChecksum   : %f\n", olshdr.LSChecksum );
	printf("Length       : %f\n", olshdr.Length );
	printf("FloddingScope: %d\n", olshdr.FloddingScope );
}
void
PrintLinkResvInfo( linkresvinfo* RLI, int Nlinks )
{
	int i;
	printf("Reserved Link Information \n");
	for(i = 0; i < Nlinks; i++ )
	{
		printf("LinkId       		: %d.%d\n", 
				RLI[i].LinkId.NetNum, RLI[i].LinkId.NodeNum );
		printf("LinkData     		: %d.%d\n", 
				RLI[i].LinkData.NetNum, RLI[i].LinkData.NodeNum );
		printf("Type         		: %d\n", RLI[i].Type );
		printf("PinFlag      		: %d\n", RLI[i].PinFlag );
		printf("ReservedBandwidth   : %f\n", RLI[i].ReservedBandwidth );
		printf("ReservedBufferSize  : %f\n", RLI[i].ReservedBufferSize );
	}
}
void
PrintVertexInfo( vertexinfo* VI )
{
	int i = 0;
	printf("Vertex Information\n");
	while( !(VI[i].VertexId.NetNum == 0 || VI[i].VertexId.NodeNum == 0) )
	{
		printf("VertexType	: %d\n", VI[i].VertexType );
		printf("VertexId	: %d.%d\n", VI[i].VertexId.NetNum,
				VI[i].VertexId.NodeNum);
		i++;	
	}
}
void
PrintLsa( char * MsgPk, int Type )
{
	RouterLinkAdv * RouterLsa;
	NetworkLinkAdv * NetworkLsa;
	SummaryLinkAdv * SummaryLsa;
	reslsa * ResourceLsa;
	rralsa * RraLsa;
	grpmbrlsa * GrpMbrShipLsa;

    switch( Type )
    {
        case ROUTERLSA :
            printf("\n---------------- (%d.%d)Router Lsa-------------------------\n",
					RouterId.NetNum, RouterId.NodeNum );
            RouterLsa = (RouterLinkAdv *)MsgPk;
			printf("RouterLsa->LinkStateHeader.AdvRouter = %d.%d\n",
					RouterLsa->LinkStateHeader.AdvRouter.NetNum, 
					RouterLsa->LinkStateHeader.AdvRouter.NodeNum );
            PrintLinkStateHdr( RouterLsa->LinkStateHeader );
            printf( "No_of_Links : %d\n", RouterLsa->No_of_Links );
            PrintLinkInfo( RouterLsa->Link_Info, RouterLsa->No_of_Links );
        break;
        case NETWORKLSA :
            printf("\n----------------Network Lsa------------------------\n ");
            NetworkLsa = (NetworkLinkAdv *)MsgPk;
            PrintLinkStateHdr( NetworkLsa->LinkStateHeader );
            printf( "ThisNetworkMask : %d\n", NetworkLsa->ThisNetworkMask );
            PrintAttRtrs( NetworkLsa->AttachedRouter );
        break;
        case SUMMARYLSA :
            printf("\n----------------Summary Lsa-------------------------\n");
            SummaryLsa = (SummaryLinkAdv *)MsgPk;
            PrintLinkStateHdr( SummaryLsa->LinkStateHeader );
            printf( "ThisNetworkMask : %d\n", SummaryLsa->ThisNetworkMask );
            PrintTosInfo( SummaryLsa->Tos_Info );
        break;
        case RESLSA :
            printf("\n----------------Resource Lsa------------------------\n");
            ResourceLsa = (reslsa *)MsgPk;
            PrintResLsaHdr( ResourceLsa->ResLsaHeader );
            PrintResLinkInfo( ResourceLsa->ResLinkInfo );
        break;
        case RRALSA :
            printf("\n----------------Rra Lsa-----------------------------\n");
            RraLsa = (rralsa *)MsgPk;
            PrintOpaqueLsaHdr( RraLsa->OpaqueLsaHeader );
			printf("DestAddr : %d.%d\n", 
				RraLsa->DestAddr.NetNum, RraLsa->DestAddr.NodeNum );
			printf("SourceAddr : %d.%d\n", 
				RraLsa->SourceAddr.NetNum, RraLsa->SourceAddr.NodeNum );
			printf("No_of_Links : %d\n", RraLsa->No_of_Links );
			PrintLinkResvInfo( RraLsa->LinkResvInfo, RraLsa->No_of_Links );
        break;
        case GRPMBRSHIPLSA :
            printf("\n----------------Group Membership Lsa----------------\n");
            GrpMbrShipLsa = (grpmbrlsa *)MsgPk;
            PrintLinkStateHdr( GrpMbrShipLsa->LinkStateHeader );
			PrintVertexInfo( GrpMbrShipLsa->VertexInfo );
        break;
        default :
            printf("\n----------------Unrecognized Message----------------\n");
            VL  printf("Could not Recognize the Message\n");
        break;
    }
    printf("-----------------------------------------------------\n");
}

